#!/bin/bash
#
#  libflame
#  An object-based infrastructure for developing high-performance
#  dense linear algebra libraries.
#
#  Copyright (C) 2008, The University of Texas
#
#  libflame is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of
#  the License, or (at your option) any later version.
#
#  libflame is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with libflame; if you did not receive a copy, see
#  http://www.gnu.org/licenses/.
#
#  For more information, please contact us at flame@cs.utexas.edu or
#  send mail to:
#
#  Field G. Van Zee and/or
#  Robert A. van de Geijn
#  The University of Texas at Austin
#  Department of Computer Sciences
#  1 University Station C0500
#  Austin TX 78712
#

#
# recursive-sed.sh
#
# Field G. Van Zee
#

print_usage()
{
	local script_name
	
	# Echo usage info
	echo " "
	echo " "$script_name
	echo " "
	echo " Field G. Van Zee"
	echo " "
	echo " Recusively descend a directory tree and perform sed commands, either on"
	echo " the filename or the file contents, or both." 
	echo " "
	echo " Usage:"
	echo "   ${script_name} [options]"
	echo " "
	echo " The following options are accepted:"
	echo " "
	echo "   -d "
	echo "                 Dry run. Go through all the motions, but don't actually"
	echo "                 apply any of the sed expressions to file names or contents."
	echo "   -h "
	echo "                 Consider hidden files and directories. Default behavior is"
	echo "                 to ignore them while recursing."
	echo "   -x "
	echo "                 Delete the original file after the new file is created."
	echo "                 Notice that this only makes sense if the file path changes"
	echo "                 (ie: if the directory and/or name changes)."
	echo "   -p pattern "
	echo "                 Specifies the filename pattern, as would be given to the"
	echo "                 ls utility, to limit which files are affected. Default is"
	echo "                 the to consider all files present."
	echo "   -r dir"
	echo "                 The root directory for the recursive action to be performed."
	echo "                 Default is to use the current working directory."
	echo "   -v [0|1|2]"
	echo "                 verboseness level"
	echo "                 level 0: silent  (no output)"
	echo "                 level 1: default (one line per directory; supress ls stderr)"
	echo "                 level 2: verbose (one line per directory; show ls stderr)"
	echo " "
	echo " At least one of the following option-argument pairs is required:"
	echo " "
	echo "   -f sed_expr "
	echo "                 Specifies the sed expression that will be applied to the"
	echo "                 filenames of the files touched by the script. This expression"
    echo "                 must be a search-and-replace pattern."
	echo "   -c sed_expr "
	echo "                 Specifies the sed expression that will be applied to the"
	echo "                 contents of the files touched by the script. This expression"
    echo "                 should be a search-and-replace pattern."
	echo "   -s sed_script"
	echo "                 Specifies an arbitrary sed script that will be applied to the"
	echo "                 file contents of the files touched by the script."
	echo " "
	echo " Note: -c and -s options are mutually exclusive."
	echo " "
	
	# Exit with non-zero exit status
	exit 1
}



main()
{
	# Variables set by getopts.
	dry_run_flag=""
	hidden_files_dirs_flag=""
	delete_original_flag=""
	filename_pattern=""
	root_dir=""
	verbose_level=""
	filename_sed_expr=""
	contents_sed_expr=""
	contents_sed_script=""

	recursive_flag="1"	

	
	# Get the script name
	script_name=${0##*/}
	
	
	# Local variable declarations.
	local item sub_items this_dir
	
	
	# Process our command line options.
	while getopts ":c:df:hp:r:s:xv:" opt; do
		case $opt in
			d  ) dry_run_flag="-d" ;;
			h  ) hidden_files_dirs_flag="-h" ;;
			x  ) delete_original_flag="-x" ;;
			v  ) verbose_level="$OPTARG" ;;
			p  ) filename_pattern="$OPTARG" ;;
			r  ) root_dir="$OPTARG" ;;
			f  ) filename_sed_expr="$OPTARG" ;;
			c  ) contents_sed_expr="$OPTARG" ;;
			s  ) contents_sed_script="$OPTARG" ;;
			\? ) print_usage
		esac
	done
	shift $(($OPTIND - 1))
	
	
	# Make sure we've parsed all command line arguments by now.
	if [ $# != "0" ]; then
		echo "${script_name}: Unparsed command line arguments! Try running with no arguments for help."
		exit 1
	fi
	
	
	# Make sure we received at least one of the required options.
	if [ -z "$filename_sed_expr" ] &&
	   [ -z "$contents_sed_expr" ] &&
	   [ -z "$contents_sed_script" ]; then
		print_usage
	fi
	
	
	# Make sure that both a file contents sed expression and sed script were
	# not given.
	if [ "$contents_sed_expr"   != "" ] &&
	   [ "$contents_sed_script" != "" ] ; then
		echo "${script_name}: The -c and -s options may not be used at the same time."
		exit 1
	fi


	# Make sure that verboseness level is valid.
	if [ "$verbose_level" != "0" ] && 
	   [ "$verbose_level" != "1" ] &&
	   [ "$verbose_level" != "2" ]; then
		verbose_level="1"
	fi
	
	# Add the dry run flag.
	if [ "$dry_run_flag" != "" ]; then
		dry_run_flag="-d"
	fi
	
	# Add the delete original flag.
	if [ "$delete_original_flag" != "" ]; then
		delete_original_flag="-x"
	fi
	
	# Prepare the verboseness level arguments to perform-sed.sh.
	if [ "$verbose_level" != "" ] ; then
		verbose_level_cmd="-v $verbose_level"
	fi

	# Prepare the filename pattern arguments to perform-sed.sh.
	if [ "$filename_pattern" != "" ] ; then
		filename_pattern_cmd="-p$filename_pattern"
	else
		filename_pattern_cmd='-p*'
	fi

	# Prepare the directory arguments to perform-sed.sh.
	if [ "$root_dir" != "" ] ; then
		
		# Strip / from end of directory paths, if there is one.
		root_dir=${root_dir%/}
		
		exist_dir_cmd="-e $root_dir"
		new_dir_cmd="-n $root_dir"
	else
		root_dir=$PWD
		exist_dir_cmd="-e $PWD"
		new_dir_cmd="-n $PWD"
	fi
	
	# Prepare the filename sed expression argument to perform-sed.sh.
	if [ "$filename_sed_expr" != "" ] ; then
		filename_sed_expr_cmd="-f $filename_sed_expr"
	fi
	
	# Prepare the file contents sed expression argument to perform-sed.sh.
	if [ "$contents_sed_expr" != "" ] ; then
		contents_sed_expr_cmd="-c $contents_sed_expr"
	else
		contents_sed_expr_cmd="-c s/123/123/g"
	fi
	
	# Prepare the file contents sed script name argument to perform-sed.sh.
	if [ "$contents_sed_script" != "" ] ; then
		contents_sed_script_cmd="-s $contents_sed_script"
	fi
	

	# Call our function to perform the sed operations on the files in the
	# directory given.
	perform-sed.sh ${dry_run_flag} ${delete_original_flag} ${verbose_level_cmd} "${filename_pattern_cmd}" ${exist_dir_cmd} ${new_dir_cmd} ${filename_sed_expr_cmd} "${contents_sed_expr_cmd}" ${contents_sed_script_cmd}

	
	# If we were asked to act recursively, then continue processing
	# src_dir's contents.
	if [ -n "$recursive_flag" ]; then
		
		# Get a listing of items in the directory according to the hidden
		# files/directories flag.
		if [ -n "$hidden_files_dirs_flag" ]; then

			# Get a listing of the directories in src_dir (including hidden
			# files and directories).
			sub_items=$(ls -a $root_dir)

		else

			# Get a listing of the directories in src_dir.
			sub_items=$(ls $root_dir)
		fi
		
		# Descend into the contents of src_dir to prepend the license info.
		for item in $sub_items; do
			
			# If item is a directory, descend into it.
			if [ -d "$root_dir/$item" ] &&
			   [ "$item" != ".." ] &&
			   [ "$item" != "." ]; then
				
				this_dir=$root_dir/$item
				recursive_sed $this_dir
			fi
		done
	fi
	
	
	# Exit peacefully
	return 0
}



recursive_sed()
{
	# Local variable declarations
	local item sub_items curr_dir this_dir
	
	
	# Extract our argument
	curr_dir=$1
	
	# Build the existing and old directory strings.
	exist_dir_cmd="-e $curr_dir"
	new_dir_cmd="-n $curr_dir"
	
	
	# Call our function to perform the sed operations on the files in the
	# directory given.
	perform-sed.sh ${dry_run_flag} ${delete_original_flag} ${verbose_level_cmd} "${filename_pattern_cmd}" ${exist_dir_cmd} ${new_dir_cmd} ${filename_sed_expr_cmd} "${contents_sed_expr_cmd}" ${contents_sed_script_cmd}
	

	# Get a listing of items in the directory according to the hidden
	# files/directories flag.
	if [ -n "$hidden_files_dirs_flag" ]; then

		# Get a listing of the directories in src_dir (including hidden
		# files and directories).
		sub_items=$(ls -a $curr_dir)

	else

		# Get a listing of the directories in src_dir.
		sub_items=$(ls $curr_dir)
	fi
	
	# Descend into the contents of curr_dir to prepend the license info to 
	# files residing  in subdirectories of the current directory.
	for item in $sub_items; do
		
		# If item is a directory, descend into it.
		if [ -d "$curr_dir/$item" ] &&
		   [ "$item" != ".." ] &&
		   [ "$item" != "." ]; then
			
			this_dir=$curr_dir/$item
			recursive_sed $this_dir
		fi
	done
	
	
	# Return peacefully
	return 0
}



# The script's main entry point, passing all parameters given.
main "$@"

