/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"


#ifdef FLA_ENABLE_MULTITHREADING


typedef struct FLASH_Obj_variables
{
      // Lock to protect reading and writing to the queue of tasks.
      FLA_Lock         lock;

      // Pointer to function to execute.
      void*            func;

      // Queue of all the tasks.
      FLASH_Obj_queue* queue;
      
} FLASH_Obj_vars;


void FLASH_Obj_exec( void* func, FLASH_Obj_queue* queue )
{
   int i;
   int n_threads = FLASH_Queue_get_num_threads();
   void* (*thread_entry_point)( void* );

   // Return if there are no tasks.
   if ( queue->n_tasks == 0 )
      return;

   // Determine which function to send threads to.
   thread_entry_point = FLASH_Obj_exec_parallel;

   // The structre to save all the execution variables.
   FLASH_Obj_vars args;
   
   // Save the queue and function pointer.
   args.func  = func;
   args.queue = queue;

   // Initialize the lock.
   FLA_Lock_init( &(args.lock) );

   // Allocate and initialize the thread structures array. Here, an array of
   // FLASH_Thread structures of length n_threads is allocated and the fields
   // of each structure set to appropriate values.
   FLASH_Thread thread[n_threads];

   for ( i = 0; i < n_threads; i++ )
   {
      // Save the thread's identifier.
      thread[i].id = i;

      // Save the pointer to the necessary variables with the thread.
      thread[i].args = ( void* ) &args;

      // The pthread object, if it was even compiled into the FLASH_Thread
      // structure, will be initialized by the pthread implementation when we
      // call pthread_create() and does not need to be touched at this time.
   }

#if FLA_MULTITHREADING_MODEL == FLA_OPENMP

   // An OpenMP parallel for region spawns n_threads threads. Each thread
   // executes the work function with a different FLASH_Thread argument.
   // An implicit synchronization point exists at the end of the curly
   // brace scope.
   #pragma omp parallel for \
           private( i ) \
           shared( thread, n_threads, thread_entry_point ) \
           schedule( static, 1 ) \
           num_threads( n_threads )
   for ( i = 0; i < n_threads; ++i )
   {
      thread_entry_point( ( void* ) &thread[i] );
   }

#elif FLA_MULTITHREADING_MODEL == FLA_PTHREADS

   // Create each POSIX thread needed in addition to the main thread.
   for ( i = 1; i < n_threads; i++ )
   {
      int pthread_e_val;

      // Create thread i with default attributes.
      pthread_e_val = pthread_create( &(thread[i].pthread_obj),
                                      NULL,
                                      thread_entry_point,
                                      ( void* ) &thread[i] );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
      FLA_Error e_val = FLA_Check_pthread_create_result( pthread_e_val );
      FLA_Check_error_code( e_val );
#endif
   }

   // The main thread is assigned the role of thread 0. Here we manually
   // execute it as a worker thread.
   thread_entry_point( ( void* ) &thread[0] );

   // Wait for non-main threads to finish.
   for ( i = 1; i < n_threads; i++ )
   {
      // These two variables are declared local to this for loop since this
      // is the only place they are needed, and since they would show up as
      // unused variables if FLA_MULTITHREADING_MODEL == FLA_PTHREADS.
      // Strangely, the Intel compiler produces code that results in an
      // "unaligned access" runtime message if thread_status is declared as
      // an int. Declaring it as a long or void* appears to force the
      // compiler (not surprisingly) into aligning it to an 8-byte boundary.
      int   pthread_e_val;
      void* thread_status;

      // Wait for thread i to invoke its respective pthread_exit().
      // The return value passed to pthread_exit() is provided to us
      // via status, if one was given.
      pthread_e_val = pthread_join( thread[i].pthread_obj,
                                    ( void** ) &thread_status );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
      FLA_Error e_val = FLA_Check_pthread_join_result( pthread_e_val );
      FLA_Check_error_code( e_val );
#endif
   }

#endif

   // Destroy the lock.
   FLA_Lock_destroy( &(args.lock) );

   return;
}


void* FLASH_Obj_exec_parallel( void* arg )
{
   FLASH_Thread*   me;
   FLASH_Obj_vars* args;
   FLASH_Obj_task* t;
   FLA_Bool        condition = TRUE;

   // Interpret the thread argument as what it really is--a pointer to an
   // FLASH_Thread structure.
   me = ( FLASH_Thread* ) arg;

   // Extract the variables from the current thread.
   args = ( FLASH_Obj_vars* ) me->args;
   
   // Loop until all the tasks have committed.
   while ( condition )
   {
      t = NULL;

      FLA_Lock_acquire( &(args->lock) ); // L ***

      // If there are no more tasks, stop execution.
      if ( args->queue->n_tasks == 0 )
      {
         condition = FALSE;
      }
      else
      {
         // Dequeue the first task.
         t = args->queue->head;

         if ( args->queue->n_tasks == 1 )
         {
            // Clear the queue of its only task.
            args->queue->head = NULL;
            args->queue->tail = NULL;
         }
         else
         {
            // Adjust pointers in queue.
            args->queue->head = t->next_task;
            args->queue->head->prev_task = NULL;
         }

         // Decrement number of tasks on queue.
         args->queue->n_tasks--;
      }

      FLA_Lock_release( &(args->lock) ); // L ***

      // Execute the task.
      if ( t != NULL )
      {
         // FLASH_Axpy
         if      ( args->func == (void *) FLASH_Axpy_hierarchy )
         {
            if      ( t->direction == FLA_FLAT_TO_HIER )
            {
               FLA_Axpy( t->alpha, t->F, t->H );
            }
            else if ( t->direction == FLA_HIER_TO_FLAT )
            {
               FLA_Axpy( t->alpha, t->H, t->F );
            }
         }
         // FLASH_Copy
         else if ( args->func == (void *) FLASH_Copy_hierarchy )
         {
            if      ( t->direction == FLA_FLAT_TO_HIER )
            {
               FLA_Copy( t->F, t->H );
            }
            else if ( t->direction == FLA_HIER_TO_FLAT )
            {
               FLA_Copy( t->H, t->F );
            }
         }
         else
         {
            FLA_Check_error_code( FLA_NOT_YET_IMPLEMENTED );
         }

         // Free the task structure in parallel.
         FLA_free( t );
      }
   }

#if FLA_MULTITHREADING_MODEL == FLA_PTHREADS
   // If this is a non-main thread, then exit with a zero (normal) error code.
   // The main thread cannot call pthread_exit() because this routine never
   // returns. The main thread must proceed so it can oversee the joining of
   // the exited non-main pthreads.
   if ( me->id != 0 )
      pthread_exit( ( void* ) NULL );
#endif

   return ( void* ) NULL;
}


void FLASH_Obj_push( int direction, FLA_Obj alpha, FLA_Obj F, FLA_Obj H, 
                     FLASH_Obj_queue* queue )
{
   FLASH_Obj_task* t = (FLASH_Obj_task *) FLA_malloc( sizeof(FLASH_Obj_task) );

   // Initialize the task elements.
   t->direction = direction;
   t->alpha     = alpha;
   t->F         = F;
   t->H         = H;
   t->prev_task = NULL;
   t->next_task = NULL;
  
   // Add the task to the tail of the queue (and the head if queue is empty).
   if ( queue->n_tasks == 0 )
   {
      queue->head = t;
      queue->tail = t;
   }
   else
   {
      t->prev_task = queue->tail;
      queue->tail->next_task = t;
      queue->tail            = t;
   }

   // Increment the number of tasks.
   queue->n_tasks++;

   return;
}


#endif
