/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

// Internal array to hold error strings.
char fla_error_string[FLA_MAX_NUM_ERROR_MSGS][FLA_MAX_ERROR_MSG_LENGTH];


/* ***************************************************************************

   FLA_Error_string_for_code()

 *************************************************************************** */

char* FLA_Error_string_for_code( int code )
{
	return fla_error_string[-code];
}

/* ***************************************************************************

   FLA_Error_messages_init()

 *************************************************************************** */

void FLA_Error_messages_init( void )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
	sprintf( FLA_Error_string_for_code(FLA_INVALID_SIDE),
             "Invalid side parameter value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_UPLO),
             "Invalid uplo parameter value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_TRANS),
             "Invalid trans parameter value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_TRANS_GIVEN_DATATYPE),
	         "Invalid trans value (FLA_CONJ_TRANSPOSE|FLA_CONJ_NO_TRANSPOSE) for given non-complex object datatype" );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_CONJ),
             "Invalid conjugate parameter value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_DIRECT),
             "Invalid direction parameter value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_STOREV),
             "Invalid storev parameter value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_DATATYPE),
             "Invalid datatype value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_INTEGER_DATATYPE),
             "Invalid integer datatype value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_REAL_DATATYPE),
             "Invalid real datatype value." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_COMPLEX_DATATYPE),
             "Invalid complex datatype value." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_INTEGER),
             "Expected integer object." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_REAL),
             "Expected real object." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_COMPLEX),
             "Expected complex object." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_SQUARE),
             "Expected square matrix object." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_SCALAR),
             "Expected scalar object." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_VECTOR),
             "Expected vector object." );
	sprintf( FLA_Error_string_for_code(FLA_INCONSISTENT_DATATYPES),
             "Detected inconsistent object datatypes." );
	sprintf( FLA_Error_string_for_code(FLA_NONCONFORMAL_DIMENSIONS),
             "Detected inconsistent object dimensions." );
	sprintf( FLA_Error_string_for_code(FLA_UNEQUAL_VECTOR_LENGTHS),
             "Detected vectors of unequal length." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_HESSENBERG_INDICES),
             "Detected invalid Hessenberg ilo/ihi indices." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_VECTOR_LENGTH),
             "Detected vector of unexpected length." );
	sprintf( FLA_Error_string_for_code(FLA_NULL_POINTER),
             "Encountered NULL pointer." );
	sprintf( FLA_Error_string_for_code(FLA_SPECIFIED_OBJ_DIM_MISMATCH),
             "Specified dimensions do not match object dimensions." );
	sprintf( FLA_Error_string_for_code(FLA_SPECIFIED_OBJ_LDIM_MISMATCH),
             "Specified leading dim dimension does not match object leading dimension." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_PIVOT_TYPE),
             "Invalid pivot index type specified." );
	sprintf( FLA_Error_string_for_code(FLA_BLOCK_HOUSEH_DIM_MISMATCH),
             "Detected inconsistent dimensions between matrix and block Householder object." );
	sprintf( FLA_Error_string_for_code(FLA_MALLOC_RETURNED_NULL_POINTER),
             "malloc() returned NULL pointer." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_BASE_BUFFER_MISMATCH),
             "Detected a buffer address mismatch between adjacent objects." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECTS_NOT_VERTICALLY_ADJ),
             "Object partitions not vertically adjacent." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECTS_NOT_HORIZONTALLY_ADJ),
             "Object partitions not horizontally adjacent." );
	sprintf( FLA_Error_string_for_code(FLA_ADJACENT_OBJECT_DIM_MISMATCH),
             "Object partitions have mismatched dimensions." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECTS_NOT_VERTICALLY_ALIGNED),
             "Object partitions not vertically aligned." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECTS_NOT_HORIZONTALLY_ALIGNED),
             "Object partitions not horizontally aligned." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_FLOATING_DATATYPE),
             "Expected single or double-precision real or complex datatype value." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_FLOATING_POINT),
             "Expected single or double-precision real or complex object." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_BLOCKSIZE_VALUE),
             "Detected blocksize value less than or equal to zero." );
	sprintf( FLA_Error_string_for_code(FLA_OPEN_RETURNED_ERROR),
             "open() returned bad file descriptor." );
	sprintf( FLA_Error_string_for_code(FLA_LSEEK_RETURNED_ERROR),
             "lseek() returned error." );
	sprintf( FLA_Error_string_for_code(FLA_CLOSE_RETURNED_ERROR),
             "close() returned error." );
	sprintf( FLA_Error_string_for_code(FLA_UNLINK_RETURNED_ERROR),
             "unlink() returned error." );
	sprintf( FLA_Error_string_for_code(FLA_READ_RETURNED_ERROR),
             "read() returned error." );
	sprintf( FLA_Error_string_for_code(FLA_WRITE_RETURNED_ERROR),
             "write() returned error." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_QUADRANT),
             "Invalid quadrant parameter value." );
	sprintf( FLA_Error_string_for_code(FLA_NOT_YET_IMPLEMENTED),
             "Function or conditional branch/case not yet implemented." );
	sprintf( FLA_Error_string_for_code(FLA_EXPECTED_NONNEGATIVE_VALUE),
             "Expected non-negative value." );
	sprintf( FLA_Error_string_for_code(FLA_SUPERMATRIX_NOT_ENABLED),
             "SuperMatrix support must be enabled for this code to execute." );
	sprintf( FLA_Error_string_for_code(FLA_UNDEFINED_ERROR_CODE),
             "Undefined error code passed to FLA_Check_error_code()." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_DIAG),
             "Invalid diag parameter value." );
	sprintf( FLA_Error_string_for_code(FLA_INCONSISTENT_OBJECT_PRECISION),
             "Inconsistent precisions between objects." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_BLOCKSIZE_OBJ),
             "Encountered blocksize object containing value less than or equal to zero." );
	sprintf( FLA_Error_string_for_code(FLA_VECTOR_LENGTH_BELOW_MIN),
             "Detected vector length below pre-determined minimum." );
	sprintf( FLA_Error_string_for_code(FLA_PTHREAD_CREATE_RETURNED_ERROR),
             "pthread_create() returned error." );
	sprintf( FLA_Error_string_for_code(FLA_PTHREAD_JOIN_RETURNED_ERROR),
             "pthread_join() returned error." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_ISGN_VALUE),
             "Invalid value for isgn parameter (ie: |isgn| != 1)." );
	sprintf( FLA_Error_string_for_code(FLA_CHOL_FAILED_MATRIX_NOT_SPD),
             "FLA_Chol() failed due to negative diagonal element; matrix not SPD." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_ELEMTYPE),
             "Invalid object element type value." );
	sprintf( FLA_Error_string_for_code(FLA_POSIX_MEMALIGN_FAILED),
             "posix_memalign() returned error." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_SUBMATRIX_DIMS),
             "Invalid submatrix dimensions relative to parent matrix." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_SUBMATRIX_OFFSET),
             "Invalid submatrix offset relative to dimensions of submatrix and parent." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_SCALAR_ELEMTYPE),
             "Object element type is not FLA_SCALAR as expected." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_MATRIX_ELEMTYPE),
             "Object element type is not FLA_MATRIX as expected." );
	sprintf( FLA_Error_string_for_code(FLA_ENCOUNTERED_NON_POSITIVE_NTHREADS),
             "Encountered non-positive (zero) value for number of threads." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_CONJ_GIVEN_DATATYPE),
	         "Invalid conj value (FLA_CONJUGATE) for given non-complex object datatype" );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_COMPLEX_TRANS),
	         "Invalid complex trans parameter value" );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_REAL_TRANS),
	         "Invalid real trans parameter value" );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_BLAS_TRANS),
	         "Invalid BLAS-style trans parameter value" );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_NONCONSTANT_DATATYPE),
             "Invalid non-constant datatype value." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_NOT_NONCONSTANT),
             "Expected non-constant object." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_M_GIVEN_LDIM),
             "Given m dimension exceeds leading dimension." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_DATATYPES_NOT_EQUAL),
             "Detected unequal object datatypes." );
	sprintf( FLA_Error_string_for_code(FLA_DIVIDE_BY_ZERO),
             "Encountered request to invert zero scalar object." );
	sprintf( FLA_Error_string_for_code(FLA_OBJECT_ELEMTYPES_NOT_EQUAL),
             "Detected unequal object elemtypes." );
	sprintf( FLA_Error_string_for_code(FLA_INVALID_PIVOT_INDEX_RANGE),
             "Invalid pivot index range." );
#endif
}

/* ***************************************************************************

   FLA_Print_message()

 *************************************************************************** */

void FLA_Print_message( char *str, char *file, int line )
{
	fprintf( stderr, "\n" );
	fprintf( stderr, "libflame: %s (line %d):\n", file, line );
	fprintf( stderr, "libflame: %s\n", str );
	fflush( stderr );
}

/* ***************************************************************************

   FLA_Abort()

 *************************************************************************** */

void FLA_Abort( void )
{
	fprintf( stderr, "libflame: Aborting.\n");
	//raise( SIGABRT );
	abort();
}

void FLA_F2C( fla_abort_f )( void )
{
	FLA_Abort();
}

