/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

#ifdef FLA_ENABLE_NON_CRITICAL_CODE
#ifndef FLA_ENABLE_WINDOWS_BUILD

/* ***************************************************************************
 
   FLA_Obj_OOC_create( )
   FLA_Obj_OOC_to_INC( )
   FLA_Obj_INC_to_OOC( )

 *************************************************************************** */

FLA_Error FLA_Obj_OOC_create( FLA_Datatype datatype, int m, int n, off_t offset, int fd, FLA_Obj *obj )
{
  FLA_Obj_create_without_buffer( datatype, m, n, obj );

  obj->base->offset = offset;
  obj->base->fp     = fd;

  return FLA_SUCCESS;
}

FLA_Error FLA_Obj_OOC_to_INC( FLA_Obj obj_ooc, FLA_Obj obj_inc )
{
  void *buf_inc;
  int typesize, offset;

  buf_inc = FLA_Obj_buffer( obj_inc );

  typesize = FLA_Obj_datatype_size( FLA_Obj_datatype( obj_ooc ) );

  offset = (obj_ooc.base)->offset + 
           ( FLA_Obj_ldim(obj_ooc) * (obj_ooc.offn) + (obj_ooc.offm) );

  return FLA_Read_matrix( typesize, FLA_Obj_length(obj_ooc), FLA_Obj_width(obj_ooc),
                          (obj_ooc.base)->fp, offset*typesize, ((obj_ooc.base)->ldim),
                          (char *) buf_inc, ((obj_inc.base)->ldim) );
}

FLA_Error FLA_Obj_INC_to_OOC( FLA_Obj obj_inc, FLA_Obj obj_ooc )
{
  void* buf_inc;
  int typesize, offset;

  buf_inc = FLA_Obj_buffer( obj_inc );
  typesize = FLA_Obj_datatype_size( FLA_Obj_datatype( obj_inc ) );

  offset = (obj_ooc.base)->offset + 
           ( FLA_Obj_ldim(obj_ooc) * (obj_ooc.offn) + (obj_ooc.offm) );

  return FLA_Write_matrix( typesize, FLA_Obj_length(obj_inc), FLA_Obj_width(obj_inc),
                           (obj_ooc.base)->fp, (char *) buf_inc, ((obj_inc.base)->ldim), 
                           offset*typesize, ((obj_ooc.base)->ldim) );
}

/* ***************************************************************************

   FLA_File_create( )
   FLA_File_open( )
   FLA_File_close( )
   FLA_File_unlink( )

 *************************************************************************** */

FLA_Error FLA_File_create( char* filename, FLA_Datatype datatype, size_t size, int* fd )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;
#endif
  off_t     offset;
  off_t     r_val;

  *fd = open( filename, O_RDWR|O_CREAT|O_TRUNC, 0700 );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_file_descriptor( *fd );
  FLA_Check_error_code( e_val );
#endif

  offset = size * FLA_Obj_datatype_size( datatype );

  r_val = lseek( *fd, offset, SEEK_SET );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_lseek_result( offset, r_val );
  FLA_Check_error_code( e_val );
#endif

  return FLA_SUCCESS;
}

FLA_Error FLA_File_open( char *filename, int *fd )
{
  *fd = open( filename, O_RDWR );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;
  e_val = FLA_Check_file_descriptor( *fd );
  FLA_Check_error_code( e_val );
#endif

  return FLA_SUCCESS;
}

FLA_Error FLA_File_close( int fd )
{
  int       r_val;
  r_val = close( fd );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;
  e_val = FLA_Check_close_result( r_val );
  FLA_Check_error_code( e_val );
#endif

  return FLA_SUCCESS;
}

FLA_Error FLA_File_unlink( char *filename )
{
  int       r_val;
  r_val = unlink( filename );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;
  e_val = FLA_Check_unlink_result( r_val );
  FLA_Check_error_code( e_val );
#endif

  return FLA_SUCCESS;
}

/* ***************************************************************************

   FLA_Obj_Read_matrix( )
   FLA_Obj_Write_matrix( )

 *************************************************************************** */

FLA_Error FLA_Read( size_t size, int fd, off_t offset, void *buffer )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;
#endif
  double    time;
  off_t     r_val;
  ssize_t   r_val2;

  time = FLA_Clock();

  r_val = lseek( fd, offset, SEEK_SET );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_lseek_result( offset, r_val );
  FLA_Check_error_code( e_val );
#endif

  r_val2 = read( fd, buffer, size );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_read_result( size, r_val2 );
  FLA_Check_error_code( e_val );
#endif

  time = FLA_Clock() - time;
  printf("size = %lu time = %le Read at %lf\n", size, time, size / time / 1000000.0 );

  return FLA_SUCCESS;
}

FLA_Error FLA_Write( size_t size, void *buffer, int fd, off_t offset )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;
#endif
  double    time, time_seek;
  off_t     r_val;
  ssize_t   r_val2;

  time = FLA_Clock();
  r_val = lseek( fd, offset, SEEK_SET );
  time_seek = FLA_Clock() - time;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_lseek_result( offset, r_val );
  FLA_Check_error_code( e_val );
#endif

  r_val2 = write( fd, buffer, size );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_write_result( size, r_val2 );
  FLA_Check_error_code( e_val );
#endif

  // fdatasync( fd );
  time = FLA_Clock()-time;
  printf("size = %lu time seek = %le time = %le Write at %lf\n", size, time_seek, time, size / time / 1000000.0 );

  return FLA_SUCCESS;
}

FLA_Error FLA_Read_matrix( int typesize, int length, int width, int fd,
                           int buffer_from, int ldim_from,
                           char *buffer_to, int ldim_to )
{
  int j;

  if ( ldim_from == length && ldim_to == length ) {
    FLA_Read( ( long ) ( length*width*typesize ), fd, buffer_from, buffer_to );
  }
  else {
    printf("read: warning\n");
    for ( j=0; j<width; j++ ) {
      FLA_Read( ( long ) ( length*typesize ), fd, buffer_from, buffer_to );
      buffer_from += ldim_from*typesize;
      buffer_to +=   ldim_to*typesize;
    } 
  } 
  return FLA_SUCCESS;
}

FLA_Error FLA_Write_matrix( int typesize, int length, int width, int fd,
                            char *buffer_from, int ldim_from,
                            int buffer_to, int ldim_to )
{
  int j;

  if ( ldim_from == length && ldim_to == length ) {
    FLA_Write( ( long ) ( length*width*typesize ), buffer_from, fd, buffer_to );
  }
  else {
    printf("write: warning\n");
    for ( j=0; j<width; j++ ){
      FLA_Write( ( long ) ( length*typesize ), buffer_from, fd, buffer_to );
      buffer_from += ldim_from*typesize;
      buffer_to +=   ldim_to*typesize;
    }
  } 
  return FLA_SUCCESS;
}

#endif
#endif
