/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"


FLA_Bool FLA_initialized = FALSE;

int
  FLA_i_minus_one,
  FLA_i_zero,
  FLA_i_one,
  FLA_i_two;
float
  FLA_s_minus_one,
  FLA_s_zero,
  FLA_s_one,
  FLA_s_two;
double
  FLA_d_minus_one,
  FLA_d_zero,
  FLA_d_one,
  FLA_d_two;
scomplex
  FLA_c_minus_one,
  FLA_c_zero,
  FLA_c_one,
  FLA_c_two;
dcomplex
  FLA_z_minus_one,
  FLA_z_zero,
  FLA_z_one,
  FLA_z_two;

FLA_Obj 
  FLA_MINUS_ONE,
  FLA_ZERO,
  FLA_ONE,
  FLA_TWO;

int* fla_lu_pivots_temp;
int  fla_lu_pivots_temp_length;

#ifdef FLA_ENABLE_MEM_COUNTER
extern int      memcounter;
#ifdef FLA_ENABLE_MULTITHREADING
extern FLA_Lock memcounter_lock;
#endif
#endif




/* *************************************************************************

   FLA_Init()

 *************************************************************************** */

void FLA_Init()
{
  if ( FLA_initialized ) return;
  
  FLA_initialized = TRUE;

  FLA_Error_messages_init();

  FLA_Init_memcounter();

  FLA_Init_constants();

  FLA_Init_temp();

  FLA_Cntl_init();

#ifdef FLA_ENABLE_NON_CRITICAL_CODE
  FLA_F2C( fla_init_fortran )( &FLA_MINUS_ONE, &FLA_ZERO, &FLA_ONE, &FLA_TWO );
#endif

#ifdef FLA_ENABLE_SUPERMATRIX
  FLASH_Queue_init( );
#endif
}

void FLA_F2C( fla_init_f )( void )
{
  FLA_Init();
}

/* *************************************************************************

   FLA_Init_constants()

 *************************************************************************** */

void FLA_Init_constants()
{
  int      *temp_i;
  float    *temp_s;
  double   *temp_d;
  scomplex *temp_c;
  dcomplex *temp_z;

  FLA_i_minus_one      = -1;
  FLA_i_zero           =  0;
  FLA_i_one            =  1;
  FLA_i_two            =  2;

  FLA_s_minus_one      = -1.0;
  FLA_s_zero           =  0.0;
  FLA_s_one            =  1.0;
  FLA_s_two            =  2.0;

  FLA_d_minus_one      = -1.0;
  FLA_d_zero           =  0.0;
  FLA_d_one            =  1.0;
  FLA_d_two            =  2.0;

  FLA_c_minus_one.real = -1.0;
  FLA_c_minus_one.imag =  0.0;
  FLA_c_zero.real      =  0.0;
  FLA_c_zero.imag      =  0.0;
  FLA_c_one.real       =  1.0;
  FLA_c_one.imag       =  0.0;
  FLA_c_two.real       =  2.0;
  FLA_c_two.imag       =  0.0;

  FLA_z_minus_one.real = -1.0;
  FLA_z_minus_one.imag =  0.0;
  FLA_z_zero.real      =  0.0;
  FLA_z_zero.imag      =  0.0;
  FLA_z_one.real       =  1.0;
  FLA_z_one.imag       =  0.0;
  FLA_z_two.real       =  2.0;
  FLA_z_two.imag       =  0.0;

  FLA_Obj_create( FLA_CONSTANT, 1, 1, &FLA_MINUS_ONE );

  temp_i = FLA_INT_PTR( FLA_MINUS_ONE );
  temp_s = FLA_FLOAT_PTR( FLA_MINUS_ONE );
  temp_d = FLA_DOUBLE_PTR( FLA_MINUS_ONE );
  temp_c = FLA_COMPLEX_PTR( FLA_MINUS_ONE );
  temp_z = FLA_DOUBLE_COMPLEX_PTR( FLA_MINUS_ONE );

  *temp_i = FLA_i_minus_one;
  *temp_s = FLA_s_minus_one;
  *temp_d = FLA_d_minus_one;
  *temp_c = FLA_c_minus_one;
  *temp_z = FLA_z_minus_one;

  FLA_Obj_create( FLA_CONSTANT, 1, 1, &FLA_ZERO );

  temp_i = FLA_INT_PTR( FLA_ZERO );
  temp_s = FLA_FLOAT_PTR( FLA_ZERO );
  temp_d = FLA_DOUBLE_PTR( FLA_ZERO );
  temp_c = FLA_COMPLEX_PTR( FLA_ZERO );
  temp_z = FLA_DOUBLE_COMPLEX_PTR( FLA_ZERO );

  *temp_i = FLA_i_zero;
  *temp_s = FLA_s_zero;
  *temp_d = FLA_d_zero;
  *temp_c = FLA_c_zero;
  *temp_z = FLA_z_zero;

  FLA_Obj_create( FLA_CONSTANT, 1, 1, &FLA_ONE );

  temp_i = FLA_INT_PTR( FLA_ONE );
  temp_s = FLA_FLOAT_PTR( FLA_ONE );
  temp_d = FLA_DOUBLE_PTR( FLA_ONE );
  temp_c = FLA_COMPLEX_PTR( FLA_ONE );
  temp_z = FLA_DOUBLE_COMPLEX_PTR( FLA_ONE );

  *temp_i = FLA_i_one;
  *temp_s = FLA_s_one;
  *temp_d = FLA_d_one;
  *temp_c = FLA_c_one;
  *temp_z = FLA_z_one;

  FLA_Obj_create( FLA_CONSTANT, 1, 1, &FLA_TWO );

  temp_i = FLA_INT_PTR( FLA_TWO );
  temp_s = FLA_FLOAT_PTR( FLA_TWO );
  temp_d = FLA_DOUBLE_PTR( FLA_TWO );
  temp_c = FLA_COMPLEX_PTR( FLA_TWO );
  temp_z = FLA_DOUBLE_COMPLEX_PTR( FLA_TWO );

  *temp_i = FLA_i_two;
  *temp_s = FLA_s_two;
  *temp_d = FLA_d_two;
  *temp_c = FLA_c_two;
  *temp_z = FLA_z_two;
}

/* *************************************************************************

   FLA_Finalize_constants()

 *************************************************************************** */

void FLA_Finalize_constants()
{
  FLA_Obj_free( &FLA_MINUS_ONE );
  FLA_Obj_free( &FLA_ZERO );
  FLA_Obj_free( &FLA_ONE );
  FLA_Obj_free( &FLA_TWO );
}

/* *************************************************************************

   FLA_Init_memcounter()

 *************************************************************************** */

void FLA_Init_memcounter()
{
#ifdef FLA_ENABLE_MEM_COUNTER
  memcounter = 0;
#ifdef FLA_ENABLE_MULTITHREADING
  FLA_Lock_init( &memcounter_lock );
#endif
#endif
}

/* *************************************************************************

   FLA_Finalize_memcounter()

 *************************************************************************** */

void FLA_Finalize_memcounter()
{
#ifdef FLA_ENABLE_MEM_COUNTER
#ifndef FLA_ENABLE_PLAPACK2E 
  fprintf( stderr, "libflame: memory leak counter: %d\n", memcounter );
  fflush( stderr );
#ifdef FLA_ENABLE_MULTITHREADING
  FLA_Lock_init( &memcounter_lock );
#endif
#endif
#endif
}

/* *************************************************************************

   FLA_Init_temp()

 *************************************************************************** */

void FLA_Init_temp()
{
  fla_lu_pivots_temp_length = FLA_MAX_LU_PIVOT_LENGTH;
  fla_lu_pivots_temp        = ( int * ) FLA_malloc( ( size_t ) fla_lu_pivots_temp_length );
}

/* *************************************************************************

   FLA_Finalize_temp()

 *************************************************************************** */
void FLA_Finalize_temp()
{
  FLA_free( fla_lu_pivots_temp );
}

/* *************************************************************************

  FLA_Finalize()

 *************************************************************************** */

void FLA_Finalize()
{
  if ( FLA_initialized == FALSE )
    return;

  FLA_initialized = FALSE;

  FLA_Finalize_constants();

  FLA_Finalize_temp();
  
  FLA_Cntl_finalize();

#ifdef FLA_ENABLE_NON_CRITICAL_CODE
  FLA_F2C( fla_finalize_fortran )();
#endif
  
#ifdef FLA_ENABLE_SUPERMATRIX
  FLASH_Queue_finalize();
#endif

  FLA_Finalize_memcounter();
}

void FLA_F2C( fla_finalize_f )( void )
{
  FLA_Finalize();
}

/* *************************************************************************

   FLA_Initialized()

 *************************************************************************** */

FLA_Bool FLA_Initialized( void )
{
  return FLA_initialized;
}

F_INT FLA_F2C( fla_initialized_f )( void )
{
  return ( F_INT ) FLA_initialized;
}

