/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"


#ifdef FLA_ENABLE_MEM_COUNTER
int memcounter = 0;
#ifdef FLA_ENABLE_MULTITHREADING
FLA_Lock memcounter_lock;
#endif
#endif


/* ***************************************************************************

   FLA_malloc()

 *************************************************************************** */

void* FLA_malloc( size_t size )
{
  void*     ptr = NULL;
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;
#endif
#ifdef FLA_ENABLE_MEMORY_ALIGNMENT
  int       r_val;
#endif

  // In practice, the size argument should very rarely be zero. However, if the
  // calling code does request a memory region of zero length, we short-circut
  // the actual allocation request and just return NULL. Hopefully, the calling
  // code is written such that the pointer is never dereferenced. At free()-time
  // everything will be fine, as calling free() with a NULl pointer is safe.
  // Also note that we do NOT increment the memory counter before returning.
  // Likewise, we will not decrement the counter when a NULL pointer is freed.
  if ( size == 0 ) return NULL;

#ifdef FLA_ENABLE_MEMORY_ALIGNMENT

  // Allocate size bytes of memory. Here, we call posix_memalign() if
  // memory alignment was requested at configure-time, providing the
  // alignment boundary value given by the user. posix_memalign() also
  // returns an error code, which is how it signals that something
  // went wrong. Compare to malloc(), which does this by simply returning
  // a NULL pointer.
  r_val = posix_memalign( &ptr, ( size_t ) FLA_MEMORY_ALIGNMENT_BOUNDARY, size );

  // Check the return value of posix_memalign() for evidence that the
  // request failed.
  #ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_posix_memalign_failure( r_val );
  FLA_Check_error_code( e_val );
  #endif

#else

  // Allocate size bytes of memory. Note that malloc() only guarantees 8-byte
  // alignment.
  ptr = malloc( size );

  // It may not seem useful to have a check for a null pointer here, given
  // that such an occurance would cause the file and line of the error to
  // be reported as the below line of the current file instead of the file
  // and line number of the calling code. However, consider that in the
  // unlikely event that malloc() does return a null pointer, the user will
  // have much bigger problems on his hands (e.g. an exhausted memory heap)
  // than needing to know exactly what line in the library triggered error.
  // Note that such a line in the application code is likely not the root
  // source of the problem anyway (ie: not the reason why the heap is full).
  #ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  e_val = FLA_Check_malloc_pointer( ptr );
  FLA_Check_error_code( e_val );
  #endif

#endif

  // Update the memory counter if it is enabled, and do so thread-safely if
  // multithreading is enabled.
#ifdef FLA_ENABLE_MEM_COUNTER
  #ifdef FLA_ENABLE_MULTITHREADING
  FLA_Lock_acquire( &memcounter_lock );
  memcounter += 1;
  FLA_Lock_release( &memcounter_lock );
  #else
  memcounter += 1;
  #endif
#endif
  
  // Return the pointer to the new memory region returned by malloc().
  return ptr;
}

/* ***************************************************************************

   FLA_realloc()

 *************************************************************************** */

void* FLA_realloc( void* old_ptr, size_t size )
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Error e_val;
#endif
  void*     new_ptr;

  // We can't do much if size is zero. To emulate realloc(), we must
  // return a NULL pointer, regardless of the value of old_ptr.
  if ( size == 0 )
  {
    // If the pointer is valid, free() it.
    if ( old_ptr != NULL )
      FLA_free( old_ptr );

    // If size is zero, we should return a NULL pointer.
    new_ptr = NULL;
  }
  else
  {
    // If old_ptr is NULL, allocate size bytes as if it were a first-time
    // FLA_malloc() request. Otherwise, proceed to realloc() the memory.
    if ( old_ptr == NULL )
      new_ptr = FLA_malloc( size );
    else
    {
      // At this point, we know that size is non-zero and old_ptr is valid.

      // Since we may need aligned addresses, we don't really want to call
      // realloc(), since it does not guarantee arbitrary aligned pointers.
      // But we can't implement it ourselves either, because we don't know
      // how large the original buffer is, therefor we don't know how much
      // to copy over after the new buffer is allocated. So we're stuck with
      // the system implementation.
      new_ptr = realloc( old_ptr, size );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
      e_val = FLA_Check_malloc_pointer( new_ptr );
      FLA_Check_error_code( e_val );
#endif
    }
  }

  // Return the pointer (either NULL, or the return value from FLA_malloc()
  // or realloc()).
  return new_ptr;
}

/* ***************************************************************************

   FLA_free()

 *************************************************************************** */

void FLA_free( void* ptr )
{
  // Free the memory addressed by ptr. It doesn't matter if ptr is NULL;
  // free() will do the right thing (in this case, nothing).
  free( ptr );

#ifdef FLA_ENABLE_MEM_COUNTER
  // We don't want to decrement the counter if the buffer is NULL.
  // This is because it's likely that the buffer was never allocated
  // a valid pointer to begin with, which means FLA_malloc() was never
  // called and thus the counter was never incremented.
  if ( ptr != NULL )
  {
  #ifdef FLA_ENABLE_MULTITHREADING
    FLA_Lock_acquire( &memcounter_lock );
    memcounter -= 1;
    FLA_Lock_release( &memcounter_lock );
  #else
    memcounter -= 1;
  #endif
  }
#endif
}

