/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"


#if   FLA_MULTITHREADING_MODEL == FLA_OPENMP
#include <omp.h>
#elif FLA_MULTITHREADING_MODEL == FLA_PTHREADS
#include <pthread.h>
#endif


#ifdef FLA_ENABLE_SUPERMATRIX

FLASH_Queue           _tq;

static FLA_Bool       flash_queue_initialized    = FALSE;

static int            flash_queue_n_read_blocks  = 0;
static int            flash_queue_n_write_blocks = 0;

static FLA_Bool       flash_queue_verbose        = FALSE;
static FLA_Bool       flash_queue_sorting        = FALSE;
static FLA_Bool       flash_queue_caching        = FALSE;
static FLASH_Data_aff flash_queue_data_affinity  = FLASH_QUEUE_AFFINITY_NONE;

static double         flash_queue_total_time     = 0.0;
static double         flash_queue_parallel_time  = 0.0;

static int            flash_queue_block_size     = 0;
static int            flash_queue_cache_size     = 4 * 1024 * 1024;

#endif


static int            flash_queue_stack          = 0;
static FLA_Bool       flash_queue_enabled        = TRUE;

static unsigned int   flash_queue_n_threads      = 1;


void FLASH_Queue_begin( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_begin

----------------------------------------------------------------------------*/
{
#ifdef FLA_ENABLE_SUPERMATRIX
   if ( flash_queue_stack == 0 )
   {
      // Reset the value of the parallel execution timer.
      flash_queue_parallel_time = 0.0;

      // Save the starting time for the total execution time.
      flash_queue_total_time = FLA_Clock();
   }
#endif

   // Push onto the stack.
   flash_queue_stack++;

   return;
}


void FLASH_Queue_end( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_end

----------------------------------------------------------------------------*/
{
   // Pop off the stack.
   flash_queue_stack--;

#ifdef FLA_ENABLE_SUPERMATRIX
   if ( flash_queue_stack == 0 )
   {
      // Execute tasks if encounter the outermost parallel region.
      FLASH_Queue_exec();

      // Find the total execution time.
      flash_queue_total_time = FLA_Clock() - flash_queue_total_time;
   }
#endif

   return;
}


FLA_Error FLASH_Queue_enable( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_enable

----------------------------------------------------------------------------*/
{
#ifdef FLA_ENABLE_SUPERMATRIX
   if ( flash_queue_stack == 0 )
   {
      // Enable if not begin parallel region yet.
      flash_queue_enabled = TRUE;   
      return FLA_SUCCESS;
   }
   else
   {
      // Cannot change status during parallel region.
      return FLA_FAILURE;
   }
#else
   // Raise an exception when SuperMatrix is not configured.
   FLA_Check_error_code( FLA_SUPERMATRIX_NOT_ENABLED );
   return FLA_FAILURE;
#endif
}


FLA_Error FLASH_Queue_disable( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_disable

----------------------------------------------------------------------------*/
{
#ifdef FLA_ENABLE_SUPERMATRIX
   if ( flash_queue_stack == 0 )
   {
      // Disable if not begin parallel region yet.
      flash_queue_enabled = FALSE;   
      return FLA_SUCCESS;      
   }
   else
   {
      // Cannot change status during parallel region.
      return FLA_FAILURE;
   }
#else
   // Allow disabling enqueuing even when SuperMatrix is not configured.
   flash_queue_enabled = FALSE;   
   return FLA_SUCCESS;
#endif
}


FLA_Bool FLASH_Queue_get_enabled( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_enabled

----------------------------------------------------------------------------*/
{
   // Return if enabled, but always false if SuperMatrix is not configured.
#ifdef FLA_ENABLE_SUPERMATRIX
   return flash_queue_enabled;
#else
   return FALSE;
#endif
}


void FLASH_Queue_set_num_threads( unsigned int n_threads )
/*----------------------------------------------------------------------------

   FLASH_Queue_set_num_threads

----------------------------------------------------------------------------*/
{
   FLA_Error e_val;

   // Verify that the number of threads is positive. 
   e_val = FLA_Check_num_threads( n_threads );
   FLA_Check_error_code( e_val );

   // Keep track of the number of threads internally.
   flash_queue_n_threads = n_threads;

#if   FLA_MULTITHREADING_MODEL == FLA_OPENMP

   // No additional action is necessary to set the number of OpenMP threads
   // since setting the number of threads is handled at the parallel for loop
   // with a num_threads() clause. This gives the user more flexibility since
   // he can use the OMP_NUM_THREADS environment variable or the
   // omp_set_num_threads() function to set the global number of OpenMP threads
   // independently of the number of SuperMatrix threads.
   
#elif FLA_MULTITHREADING_MODEL == FLA_PTHREADS

   // No additional action is necessary to set the number of pthreads
   // since setting the number of threads is handled entirely on our end.

#endif

   return;
}


unsigned int FLASH_Queue_get_num_threads( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_num_threads

----------------------------------------------------------------------------*/
{
   return flash_queue_n_threads;
}


#ifdef FLA_ENABLE_SUPERMATRIX


void FLASH_Queue_init( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_init

----------------------------------------------------------------------------*/
{
   // Exit early if we're already initialized.
   if ( flash_queue_initialized == TRUE )
      return;
   
   // Reset all the initial values.
   FLASH_Queue_reset();

   // Set the initialized flag.
   flash_queue_initialized = TRUE;

   return;
}


void FLASH_Queue_finalize( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_finalize

----------------------------------------------------------------------------*/
{
   // Exit early if we're not already initialized.
   if ( flash_queue_initialized == FALSE )
      return;

   // Clear the initialized flag.
   flash_queue_initialized = FALSE;

   return;
}


int FLASH_Queue_get_num_tasks( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_num_tasks

----------------------------------------------------------------------------*/
{
   return _tq.n_tasks;
}


void FLASH_Queue_set_verbose_output( FLA_Bool verbose )
/*----------------------------------------------------------------------------

   FLASH_Queue_set_verbose_output

----------------------------------------------------------------------------*/
{ 
   flash_queue_verbose = verbose;

   return;
}


FLA_Bool FLASH_Queue_get_verbose_output( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_verbose_output

----------------------------------------------------------------------------*/
{ 
   return flash_queue_verbose;
}


void FLASH_Queue_set_sorting( FLA_Bool sorting )
/*----------------------------------------------------------------------------

   FLASH_Queue_set_sorting

----------------------------------------------------------------------------*/
{ 
   flash_queue_sorting = sorting; 

   return;
}


FLA_Bool FLASH_Queue_get_sorting( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_sorting

----------------------------------------------------------------------------*/
{ 
   return flash_queue_sorting;
}


void FLASH_Queue_set_caching( FLA_Bool caching )
/*----------------------------------------------------------------------------

   FLASH_Queue_set_caching

----------------------------------------------------------------------------*/
{ 
   flash_queue_caching = caching; 

   return;
}


FLA_Bool FLASH_Queue_get_caching( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_caching

----------------------------------------------------------------------------*/
{ 
   return flash_queue_caching;
}


void FLASH_Queue_set_data_affinity( FLASH_Data_aff data_affinity )
/*----------------------------------------------------------------------------

   FLASH_Queue_set_data_affinity

----------------------------------------------------------------------------*/
{ 
   flash_queue_data_affinity = data_affinity; 

   return;
}


FLASH_Data_aff FLASH_Queue_get_data_affinity( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_data_affinity

----------------------------------------------------------------------------*/
{ 
   return flash_queue_data_affinity;
}


double FLASH_Queue_get_total_time( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_total_time

----------------------------------------------------------------------------*/
{
   // Only return time if out of parallel region.
   if ( flash_queue_stack == 0 )
      return flash_queue_total_time;

   return 0.0;
}


double FLASH_Queue_get_parallel_time( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_parallel_time

----------------------------------------------------------------------------*/
{
   // Only return time if out of parallel region.
   if ( flash_queue_stack == 0 )
      return flash_queue_parallel_time;

   return 0.0;
}


// --- helper functions --- ===================================================


void FLASH_Queue_set_parallel_time( double dtime )
/*----------------------------------------------------------------------------

   FLASH_Queue_set_parallel_time

----------------------------------------------------------------------------*/
{
   flash_queue_parallel_time += dtime;

   return;
}


void FLASH_Queue_set_block_size( int size )
/*----------------------------------------------------------------------------

   FLASH_Queue_set_block_size

----------------------------------------------------------------------------*/
{
   // Only adjust the block size if the new block is larger.
   if ( flash_queue_block_size < size )
      flash_queue_block_size = size;

   return;
}


int FLASH_Queue_get_block_size( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_block_size

----------------------------------------------------------------------------*/
{
   return flash_queue_block_size;
}


void FLASH_Queue_set_cache_size( int size )
/*----------------------------------------------------------------------------

   FLASH_Queue_set_cache_size

----------------------------------------------------------------------------*/
{
   flash_queue_cache_size = size;

   return;
}


int FLASH_Queue_get_cache_size( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_cache_size

----------------------------------------------------------------------------*/
{
   return flash_queue_cache_size;
}


void FLASH_Queue_reset( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_reset

----------------------------------------------------------------------------*/
{
   // Clear the other fields of the FLASH_Queue structure.
   _tq.n_tasks = 0;
   _tq.head    = NULL;
   _tq.tail    = NULL;

   // Reset the number of blocks.
   flash_queue_n_read_blocks  = 0;
   flash_queue_n_write_blocks = 0;

   return;
}


FLASH_Task* FLASH_Queue_get_head_task( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_head_task

----------------------------------------------------------------------------*/
{
   return _tq.head;
}


FLASH_Task* FLASH_Queue_get_tail_task( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_get_tail_task

----------------------------------------------------------------------------*/
{
   return _tq.tail;
}


void FLASH_Queue_push( void* func, 
                       void* cntl,
                       char* name,
                       double cost,
                       int n_int_args, 
                       int n_fla_args,
                       int n_input_args,
                       int n_output_args,
                       ... )
/*----------------------------------------------------------------------------

   FLASH_Queue_push

----------------------------------------------------------------------------*/
{
   int         i, j;
   va_list     var_arg_list;
   FLASH_Task* t;
   FLASH_Task* task;
   FLASH_Dep*  d;
   FLASH_Dep*  next_dep;
   FLA_Obj     obj;

   // Allocate a new FLA_Task and populate its fields with appropriate values.
   t = FLASH_Task_alloc( func, cntl, name, cost,
                         n_int_args, n_fla_args, 
                         n_input_args, n_output_args );
   
   // Initialize variable argument environment. In case you're wondering, the
   // second argument in this macro invocation of va_start() is supposed to be
   // the parameter that immediately preceeds the variable argument list
   // (ie: the ... above ).
   va_start( var_arg_list, n_output_args );

   // Extract the integer arguments.
   for ( i = 0; i < n_int_args; i++ )
      t->int_arg[i] = va_arg( var_arg_list, int );
   
   // Extract the FLA_Obj arguments.
   for ( i = 0; i < n_fla_args; i++ )
      t->fla_arg[i] = va_arg( var_arg_list, FLA_Obj );

   // Extract the input FLA_Obj arguments.
   for ( i = 0; i < n_input_args; i++ )
   {
      obj = va_arg( var_arg_list, FLA_Obj );
      t->input_arg[i] = obj;

      // Find dependence information.
      if ( obj.base->enqueue->write_task == NULL )
      {
         t->n_ready--;

         // Add to number of blocks read if not written and not read before.
         if ( obj.base->enqueue->n_read_tasks == 0 )
         {
            // Identify each read block with an id for freeing.
            obj.base->n_read_blocks = flash_queue_n_read_blocks;

            flash_queue_n_read_blocks++;            
         }
      }
      else
      { // Flow dependence.
         task = obj.base->enqueue->write_task;

         d = (FLASH_Dep *) FLA_malloc( sizeof(FLASH_Dep) );

         d->deptype  = 0; // Flow
         d->task     = t;
         d->next_dep = NULL;
         
         if ( task->n_dep_args == 0 )
         {
            task->dep_arg_head = d;
            task->dep_arg_tail = d;
         }
         else
         {
            task->dep_arg_tail->next_dep = d;
            task->dep_arg_tail           = d;
         }

         task->n_dep_args++;
      }

      // Add task to the read task in the object if not already there.
      if ( obj.base->enqueue->n_read_tasks == 0 ||
           obj.base->enqueue->read_task_tail->task != t )
      { // Anti-dependence potentially.
         d = (FLASH_Dep *) FLA_malloc( sizeof(FLASH_Dep) );

         d->deptype  = 1; // Anti
         d->task     = t;
         d->next_dep = NULL;

         if ( obj.base->enqueue->n_read_tasks == 0 )
         {
            obj.base->enqueue->read_task_head = d;
            obj.base->enqueue->read_task_tail = d;
         }
         else
         {
            obj.base->enqueue->read_task_tail->next_dep = d;
            obj.base->enqueue->read_task_tail           = d;
         }
         
         obj.base->enqueue->n_read_tasks++;
      }      
   }

   // Extract the output FLA_Obj arguments.
   for ( i = 0; i < n_output_args; i++ )
   {
      obj = va_arg( var_arg_list, FLA_Obj );
      t->output_arg[i] = obj;

      // Assign tasks to threads with data affinity.
      if ( obj.base->enqueue->write_task == NULL )
      {
         t->n_ready--;

         // Only assign data affinity to the first output block.
         if ( i == 0 )
            t->queue = flash_queue_n_write_blocks;

         // Number of blocks written if not written before.
         flash_queue_n_write_blocks++;

         // Add to number of blocks read if not written or read before.
         if ( obj.base->enqueue->n_read_tasks == 0 )
         {
            // Identify each read block with an id for freeing.
            obj.base->n_read_blocks = flash_queue_n_read_blocks;

            flash_queue_n_read_blocks++;
         }
      }
      else
      { // Flow dependence
         // Create dependency from task that last wrote the block.
         task = obj.base->enqueue->write_task;
         
         d = (FLASH_Dep *) FLA_malloc( sizeof(FLASH_Dep) );

         d->deptype  = 0; // Flow; 2 (possibly output dependence)
         d->task     = t;
         d->next_dep = NULL;
         
         if ( task->n_dep_args == 0 )
         {
            task->dep_arg_head = d;
            task->dep_arg_tail = d;
         }
         else
         {
            task->dep_arg_tail->next_dep = d;
            task->dep_arg_tail           = d;
         }
         
         task->n_dep_args++;

         // Only assign data affinity to the first output block.
         if ( i == 0 )
            t->queue = task->queue;
      }
      
      // Clear read task for next set of reads and record the anti-dependence.
      d = obj.base->enqueue->read_task_head;

      for ( j = 0; j < obj.base->enqueue->n_read_tasks; j++ )
      {
         task     = d->task;
         next_dep = d->next_dep;

         // If the last task to read is not the current task, add dependence.
         if ( task != t )
         {
            d->deptype  = 1; // Anti
            d->task     = t;
            d->next_dep = NULL;
            
            if ( task->n_dep_args == 0 )
            {
               task->dep_arg_head = d;
               task->dep_arg_tail = d;
            }
            else
            {
               task->dep_arg_tail->next_dep = d;
               task->dep_arg_tail           = d;
            }
            
            task->n_dep_args++;

            t->n_war_args++;
         }  
         else
         {
            FLA_free( d );
         }

         d = next_dep;
      }
      
      obj.base->enqueue->n_read_tasks   = 0;
      obj.base->enqueue->read_task_head = NULL;
      obj.base->enqueue->read_task_tail = NULL;

      // Record this task as the last to write to this block.
      obj.base->enqueue->write_task = t;
   }      

   // Finalize the variable argument environment.
   va_end( var_arg_list );
  
   // Add the task to the tail of the queue (and the head if queue is empty).
   if ( _tq.n_tasks == 0 )
   {
      _tq.head = t;
      _tq.tail = t;
   }
   else
   {
      t->prev_task = _tq.tail;
      _tq.tail->next_task = t;
      _tq.tail            = t;
   }
   
   // Increment the number of tasks.
   _tq.n_tasks++;

   return;
}


FLASH_Task* FLASH_Task_alloc( void *func, 
                              void *cntl,
                              char *name,
                              double cost,
                              int n_int_args, 
                              int n_fla_args,
                              int n_input_args,
                              int n_output_args )
/*----------------------------------------------------------------------------

   FLASH_Task_alloc

----------------------------------------------------------------------------*/
{
   FLASH_Task* t;

   // Allocate space for the task structure t.
   t             = (FLASH_Task *) FLA_malloc( sizeof(FLASH_Task) );

   // Allocate space for the task's integer arguments.
   t->int_arg    = (int *) FLA_malloc( n_int_args * sizeof(int) );

   // Allocate space for the task's FLA_Obj arguments.
   t->fla_arg    = (FLA_Obj *) FLA_malloc( n_fla_args * sizeof(FLA_Obj) );

   // Allocate space for the task's input FLA_Obj arguments.
   t->input_arg  = (FLA_Obj *) FLA_malloc( n_input_args * sizeof(FLA_Obj) );

   // Allocate space for the task's output FLA_Obj arguments.
   t->output_arg = (FLA_Obj *) FLA_malloc( n_output_args * sizeof(FLA_Obj) );
   
   // Initialize other fields of the structure.
   t->n_ready       = 0;
   t->order         = 0;
   t->queue         = 0;
   t->thread        = 0;
   t->sort          = 0;
   t->weighted_sort = 0.0;

   t->func          = func;
   t->cntl          = cntl;
   t->name          = name;
   t->cost          = cost;
   t->n_int_args    = n_int_args;
   t->n_fla_args    = n_fla_args;
   t->n_input_args  = n_input_args;
   t->n_output_args = n_output_args;
   
   t->n_war_args    = 0;
   t->n_dep_args    = 0;
   t->dep_arg_head  = NULL;
   t->dep_arg_tail  = NULL;
   t->prev_task     = NULL;
   t->next_task     = NULL;
   t->prev_wait     = NULL;
   t->next_wait     = NULL;
   
   // Return a pointer to the initialized structure.
   return t;
}


void FLASH_Task_free( FLASH_Task *t )
/*----------------------------------------------------------------------------

   FLASH_Task_free

----------------------------------------------------------------------------*/
{
   int        i, j, k;
   FLASH_Dep* d;
   FLASH_Dep* next_dep;
   
   // Clearing the last write task in each output block.
   for ( i = 0; i < t->n_output_args; i++ )
      t->output_arg[i].base->enqueue->write_task = NULL;

   // Cleaning the last read tasks in each input block.
   for ( i = 0; i < t->n_input_args; i++ )
   {
      k = t->input_arg[i].base->enqueue->n_read_tasks;
      d = t->input_arg[i].base->enqueue->read_task_head;

      t->input_arg[i].base->enqueue->n_read_tasks   = 0;
      t->input_arg[i].base->enqueue->read_task_head = NULL;
      t->input_arg[i].base->enqueue->read_task_tail = NULL;

      for ( j = 0; j < k; j++ )
      {
         next_dep = d->next_dep;
         FLA_free( d );
         d = next_dep;
      }
   }

   // Free the dep_arg field of t.
   d = t->dep_arg_head;

   for ( i = 0; i < t->n_dep_args; i++ )
   {
      next_dep = d->next_dep;
      FLA_free( d );
      d = next_dep;
   }   

   // Free the int_arg field of t.
   FLA_free( t->int_arg );
   
   // Free the fla_arg field of t.
   FLA_free( t->fla_arg );

   // Free the input_arg field of t.
   FLA_free( t->input_arg );

   // Free the output_arg field of t.
   FLA_free( t->output_arg );

   // Finally, free the struct itself.
   FLA_free( t );

   return;
}


void FLASH_Queue_exec_task( FLASH_Task* t )
/*----------------------------------------------------------------------------

   FLASH_Queue_exec_task

----------------------------------------------------------------------------*/
{
   // Only execute task if it is not NULL.
   if ( t == NULL )
      return;
   
#ifdef FLA_ENABLE_SUPERMATRIX_VISUALIZATION
   t->begin_time = FLA_Clock();
#endif

   // Define local function pointer types.
   typedef FLA_Error(*flash_lu_nopiv_p)(FLA_Obj A, fla_lu_t* cntl);
   typedef FLA_Error(*flash_trinv_p)(FLA_Uplo uplo, FLA_Diag diag, FLA_Obj A, fla_trinv_t* cntl);
   typedef FLA_Error(*flash_ttmm_p)(FLA_Uplo uplo, FLA_Obj A, fla_ttmm_t* cntl);
   typedef FLA_Error(*flash_chol_p)(FLA_Uplo uplo, FLA_Obj A, fla_chol_t* cntl);
   typedef FLA_Error(*flash_sylv_p)(FLA_Trans transa, FLA_Trans transb, FLA_Obj isgn, FLA_Obj A, FLA_Obj B, FLA_Obj C, FLA_Obj scale, fla_sylv_t* cntl);
   typedef FLA_Error(*flash_gemm_p)(FLA_Trans transa, FLA_Trans transb, FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C, fla_gemm_t* cntl);
   typedef FLA_Error(*flash_hemm_p)(FLA_Side side, FLA_Uplo uplo, FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C, fla_hemm_t* cntl);
   typedef FLA_Error(*flash_herk_p)(FLA_Uplo uplo, FLA_Trans transa, FLA_Obj alpha, FLA_Obj A, FLA_Obj beta, FLA_Obj C, fla_herk_t* cntl);
   typedef FLA_Error(*flash_her2k_p)(FLA_Uplo uplo, FLA_Trans transa, FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C, fla_her2k_t* cntl);
   typedef FLA_Error(*flash_symm_p)(FLA_Side side, FLA_Uplo uplo, FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C, fla_symm_t* cntl);
   typedef FLA_Error(*flash_syrk_p)(FLA_Uplo uplo, FLA_Trans transa, FLA_Obj alpha, FLA_Obj A, FLA_Obj beta, FLA_Obj C, fla_syrk_t* cntl);
   typedef FLA_Error(*flash_syr2k_p)(FLA_Uplo uplo, FLA_Trans transa, FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C, fla_syr2k_t* cntl);
   typedef FLA_Error(*flash_trmm_p)(FLA_Side side, FLA_Uplo uplo, FLA_Trans trans, FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj C, fla_trmm_t* cntl);
   typedef FLA_Error(*flash_trsm_p)(FLA_Side side, FLA_Uplo uplo, FLA_Trans trans, FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj C, fla_trsm_t* cntl);

   // Now "switch" between the various possible task functions.
   
   // FLA_LU_nopiv
   if ( t->func == (void *) FLA_LU_nopiv_task )
   {
      flash_lu_nopiv_p func;
      func = (flash_lu_nopiv_p) t->func;

      func(               t->output_arg[0],
            ( fla_lu_t* ) t->cntl );
   }
   // FLA_Trinv
   else if ( t->func == (void *) FLA_Trinv_task )
   {
      flash_trinv_p func;
      func = (flash_trinv_p) t->func;

      func( ( FLA_Uplo     ) t->int_arg[0],
            ( FLA_Diag     ) t->int_arg[1],
                             t->output_arg[0],
            ( fla_trinv_t* ) t->cntl );
   }
   // FLA_Ttmm
   else if ( t->func == (void *) FLA_Ttmm_task )
   {      
      flash_ttmm_p func;
      func = (flash_ttmm_p) t->func;

      func( ( FLA_Uplo    ) t->int_arg[0],
                            t->output_arg[0],
            ( fla_ttmm_t* ) t->cntl );
   }
   // FLA_Chol
   else if ( t->func == (void *) FLA_Chol_task )
   {      
      flash_chol_p func;
      func = (flash_chol_p) t->func;

      func( ( FLA_Uplo    ) t->int_arg[0],
                            t->output_arg[0],
            ( fla_chol_t* ) t->cntl );
   }
   // FLA_Sylv
   else if ( t->func == (void *) FLA_Sylv_task )
   {      
      flash_sylv_p func;
      func = (flash_sylv_p) t->func;

      func( ( FLA_Trans   ) t->int_arg[0],
            ( FLA_Trans   ) t->int_arg[1],
                            t->fla_arg[0],
                            t->input_arg[0],
                            t->input_arg[1],
                            t->output_arg[0],
                            t->fla_arg[1],
            ( fla_sylv_t* ) t->cntl );
   }
   // FLA_Gemm
   else if ( t->func == (void *) FLA_Gemm_task )
   {
      flash_gemm_p func;
      func = (flash_gemm_p) t->func;

      func( ( FLA_Trans   ) t->int_arg[0],
            ( FLA_Trans   ) t->int_arg[1],
                            t->fla_arg[0],
                            t->input_arg[0],
                            t->input_arg[1],
                            t->fla_arg[1],
                            t->output_arg[0],
            ( fla_gemm_t* ) t->cntl );
   }
   // FLA_Hemm
   else if ( t->func == (void *) FLA_Hemm_task )
   {
      flash_hemm_p func;
      func = (flash_hemm_p) t->func;
      
      func( ( FLA_Side    ) t->int_arg[0],
            ( FLA_Uplo    ) t->int_arg[1],
                            t->fla_arg[0],
                            t->input_arg[0],
                            t->input_arg[1],
                            t->fla_arg[1],
                            t->output_arg[0],
            ( fla_hemm_t* ) t->cntl );
   }
   // FLA_Herk
   else if ( t->func == (void *) FLA_Herk_task )
   {
      flash_herk_p func;
      func = (flash_herk_p) t->func;

      func( ( FLA_Uplo    ) t->int_arg[0],
            ( FLA_Trans   ) t->int_arg[1],
                            t->fla_arg[0],
                            t->input_arg[0],
                            t->fla_arg[1],
                            t->output_arg[0],
            ( fla_herk_t* ) t->cntl );
   }
   // FLA_Her2k
   else if ( t->func == (void *) FLA_Her2k_task )
   {
      flash_her2k_p func;
      func = (flash_her2k_p) t->func;

      func( ( FLA_Uplo     ) t->int_arg[0],
            ( FLA_Trans    ) t->int_arg[1],
                             t->fla_arg[0],
                             t->input_arg[0],
                             t->input_arg[1],
                             t->fla_arg[1],
                             t->output_arg[0],
            ( fla_her2k_t* ) t->cntl );
   }
   // FLA_Symm
   else if ( t->func == (void *) FLA_Symm_task )
   {
      flash_symm_p func;
      func = (flash_symm_p) t->func;
      
      func( ( FLA_Side    ) t->int_arg[0],
            ( FLA_Uplo    ) t->int_arg[1],
                            t->fla_arg[0],
                            t->input_arg[0],
                            t->input_arg[1],
                            t->fla_arg[1],
                            t->output_arg[0],
            ( fla_symm_t* ) t->cntl );
   }
   // FLA_Syrk
   else if ( t->func == (void *) FLA_Syrk_task )
   {
      flash_syrk_p func;
      func = (flash_syrk_p) t->func;

      func( ( FLA_Uplo    ) t->int_arg[0],
            ( FLA_Trans   ) t->int_arg[1],
                            t->fla_arg[0],
                            t->input_arg[0],
                            t->fla_arg[1],
                            t->output_arg[0],
            ( fla_syrk_t* ) t->cntl );
   }
   // FLA_Syr2k
   else if ( t->func == (void *) FLA_Syr2k_task )
   {
      flash_syr2k_p func;
      func = (flash_syr2k_p) t->func;

      func( ( FLA_Uplo     ) t->int_arg[0],
            ( FLA_Trans    ) t->int_arg[1],
                             t->fla_arg[0],
                             t->input_arg[0],
                             t->input_arg[1],
                             t->fla_arg[1],
                             t->output_arg[0],
            ( fla_syr2k_t* ) t->cntl );
   }
   // FLA_Trmm
   else if ( t->func == (void *) FLA_Trmm_task )
   {
      flash_trmm_p func;
      func = (flash_trmm_p) t->func;

      func( ( FLA_Side    ) t->int_arg[0],
            ( FLA_Uplo    ) t->int_arg[1],
            ( FLA_Trans   ) t->int_arg[2],
            ( FLA_Diag    ) t->int_arg[3],
                            t->fla_arg[0],
                            t->input_arg[0],
                            t->output_arg[0],
            ( fla_trmm_t* ) t->cntl );
   }
   // FLA_Trsm
   else if ( t->func == (void *) FLA_Trsm_task )
   {
      flash_trsm_p func;
      func = (flash_trsm_p) t->func;

      func( ( FLA_Side    ) t->int_arg[0],
            ( FLA_Uplo    ) t->int_arg[1],
            ( FLA_Trans   ) t->int_arg[2],
            ( FLA_Diag    ) t->int_arg[3],
                            t->fla_arg[0],
                            t->input_arg[0],
                            t->output_arg[0],
            ( fla_trsm_t* ) t->cntl );
   }
   else
   {
      FLA_Check_error_code( FLA_NOT_YET_IMPLEMENTED );
   }

#ifdef FLA_ENABLE_SUPERMATRIX_VISUALIZATION
   t->end_time = FLA_Clock();
#endif
   
   return;
}


void FLASH_Queue_verbose_output( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_verbose_output

----------------------------------------------------------------------------*/
{
   int         i, j;
   int         n_tasks = FLASH_Queue_get_num_tasks();
   FLASH_Task* t;
   FLASH_Dep*  d;

   // Grab the head of the task queue.
   t = FLASH_Queue_get_head_task();

   // Iterate over linked list of tasks.
   for ( i = 0; i < n_tasks; i++ )
   {
      printf( "%s;", t->name );

      printf( "In;" );
      for ( j = 0; j < t->n_input_args; j++ )
         printf( "%d[%d,%d];", t->input_arg[j].base->id,
                 t->input_arg[j].base->m_index, 
                 t->input_arg[j].base->n_index );

      printf( "Out;" );
      for ( j = 0; j < t->n_output_args; j++ )
         printf( "%d[%d,%d];", t->output_arg[j].base->id,
                 t->output_arg[j].base->m_index, 
                 t->output_arg[j].base->n_index );

      printf( "Dep" );
      d = t->dep_arg_head;
      for ( j = 0; j < t->n_dep_args; j++ )
      {
         printf( ";%d-%d", d->task->order, d->deptype );
         d = d->next_dep;
      }

      printf( "\n" );

      // Go to the next task.
      t = t->next_task;
   }

   printf( "\n" );

   return;
}


#ifdef FLA_ENABLE_SUPERMATRIX_VISUALIZATION

void FLASH_Queue_visualization( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_visualization

----------------------------------------------------------------------------*/
{
   int         i, j;
   int         n_tasks = FLASH_Queue_get_num_tasks();
   FLASH_Task* t;

   // Grab the head of the task queue.
   t = FLASH_Queue_get_head_task();
   
   // Iterate over linked list of tasks.
   for ( i = 0; i < n_tasks; i++ )
   {
      printf( "%s;%d;%f;%f;", 
              t->name, t->thread, t->begin_time, t->end_time );

      printf( "In;" );
      for ( j = 0; j < t->n_input_args; j++ )
         printf( "%d[%d,%d];", t->input_arg[j].base->id,
                 t->input_arg[j].base->m_index, 
                 t->input_arg[j].base->n_index );
      
      printf( "Out" );
      for ( j = 0; j < t->n_output_args; j++ )
         printf( ";%d[%d,%d]", t->output_arg[j].base->id,
                 t->output_arg[j].base->m_index, 
                 t->output_arg[j].base->n_index );
      
      printf( "\n" );
      
      // Go to the next task.
      t = t->next_task;
   }

   return;
}

#endif


#endif // FLA_ENABLE_SUPERMATRIX
