/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"


#if   FLA_MULTITHREADING_MODEL == FLA_OPENMP
#include <omp.h>
#elif FLA_MULTITHREADING_MODEL == FLA_PTHREADS
#include <pthread.h>
#endif


#ifdef FLA_ENABLE_SUPERMATRIX


typedef struct FLASH_Queue_variables
{
   // A lock on the global task counter.  
   // Needed only when multithreading is enabled.
   FLA_Lock     all_lock;

   // A lock that protects the thread's waiting queue.
   // Needed only when multithreading is enabled.
   FLA_Lock*    run_lock;
   
   // A lock that allows threads to safely check for and place ready dependent
   // tasks on waiting queue.  Needed only when multithreading is enabled.
   FLA_Lock*    dep_lock;

   // A lock that allows threads to safely free the anti-dependency queue
   // within each block.  Needed only when multithreading is enabled.
   FLA_Lock*    war_lock;

   // The waiting queue of tasks for each thread.
   FLASH_Queue* wait_queue;

   // A global task counter that keeps track of how many tasks on the waiting
   // queue have been processed.
   int          pc;

} FLASH_Queue_vars;


void FLASH_Queue_exec( void )
/*----------------------------------------------------------------------------

   FLASH_Queue_exec

----------------------------------------------------------------------------*/
{
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
   FLA_Bool check_params;
#endif
   int      i;
   int      n_tasks   = FLASH_Queue_get_num_tasks();
   int      n_threads = FLASH_Queue_get_num_threads();
   FLA_Bool verbose   = FLASH_Queue_get_verbose_output();

   // If the queue is empty, return early.
   if ( n_tasks == 0 )
      return;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
   // Temporarily disable parameter checking.
   check_params = FLA_Check_parameters_set( FALSE );
#endif
   
   // All the necessary variables for the SuperMatrix mechanism.
   FLASH_Queue_vars args;

   // Allocate different number of elements in arrays if using data affinity.
   int n_memory = ( FLASH_Queue_get_data_affinity() == FLASH_QUEUE_AFFINITY_NONE ? 1 : n_threads );

#ifdef FLA_ENABLE_MULTITHREADING
   // Allocate memory for array of locks.
   FLA_Lock run_lock[n_memory];
   FLA_Lock dep_lock[n_threads];
   FLA_Lock war_lock[n_threads];

   args.run_lock = run_lock;
   args.dep_lock = dep_lock;
   args.war_lock = war_lock;

   // Initialize the all lock.
   FLA_Lock_init( &(args.all_lock) );
   
   // Initialize the run lock for thread i.
   for ( i = 0; i < n_memory; i++ )
   {
      FLA_Lock_init( &(args.run_lock[i]) );
   }

   // Initialize the dep and war locks for thread i.
   for ( i = 0; i < n_threads; i++ )
   {
      FLA_Lock_init( &(args.dep_lock[i]) );
      FLA_Lock_init( &(args.war_lock[i]) );
   }
#endif
   
   // Allocate memory for waiting queue.
   FLASH_Queue wait_queue[n_memory];
   args.wait_queue = wait_queue;

   for ( i = 0; i < n_memory; i++ )
   {
      args.wait_queue[i].n_tasks = 0;
      args.wait_queue[i].head = NULL;
      args.wait_queue[i].tail = NULL;
   }

   // Initialize the aggregate task counter.
   args.pc = 0;

   // Initialize tasks with critical information.
   FLASH_Queue_init_tasks( ( void* ) &args );
   
   // Display verbose output before free all tasks. 
   if ( verbose )
      FLASH_Queue_verbose_output();
   
   // Start timing the parallel execution.
   double dtime = FLA_Clock();
   
#ifdef FLA_ENABLE_MULTITHREADING
   // Parallel Execution!
   FLASH_Queue_exec_parallel( ( void* ) &args );
#else
   // Simulation!
   FLASH_Queue_exec_simulation( ( void* ) &args );
#endif
   
   // End timing the parallel execution.
   dtime = FLA_Clock() - dtime;
   FLASH_Queue_set_parallel_time( dtime );

#ifdef FLA_ENABLE_SUPERMATRIX_VISUALIZATION
   // Visualize all tasks.
   if ( !verbose )
      FLASH_Queue_visualization();
#endif

#ifdef FLA_ENABLE_MULTITHREADING   
   // Destroy the locks.
   FLA_Lock_destroy( &(args.all_lock) );

   for ( i = 0; i < n_memory; i++ )
   {
      FLA_Lock_destroy( &(args.run_lock[i]) );
   }

   for ( i = 0; i < n_threads; i++ )
   {
      FLA_Lock_destroy( &(args.dep_lock[i]) );
      FLA_Lock_destroy( &(args.war_lock[i]) );
   }

#ifdef FLA_ENABLE_SUPERMATRIX_VISUALIZATION
   // Now that we're done with the task array, flush the queue.
   FLASH_Task* t = FLASH_Queue_get_head_task();
   FLASH_Task* next;

   for ( i = 0; i < n_tasks; i++ )
   {
      // Obtain the next task.
      next = t->next_task;

      // Free the current task.
      FLASH_Task_free( t );

      // Move to the next task.
      t = next;      
   }
#endif

#else
   // Now that we're done with the task array, flush the queue.
   FLASH_Task* t = FLASH_Queue_get_head_task();
   FLASH_Task* next;

   for ( i = 0; i < n_tasks; i++ )
   {
      // Obtain the next task.
      next = t->next_task;

      // Free the current task.
      FLASH_Task_free( t );

      // Move to the next task.
      t = next;      
   }
#endif

   // Reset values for next call to FLASH_Queue_exec().
   FLASH_Queue_reset();

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
   // Restore parameter checking to its previous status.
   FLA_Check_parameters_set( check_params );
#endif

   return;
}


void FLASH_Queue_init_tasks( void* arg )
/*----------------------------------------------------------------------------

   FLASH_Queue_init_tasks

----------------------------------------------------------------------------*/
{
   int            i, j;
   int            n_tasks   = FLASH_Queue_get_num_tasks();
   int            n_threads = FLASH_Queue_get_num_threads();
   int            length    = 0;
   int            width     = 0;
   int            sort;
   double         cost;
   FLASH_Data_aff data_aff  = FLASH_Queue_get_data_affinity();
   FLASH_Task*    t;
   FLASH_Dep*     d;

   // Find the 2D factorization of the number of threads.
   if ( data_aff == FLASH_QUEUE_AFFINITY_2D_BLOCK_CYCLIC )
   {
      int sq_rt = 0;
      while ( sq_rt * sq_rt <= n_threads ) sq_rt++;
      sq_rt--;
      while ( n_threads % sq_rt != 0 ) sq_rt--;
      length = n_threads / sq_rt;
      width  = sq_rt;     
   }

   // Grab the tail of the task queue.
   t = FLASH_Queue_get_tail_task();

   for ( i = n_tasks - 1; i >= 0; i-- )
   {
      // Determine the index of the task in the task queue.
      t->order = i;     
      
      // Determine data affinity.
      if ( data_aff == FLASH_QUEUE_AFFINITY_NONE )
      { // No data affinity
         t->queue = 0;
      }
      else if ( data_aff == FLASH_QUEUE_AFFINITY_2D_BLOCK_CYCLIC )
      { // Two-dimensional block cyclic
         t->queue = ( t->output_arg[0].base->m_index % length ) + 
                    ( t->output_arg[0].base->n_index % width  ) * length;
      }
      else
      { // Round-robin
         t->queue = t->queue % n_threads;
      }

      // Determine sorting value.
      sort = 0;
      cost = 0.0;
      
      d = t->dep_arg_head;      
      
      // Take the maximum sort value of dependent tasks.
      for ( j = 0; j < t->n_dep_args; j++ )
      {
         sort = max( sort, d->task->sort );
         cost = max( cost, d->task->weighted_sort );
         
         d = d->next_dep;
      }
      
      t->sort          = sort + 1;
      t->weighted_sort = cost + t->cost;     

      // Find all ready tasks.
      t->n_ready += t->n_input_args + t->n_output_args + t->n_war_args;
      
      if ( t->n_ready == 0 )
      {
         FLASH_Queue_wait_enqueue( t, arg );
      }

      // Go to the previous task.
      t = t->prev_task;
   }

   return;
}


void FLASH_Queue_wait_enqueue( FLASH_Task* t, void* arg )
/*----------------------------------------------------------------------------

   FLASH_Queue_wait_enqueue

----------------------------------------------------------------------------*/
{  
   FLASH_Queue_vars* args = ( FLASH_Queue_vars* ) arg;
   int queue = t->queue;

   if ( args->wait_queue[queue].n_tasks == 0 )
   {
      args->wait_queue[queue].head = t;
      args->wait_queue[queue].tail = t;
   }
   else
   {
      t->prev_wait = args->wait_queue[queue].tail;

      // Insertion sort of tasks in waiting queue.
      if ( FLASH_Queue_get_sorting() )
      {
         while ( t->prev_wait != NULL )
         {
            if ( t->prev_wait->sort >= t->sort )
               break;
            
            t->next_wait = t->prev_wait;
            t->prev_wait = t->prev_wait->prev_wait;
         }         
      }

      // Checking if the task is the head of the waiting queue.      
      if ( t->prev_wait == NULL )
         args->wait_queue[queue].head = t;
      else
         t->prev_wait->next_wait = t;

      // Checking if the task is the tail of the waiting queue.
      if ( t->next_wait == NULL )
         args->wait_queue[queue].tail = t;
      else
         t->next_wait->prev_wait = t;
   }

   // Increment number of tasks on waiting queue.
   args->wait_queue[queue].n_tasks++;

   return;
}


FLASH_Task* FLASH_Queue_wait_dequeue( int queue, void* arg )
/*----------------------------------------------------------------------------

   FLASH_Queue_wait_dequeue

----------------------------------------------------------------------------*/
{
   FLASH_Queue_vars* args = ( FLASH_Queue_vars* ) arg;
   FLASH_Task* t = NULL;

   if ( args->wait_queue[queue].n_tasks > 0 )
   {
      // Dequeue the first task.
      t = args->wait_queue[queue].head;

      if ( args->wait_queue[queue].n_tasks == 1 )
      {
         // Clear the queue of its only task.
         args->wait_queue[queue].head = NULL;
         args->wait_queue[queue].tail = NULL;        
      }
      else
      {
         // Adjust pointers in waiting queue.
         args->wait_queue[queue].head = t->next_wait;
         args->wait_queue[queue].head->prev_wait = NULL;
      }

      // Decrement number of tasks on waiting queue.
      args->wait_queue[queue].n_tasks--;     
   }

   return t;
}


#ifdef FLA_ENABLE_MULTITHREADING

void FLASH_Queue_exec_parallel( void* arg )
/*----------------------------------------------------------------------------

   FLASH_Queue_exec_parallel

----------------------------------------------------------------------------*/
{
   int   i;
   int   n_threads = FLASH_Queue_get_num_threads();
   void* (*thread_entry_point)( void* );

   // Determine which function to send threads to.
   thread_entry_point = FLASH_Queue_exec_parallel_function;

   // Allocate and initialize the thread structures array. Here, an array of
   // FLASH_Thread structures of length n_threads is allocated and the fields
   // of each structure set to appropriate values.
   FLASH_Thread thread[n_threads];

   for ( i = 0; i < n_threads; i++ )
   {
      // Save the thread's identifier.
      thread[i].id = i;

      // Save the pointer to the necessary variables with the thread.
      thread[i].args = arg;

      // The pthread object, if it was even compiled into the FLASH_Thread
      // structure, will be initialized by the pthread implementation when we
      // call pthread_create() and does not need to be touched at this time.
   }

#if FLA_MULTITHREADING_MODEL == FLA_OPENMP

   // An OpenMP parallel for region spawns n_threads threads. Each thread
   // executes the work function with a different FLASH_Thread argument.
   // An implicit synchronization point exists at the end of the curly
   // brace scope.
   #pragma omp parallel for \
           private( i ) \
           shared( thread, n_threads, thread_entry_point ) \
           schedule( static, 1 ) \
           num_threads( n_threads )
   for ( i = 0; i < n_threads; ++i )
   {
      thread_entry_point( ( void* ) &thread[i] );
   }

#elif FLA_MULTITHREADING_MODEL == FLA_PTHREADS

   // Create each POSIX thread needed in addition to the main thread.
   for ( i = 1; i < n_threads; i++ )
   {
      int pthread_e_val;

      // Create thread i with default attributes.
      pthread_e_val = pthread_create( &(thread[i].pthread_obj),
                                      NULL,
                                      thread_entry_point,
                                      ( void* ) &thread[i] );

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
      FLA_Error e_val = FLA_Check_pthread_create_result( pthread_e_val );
      FLA_Check_error_code( e_val );
#endif
   }

   // The main thread is assigned the role of thread 0. Here we manually
   // execute it as a worker thread.
   thread_entry_point( ( void* ) &thread[0] );

   // Wait for non-main threads to finish.
   for ( i = 1; i < n_threads; i++ )
   {
      // These two variables are declared local to this for loop since this
      // is the only place they are needed, and since they would show up as
      // unused variables if FLA_MULTITHREADING_MODEL == FLA_PTHREADS.
      // Strangely, the Intel compiler produces code that results in an
      // "unaligned access" runtime message if thread_status is declared as
      // an int. Declaring it as a long or void* appears to force the
      // compiler (not surprisingly) into aligning it to an 8-byte boundary.
      int   pthread_e_val;
      void* thread_status;

      // Wait for thread i to invoke its respective pthread_exit().
      // The return value passed to pthread_exit() is provided to us
      // via status, if one was given.
      pthread_e_val = pthread_join( thread[i].pthread_obj,
                                    ( void** ) &thread_status );
      
#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
      FLA_Error e_val = FLA_Check_pthread_join_result( pthread_e_val );
      FLA_Check_error_code( e_val );
#endif
   }
   
#endif

   return;
}


//#include <sched.h>
//#include <sys/types.h>
//#include <linux/unistd.h>
//#include <errno.h>
//#include <unistd.h>
//#include <sys/syscall.h>


void* FLASH_Queue_exec_parallel_function( void* arg )
/*----------------------------------------------------------------------------

   FLASH_Queue_exec_parallel_function

----------------------------------------------------------------------------*/
{
   FLASH_Queue_vars* args;   
   int           i, queue;
   int           n_tasks   = FLASH_Queue_get_num_tasks();
   FLA_Bool      condition = TRUE;
   FLA_Bool      available;
   FLASH_Task*   t;
   FLASH_Thread* me;
   //cpu_set_t     cpu_set;

   // Interpret the thread argument as what it really is--a pointer to an
   // FLASH_Thread structure.
   me = ( FLASH_Thread* ) arg;

   // Extract the variables from the current thread.
   args = ( FLASH_Queue_vars* ) me->args;

   // Figure out the id of the current thread.
   i = me->id;

   // Use different queues depending on if using data affinity or not.
   if ( FLASH_Queue_get_data_affinity() != FLASH_QUEUE_AFFINITY_NONE )
   {
      queue = i;
   }
   else // No data affinity.
   {
      queue = 0;
   }

   // Set the CPU affinity; We want the current thread i to run only on CPU i.
   //CPU_ZERO( &cpu_set );
   //CPU_SET( i, &cpu_set );
   //sched_setaffinity( syscall( __NR_gettid ), sizeof(cpu_set_t), &cpu_set );
   
   // Loop until all the tasks have committed.
   while ( condition )
   {
      FLA_Lock_acquire( &(args->run_lock[queue]) ); // R ***

      // Obtain task to execute.
      t = FLASH_Queue_wait_dequeue( queue, ( void* ) args );

      FLA_Lock_release( &(args->run_lock[queue]) ); // R ***

      // Dequeued a task from the waiting queue.
      available = ( t != NULL );

      if ( available )
      {
         // Save the executing thread.
         t->thread = i;

         // Execute the task.
         FLASH_Queue_exec_task( t );         

         // Update task dependencies.
         FLASH_Task_update_dependencies( t, ( void* ) args );

#ifndef FLA_ENABLE_SUPERMATRIX_VISUALIZATION
         // Free the task once it executes in parallel.
         FLASH_Task_free_parallel( t, ( void* ) args );
#endif         
      }

      FLA_Lock_acquire( &(args->all_lock) ); // A ***

      // Increment program counter.
      if ( available )
         args->pc++;

      // Terminate loop.
      if ( args->pc >= n_tasks )
         condition = FALSE;
      
      FLA_Lock_release( &(args->all_lock) ); // A ***
   }
   
#if FLA_MULTITHREADING_MODEL == FLA_PTHREADS
   // If this is a non-main thread, then exit with a zero (normal) error code.
   // The main thread cannot call pthread_exit() because this routine never
   // returns. The main thread must proceed so it can oversee the joining of
   // the exited non-main pthreads.
   if ( i != 0 )
      pthread_exit( ( void* ) NULL );
#endif

   return ( void* ) NULL;
}


void FLASH_Task_update_dependencies( FLASH_Task* t, void* arg )
/*----------------------------------------------------------------------------

   FLASH_Task_update_dependencies

----------------------------------------------------------------------------*/
{
   FLASH_Queue_vars* args = ( FLASH_Queue_vars* ) arg;
   int         i, queue;
   FLA_Bool    available;
   FLASH_Task* task;
   FLASH_Dep*  d;

   // Check each dependent task.
   d = t->dep_arg_head;
   
   for ( i = 0; i < t->n_dep_args; i++ )
   {
      task  = d->task;
      queue = task->queue;
      
      FLA_Lock_acquire( &(args->dep_lock[queue]) ); // D ***
      
      task->n_ready--;
      available = ( task->n_ready == 0 );
      
      FLA_Lock_release( &(args->dep_lock[queue]) ); // D ***
      
      // Place newly ready tasks on sorted queue.
      if ( available )
      {
         FLA_Lock_acquire( &(args->run_lock[queue]) ); // R ***
         
         FLASH_Queue_wait_enqueue( task, arg );
         
         FLA_Lock_release( &(args->run_lock[queue]) ); // R ***
      }
      
      // Go to the next dep.
      d = d->next_dep;
   }

   return;
}


void FLASH_Task_free_parallel( FLASH_Task* t, void* arg )
/*----------------------------------------------------------------------------

   FLASH_Task_free_parallel

----------------------------------------------------------------------------*/
{
   FLASH_Queue_vars* args = ( FLASH_Queue_vars* ) arg;   
   int        i, j, k;
   int        queue;
   int        n_threads = FLASH_Queue_get_num_threads();
   FLASH_Dep* d;
   FLASH_Dep* next_dep;
   
   // Clearing the last write task in each output block.
   for ( i = 0; i < t->n_output_args; i++ )
      t->output_arg[i].base->enqueue->write_task = NULL;

   // Cleaning the last read tasks in each input block.
   for ( i = 0; i < t->n_input_args; i++ )
   {
      queue = t->input_arg[i].base->n_read_blocks % n_threads;

      FLA_Lock_acquire( &(args->war_lock[queue]) ); // W ***

      k = t->input_arg[i].base->enqueue->n_read_tasks;
      d = t->input_arg[i].base->enqueue->read_task_head;

      t->input_arg[i].base->enqueue->n_read_tasks   = 0;
      t->input_arg[i].base->enqueue->read_task_head = NULL;
      t->input_arg[i].base->enqueue->read_task_tail = NULL;

      FLA_Lock_release( &(args->war_lock[queue]) ); // W ***

      for ( j = 0; j < k; j++ )
      {
         next_dep = d->next_dep;
         FLA_free( d );
         d = next_dep;
      }
   }

   // Free the dep_arg field of t.
   d = t->dep_arg_head;

   for ( i = 0; i < t->n_dep_args; i++ )
   {
      next_dep = d->next_dep;
      FLA_free( d );
      d = next_dep;
   }
   
   // Free the int_arg field of t.
   FLA_free( t->int_arg );
   
   // Free the fla_arg field of t.
   FLA_free( t->fla_arg );

   // Free the input_arg field of t.
   FLA_free( t->input_arg );

   // Free the output_arg field of t.
   FLA_free( t->output_arg );

   // Finally, free the struct itself.
   FLA_free( t );

   return;
}

#endif


// ============================================================================


#ifndef FLA_ENABLE_MULTITHREADING

void FLASH_Queue_exec_simulation( void* arg )
/*----------------------------------------------------------------------------

   FLASH_Queue_exec_simulation

----------------------------------------------------------------------------*/
{
   FLASH_Queue_vars* args = ( FLASH_Queue_vars* ) arg;
   int         i, j;
   int         queue;
   int         n_stages  = 0;
   int         n_tasks   = FLASH_Queue_get_num_tasks();
   int         n_threads = FLASH_Queue_get_num_threads();
   FLA_Bool    verbose   = FLASH_Queue_get_verbose_output();
   FLASH_Task* task;
   FLASH_Task* t;
   FLASH_Dep*  d;

   // An array to hold tasks to be executed during of simulation.
   FLASH_Task* exec_array[n_threads];

   // Initialize all exec_array to NULL.
   for ( i = 0; i < n_threads; i++ )
      exec_array[i] = NULL;
   
   // Loop until all the tasks have committed.
   while ( args->pc < n_tasks )
   {
      for ( i = 0; i < n_threads; i++ )
      {
         // Update waiting queue with ready tasks.
         t = exec_array[i];
         
         if ( t != NULL )
         {
            // Check each dependent task.
            d = t->dep_arg_head;
            
            for ( j = 0; j < t->n_dep_args; j++ )
            {
               task = d->task;              
               task->n_ready--;
               
               // Place newly ready tasks on waiting queue.
               if ( task->n_ready == 0 )
               {
                  FLASH_Queue_wait_enqueue( task, arg );
               }
               
               // Go to the next dep.
               d = d->next_dep;
            }
         }
      }      
      
      n_stages++;
      if ( !verbose )
         printf( "%7d", n_stages );
      
      // Move ready tasks from the waiting queue to execution queue.
      for ( i = 0; i < n_threads; i++ )
      {
         // Use different queues depending on if using data affinity or not.
         if ( FLASH_Queue_get_data_affinity() != FLASH_QUEUE_AFFINITY_NONE )
         {
            queue = i;
         }
         else // No data affinity.
         {
            queue = 0;
         }

         t = FLASH_Queue_wait_dequeue( queue, arg );
         exec_array[i] = t;

         // Save the executing thread and increment program counter.
         if ( t != NULL )
         {
            t->thread = i;
            args->pc++;
         }
      }

      // Execute independent tasks.
      for ( i = 0; i < n_threads; i++ )
      {
         t = exec_array[i];
         FLASH_Queue_exec_task( t );
         
         if ( !verbose )
            printf( "%7s", ( t == NULL ? "     " : t->name ) );        
      }
      
      if ( !verbose ) 
         printf( "\n" );
   }
   
   if ( !verbose )
      printf( "\n" );

   return;
}

#endif


#endif // FLA_ENABLE_SUPERMATRIX
