/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

/*
    This file was originally authored by Kazushige Goto (TACC, UT-Austin).
    We use it with his permission.
*/

#include "FLAME.h"

double dclock( void );

double FLA_Clock()
{
  return dclock();
}

double FLA_F2C( fla_clock_f ) ( void )
{
  return FLA_Clock();
}

// --- Begin non-Windows build definitions -------------------------------------
#ifndef FLA_ENABLE_WINDOWS_BUILD

// This function is only used internally, which is why its prototype is not
// with the other "util" directory function prototypes.
void   detect_clocks( void );

// A global variable used when dclock() is defined in terms of gettimeofday().
double gtod_ref_time_sec = 0.0;


// --- Begin portable dclock() definitions -------------------------------------
#ifdef FLA_ENABLE_PORTABLE_DCLOCK

double dclock()
{
  double the_time, norm_sec;
  struct timeval tv;

  gettimeofday( &tv, NULL );

  // If this is the first invocation of through dclock(), then initialize the 
  // "reference time" global variable to the seconds field of the tv struct.
  if( gtod_ref_time_sec == 0.0 )
    gtod_ref_time_sec = ( double ) tv.tv_sec;

  // Normalize the seconds field of the tv struct so that it is relative to the
  // "reference time" that was recorded during the first invocation of dclock().
  norm_sec = ( double ) tv.tv_sec - gtod_ref_time_sec;

  // Compute the number of seconds since the reference time.
  the_time = norm_sec + tv.tv_usec * 1.0e-6;

  return the_time;
}

#else
// --- Begin possibly non-portable dclock() definitions ------------------------


// Global variables that are used only for non-portable dclock() definitions.
static double        clocks             = 0.0;
#ifdef __i386__
static unsigned int  initialclockoffset = 0;
#endif



// --- Begin ia64 section ------------------------------------------------------
#ifdef __ia64__


static __inline unsigned long rdtsc( void );
static __inline unsigned long rdtsc()
{
  unsigned long clocks;

#ifdef __INTEL_COMPILER
  clocks = __getReg(_IA64_REG_AR_ITC);  
#else
  __asm__ __volatile__ ("mov %0=ar.itc" : "=r"(clocks));
#endif

  return clocks;
}


double dclock()
{
  unsigned long totalclocks;

  if ( clocks == 0.0 ) detect_clocks();
  totalclocks = rdtsc();

  return (double)totalclocks / clocks;
}


// --- End ia64 section --------------------------------------------------------
#else
// --- Begin i386 section ------------------------------------------------------
#ifdef __i386__

// rdtsc was once declared static. neccessary?
inline void rdtsc( unsigned int *high, unsigned int *low );
inline void rdtsc( unsigned int *high, unsigned int *low )
{
  asm("rdtsc" : "=a" (*low), "=d"(*high): : "cc");
}


double dclock()
{
  unsigned int high, low;
  unsigned long long totalclocks;

  if (!clocks) detect_clocks();
  rdtsc(&high, &low);
  high -= initialclockoffset;
  totalclocks = (((unsigned long long)high) << 32)
                | (unsigned long long)low;

  return (double)totalclocks / clocks;
}


// --- End i386 section --------------------------------------------------------
#else
// --- Begin non-ia64, non-i386 section ----------------------------------------


double dclock()
{
  double the_time, norm_sec;
  struct timeval tv;

  gettimeofday( &tv, NULL );

  // If this is the first invocation of through dclock(), then initialize the 
  // "reference time" global variable to the seconds field of the tv struct.
  if( gtod_ref_time_sec == 0.0 )
    gtod_ref_time_sec = ( double ) tv.tv_sec;

  // Normalize the seconds field of the tv struct so that it is relative to the
  // "reference time" that was recorded during the first invocation of dclock().
  norm_sec = ( double ) tv.tv_sec - gtod_ref_time_sec;

  // Compute the number of seconds since the reference time.
  the_time = norm_sec + tv.tv_usec * 1.0e-6;

  return the_time;
}

#endif
#endif
// --- End non-ia64, non-i386 section ------------------------------------------
// --- Return to possibly non-portable dclock() definitions --------------------


// This implementation of detect_clocks() works only on operating systems that
// support the /proc filesystem (in particular, GNU/Linux).
// We place this function here at the end so that rdtsc() has been defined
// (and prototyped) by the time we get here. Otherwise the compiler spits out
// warnings.
void detect_clocks()
{
  FILE *infile;
  char buffer[256], *p;
#ifdef __i386__
  unsigned int high, low;
#endif

  if ( clocks == 0.0 )
  {
    p = (char *)NULL;
    infile = fopen("/proc/cpuinfo", "r");
    while (fgets(buffer, sizeof(buffer), infile))
    {
      if (!strncmp("cpu MHz", buffer, 6))
      {
        p = strchr(buffer, ':') + 1;
        break;
      }
    }
    clocks = 1.e6 * atof(p);
#ifdef __i386__
    rdtsc(&high, &low);
    initialclockoffset = high;
#endif
  }
}

// --- End possibly non-portable dclock() definitions --------------------------
#endif

// --- End possibly non-Windows build definitions ------------------------------
// --- Begin Windows build definitions -----------------------------------------
#else

#define WIN32_LEAN_AND_MEAN
#define VC_EXTRALEAN
#include <windows.h>

static LARGE_INTEGER clock_freq = 0;

double dclock()
{
  LARGE_INTEGER clock_val;
  BOOL          r_val;

  if ( clock_freq.QuadPart == 0 )
  {
    r_val = QueryPerformanceFrequency( &clock_freq );

    if ( r_val == 0 )
    {
      FLA_Print_message( "QueryPerformanceFrequency() failed", __FILE__, __LINE__ );
      FLA_Abort();
    }
  }

  r_val = QueryPerformanceCounter( &clock_val );

  if ( r_val == 0 )
  {
    FLA_Print_message( "QueryPerformanceCounter() failed", __FILE__, __LINE__ );
    FLA_Abort();
  }

  return ( double ) ( clock_val.QuadPart / clock_freq.QuadPart );
}

#endif
