/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Hermitianize( FLA_Uplo uplo, FLA_Obj A )
{
  FLA_Datatype datatype;
  int          j;
  int          n_A, ldim_A;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Hermitianize_check( uplo, A );
#endif

  FLA_Symmetrize( uplo, A );

  if ( FLA_Obj_is_real( A ) ) return FLA_SUCCESS;

  if ( uplo == FLA_LOWER_TRIANGULAR )
    FLA_Conjugate_t( FLA_UPPER_TRIANGULAR, A );
  else
    FLA_Conjugate_t( FLA_LOWER_TRIANGULAR, A );


  datatype = FLA_Obj_datatype( A );

  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  switch ( datatype ){

  case FLA_COMPLEX:
  {
    scomplex *buff_A = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_0 = ( scomplex * ) FLA_COMPLEX_PTR( FLA_ZERO );

    for ( j = 0; j < n_A; j++ )
      buff_A[ j*ldim_A + j ].imag = buff_0->imag;

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_0 = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( FLA_ZERO );

    for ( j = 0; j < n_A; j++ )
      buff_A[ j*ldim_A + j ].imag = buff_0->imag;

    break;
  }

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_hermitianize_f )
     ( F_INT *uplo, F_INT *A, F_INT *IERROR )
{
  *IERROR = FLA_Hermitianize( *( ( FLA_Uplo * ) uplo ),
                              *( ( FLA_Obj  * ) A    ) );
}

