/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Max_abs_value( FLA_Obj A, FLA_Obj amax )
{
  FLA_Datatype datatype;
  int          m_A, n_A, ldim_A;
  int          i, j;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Max_abs_value_check( A, amax );
#endif

  m_A    = FLA_Obj_length( A );
  n_A    = FLA_Obj_width( A );
  ldim_A = FLA_Obj_ldim( A );

  datatype = FLA_Obj_datatype( A );
  
  switch ( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_amax  = ( float * ) FLA_FLOAT_PTR( amax );
    float  curr_amax;
    float  temp_amax;

    // Initialize the search with the absolute value of the first element.
    curr_amax = ( float ) fabs( buff_A[0] );

    // Inspect each element, saving values in curr_amax that are larger than
    // the previous elements.
    for( j = 0; j < n_A; j++ )
    {
      for( i = 0; i < m_A; i++ )
      {
        temp_amax = ( float ) fabs( buff_A[ j * ldim_A + i ] );

        if ( curr_amax < temp_amax )
          curr_amax = temp_amax;
      }
    }

    // Copy the result into the amax object buffer.
    *buff_amax = curr_amax;

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_amax  = ( double * ) FLA_DOUBLE_PTR( amax );
    double  curr_amax;
    double  temp_amax;

    // Initialize the search with the absolute value of the first element.
    curr_amax = ( double ) fabs( buff_A[0] );

    // Inspect each element, saving values in curr_amax that are larger than
    // the previous elements.
    for( j = 0; j < n_A; j++ )
    {
      for( i = 0; i < m_A; i++ )
      {
        temp_amax = ( double ) fabs( buff_A[ j * ldim_A + i ] );

        if ( curr_amax < temp_amax )
          curr_amax = temp_amax;
      }
    }

    // Copy the result into the amax object buffer.
    *buff_amax = curr_amax;


    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
    float    *buff_amax  = ( float    * ) FLA_FLOAT_PTR( amax );
    scomplex *curr_value;
    scomplex *temp_value;
    float     curr_amax;
    float     temp_amax;

    curr_value = buff_A;

    // Initialize the search with the absolute value of the first element.
    curr_amax = ( float ) sqrt( curr_value->real * curr_value->real + 
                                curr_value->imag * curr_value->imag ); 

    // Inspect each element, saving values in curr_amax that are larger than
    // the previous elements.
    for( j = 0; j < n_A; j++ )
    {
      for( i = 0; i < m_A; i++ )
      {
        temp_value = buff_A + j * ldim_A + i;

        temp_amax = ( float ) sqrt( temp_value->real * temp_value->real + 
                                    temp_value->imag * temp_value->imag ); 

        if ( curr_amax < temp_amax )
          curr_amax = temp_amax;
      }
    }

    // Copy the result into the amax object buffer.
    *buff_amax = curr_amax;

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    double   *buff_amax  = ( double   * ) FLA_DOUBLE_PTR( amax );
    dcomplex *curr_value;
    dcomplex *temp_value;
    double    curr_amax;
    double    temp_amax;

    curr_value = buff_A;

    // Initialize the search with the absolute value of the first element.
    curr_amax = ( double ) sqrt( curr_value->real * curr_value->real + 
                                 curr_value->imag * curr_value->imag ); 

    // Inspect each element, saving values in curr_amax that are larger than
    // the previous elements.
    for( j = 0; j < n_A; j++ )
    {
      for( i = 0; i < m_A; i++ )
      {
        temp_value = buff_A + j * ldim_A + i;

        temp_amax = ( double ) sqrt( temp_value->real * temp_value->real + 
                                     temp_value->imag * temp_value->imag ); 

        if ( curr_amax < temp_amax )
          curr_amax = temp_amax;
      }
    }

    // Copy the result into the amax object buffer.
    *buff_amax = curr_amax;

    break;
  }

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_max_abs_value_f )
     ( F_INT *A, F_INT *amax, F_INT *IERROR )
{
  *IERROR = FLA_Max_abs_value( *( ( FLA_Obj * ) A    ),
                               *( ( FLA_Obj * ) amax ) );
}

