/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Random_tri_matrix( FLA_Uplo uplo, FLA_Diag diag, FLA_Obj A )
{
  FLA_Datatype datatype;
  int          i, j;
  int          m_A, n_A, ldim_A;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Random_tri_matrix_check( uplo, diag, A );
#endif

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  switch( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_0 = ( float * ) FLA_FLOAT_PTR( FLA_ZERO );
    float *buff_1 = ( float * ) FLA_FLOAT_PTR( FLA_ONE );

    if ( uplo == FLA_LOWER_TRIANGULAR )
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j; i++ )
          buff_A[ j*ldim_A + i ] = *buff_0;

        if      ( diag == FLA_UNIT_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_1;
        else if ( diag == FLA_ZERO_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_0;
        else
          buff_A[ j*ldim_A + j ] = FLA_random_float();

        for ( i = j + 1; i < m_A; i++ )
          buff_A[ j*ldim_A + i ] = FLA_random_float();
      } 
    }
    else if ( uplo == FLA_UPPER_TRIANGULAR ) 
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j; i++ )
          buff_A[ j*ldim_A + i ] = FLA_random_float();

        if      ( diag == FLA_UNIT_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_1;
        else if ( diag == FLA_ZERO_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_0;
        else
          buff_A[ j*ldim_A + j ] = FLA_random_float();

        for ( i = j + 1; i < m_A; i++ )
          buff_A[ j*ldim_A + i ] = *buff_0;
      } 
    }

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_0 = ( double * ) FLA_DOUBLE_PTR( FLA_ZERO );
    double *buff_1 = ( double * ) FLA_DOUBLE_PTR( FLA_ONE );

    if ( uplo == FLA_LOWER_TRIANGULAR )
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j; i++ )
          buff_A[ j*ldim_A + i ] = *buff_0;

        if      ( diag == FLA_UNIT_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_1;
        else if ( diag == FLA_ZERO_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_0;
        else
          buff_A[ j*ldim_A + j ] = FLA_random_double();

        for ( i = j + 1; i < m_A; i++ )
          buff_A[ j*ldim_A + i ] = FLA_random_double();
      } 
    }
    else if ( uplo == FLA_UPPER_TRIANGULAR ) 
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j; i++ )
          buff_A[ j*ldim_A + i ] = FLA_random_double();

        if      ( diag == FLA_UNIT_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_1;
        else if ( diag == FLA_ZERO_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_0;
        else
          buff_A[ j*ldim_A + j ] = FLA_random_double();

        for ( i = j + 1; i < m_A; i++ )
          buff_A[ j*ldim_A + i ] = *buff_0;
      } 
    }

    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_0 = ( scomplex * ) FLA_COMPLEX_PTR( FLA_ZERO );
    scomplex *buff_1 = ( scomplex * ) FLA_COMPLEX_PTR( FLA_ONE );

    if ( uplo == FLA_LOWER_TRIANGULAR )
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j; i++ )
          buff_A[ j*ldim_A + i ] = *buff_0;

        if      ( diag == FLA_UNIT_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_1;
        else if ( diag == FLA_ZERO_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_0;
        else
          buff_A[ j*ldim_A + j ] = FLA_random_scomplex();

        for ( i = j + 1; i < m_A; i++ )
          buff_A[ j*ldim_A + i ] = FLA_random_scomplex();
      } 
    }
    else if ( uplo == FLA_UPPER_TRIANGULAR ) 
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j; i++ )
          buff_A[ j*ldim_A + i ] = FLA_random_scomplex();

        if      ( diag == FLA_UNIT_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_1;
        else if ( diag == FLA_ZERO_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_0;
        else
          buff_A[ j*ldim_A + j ] = FLA_random_scomplex();

        for ( i = j + 1; i < m_A; i++ )
          buff_A[ j*ldim_A + i ] = *buff_0;
      } 
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_0 = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( FLA_ZERO );
    dcomplex *buff_1 = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( FLA_ONE );

    if ( uplo == FLA_LOWER_TRIANGULAR )
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j; i++ )
          buff_A[ j*ldim_A + i ] = *buff_0;

        if      ( diag == FLA_UNIT_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_1;
        else if ( diag == FLA_ZERO_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_0;
        else
          buff_A[ j*ldim_A + j ] = FLA_random_dcomplex();

        for ( i = j + 1; i < m_A; i++ )
          buff_A[ j*ldim_A + i ] = FLA_random_dcomplex();
      } 
    }
    else if ( uplo == FLA_UPPER_TRIANGULAR ) 
    {
      for ( j = 0; j < n_A; j++ )
      {
        for ( i = 0; i < j; i++ )
          buff_A[ j*ldim_A + i ] = FLA_random_dcomplex();

        if      ( diag == FLA_UNIT_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_1;
        else if ( diag == FLA_ZERO_DIAG )
          buff_A[ j*ldim_A + j ] = *buff_0;
        else
          buff_A[ j*ldim_A + j ] = FLA_random_dcomplex();

        for ( i = j + 1; i < m_A; i++ )
          buff_A[ j*ldim_A + i ] = *buff_0;
      } 
    }

    break;
  }

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_random_tri_matrix_f ) 
     ( F_INT *uplo, F_INT *diag, F_INT *A, F_INT *IERROR )
{
  *IERROR = FLA_Random_tri_matrix( *( ( FLA_Uplo * ) uplo ),
                                   *( ( FLA_Diag * ) diag ),
                                   *( ( FLA_Obj  * ) A    ) );
}

