/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Form_Q_blk_var1( FLA_Obj Q, FLA_Obj t, FLA_Obj S, int nb_alg )
{
  FLA_Obj QTL,   QTR,      Q00, Q01, Q02, 
          QBL,   QBR,      Q10, Q11, Q12,
                           Q20, Q21, Q22;

  FLA_Obj tT,              t0,
          tB,              t1,
                           t2;

  FLA_Obj SL,    SR,       S0,  S1,  S2;
  FLA_Obj S1T;
  FLA_Obj S2B;
  FLA_Obj Q11_cpy, U11_cpy;

  int b;

  FLA_Part_2x2( Q,    &QTL, &QTR,
                      &QBL, &QBR,     0, 0, FLA_BR );

  FLA_Part_2x1( t,    &tT, 
                      &tB,            0, FLA_BOTTOM );

  FLA_Part_1x2( S,    &SL,  &SR,      0, FLA_RIGHT );

  while ( FLA_Obj_width ( QBR ) != FLA_Obj_width ( Q )   &&
          FLA_Obj_length( QBR ) != FLA_Obj_length( Q )){

    b = FLA_Obj_width( QTL ) % nb_alg;
    if ( b == 0 ) b = nb_alg;

    FLA_Repart_2x2_to_3x3( QTL, /**/ QTR,       &Q00, &Q01, /**/ &Q02,
                                                &Q10, &Q11, /**/ &Q12,
                        /* ************* */   /* ******************** */
                           QBL, /**/ QBR,       &Q20, &Q21, /**/ &Q22,
                           b, b, FLA_TL );

    FLA_Repart_2x1_to_3x1( tT,                &t0, 
                                              &t1, 
                        /* ** */            /* ** */
                           tB,                &t2,        b, FLA_TOP );

    FLA_Repart_1x2_to_1x3( SL,  /**/ SR,        &S0, &S1, /**/ &S2,
                           b, FLA_LEFT );

    /*------------------------------------------------------------*/

    /*
    S1T = FLA_Top_part( S1, b );
    */

    FLA_Part_2x1( S1,   &S1T, 
                        &S2B, b, FLA_TOP );

    /*
    U11 = trilu( Q11 );
    U21 = Q21;
    */
    /*    
    W12 = triu( S1T ) * ( U11' * 0 + U21' * Q22 );
    W12 is stored in Q12 which is now zero
    */

    if ( FLA_Obj_width( Q12 ) > 0 ){
      FLA_Gemm( FLA_TRANSPOSE, FLA_NO_TRANSPOSE,
                FLA_ONE, Q21, Q22, FLA_ZERO, Q12 );
      FLA_Trmm( FLA_LEFT, FLA_UPPER_TRIANGULAR, 
                FLA_NO_TRANSPOSE, FLA_NONUNIT_DIAG, 
                FLA_ONE, S1T, Q12 );

      /*
        Q22 = Q22 - Q21 * W12;
        Q12 = 0 - Q11 * W12;
      */

      FLA_Gemm( FLA_NO_TRANSPOSE, FLA_NO_TRANSPOSE, 
                FLA_MINUS_ONE, Q21, Q12, FLA_ONE, Q22 );
      FLA_Trmm( FLA_LEFT, FLA_LOWER_TRIANGULAR, 
                FLA_NO_TRANSPOSE, FLA_UNIT_DIAG,
                FLA_MINUS_ONE, Q11, Q12 );
    }


    /*    
    W11 = triu( S1T ) * ( U11' * I + U21' * 0 ) = triu( S1T ) * U11';
    */

    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, Q11, &Q11_cpy );

    FLA_Copy( Q11, Q11_cpy );
    
    /*
      Q11 = I - U11 * W11 = I - U11 * triu( S1T ) * U11';
      Q21 = 0 - U21 * W11 = - U21 * triu( S1T ) * U11';
    */

    //FLA_Triangularize( Q11, FLA_UNIT_DIAG, FLA_LOWER_TRIANGULAR );
    FLA_Trsm( FLA_RIGHT, FLA_UPPER_TRIANGULAR, 
              FLA_NO_TRANSPOSE, FLA_NONUNIT_DIAG, 
              FLA_MINUS_ONE, S1T, Q11 );
    FLA_Trsm( FLA_RIGHT, FLA_LOWER_TRIANGULAR, 
              FLA_TRANSPOSE, FLA_UNIT_DIAG, 
              FLA_ONE, Q11_cpy, Q11 );
    //FLA_Add_to_diag( .... );

    FLA_Trsm( FLA_RIGHT, FLA_UPPER_TRIANGULAR, 
              FLA_NO_TRANSPOSE, FLA_NONUNIT_DIAG, 
              FLA_MINUS_ONE, S1T, Q21 );
    FLA_Trsm( FLA_RIGHT, FLA_LOWER_TRIANGULAR, 
              FLA_TRANSPOSE, FLA_UNIT_DIAG, 
              FLA_ONE, Q11_cpy, Q21 );

    FLA_Obj_free( &U11_cpy );

    /*------------------------------------------------------------*/

    FLA_Cont_with_3x3_to_2x2( &QTL, /**/ &QTR,       Q00, /**/ Q01, Q02,
                            /* ************** */  /* ****************** */
                                                     Q10, /**/ Q11, Q12,
                              &QBL, /**/ &QBR,       Q20, /**/ Q21, Q22,
                              FLA_BR );

    FLA_Cont_with_3x1_to_2x1( &tT,                t0, 
                            /* ** */           /* ** */
                                                  t1, 
                              &tB,                t2,     FLA_BOTTOM );

    FLA_Cont_with_1x3_to_1x2( &SL,  /**/ &SR,        S0, /**/ S1, S2,
                              FLA_RIGHT );

  }

  return FLA_SUCCESS;
}
