/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Copy_t( FLA_Trans trans, FLA_Obj A, FLA_Obj B )
{
  FLA_Datatype datatype;
  int          j, n_iter;
  int          num_elem;
  int          m_A, n_A, inc_A, ldim_A;
  int          m_B,      inc_B, ldim_B;
  int          ldim_B_trans, inc_B_trans;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Copy_t_check( trans, A, B );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  m_B      = FLA_Obj_length( B );
  ldim_B   = FLA_Obj_ldim( B );

  if ( FLA_Obj_is_vector( A ) )
  {
    inc_A    = ( m_A == 1 ? ldim_A : 1 );
    inc_B    = ( m_B == 1 ? ldim_B : 1 );
    n_iter   = 1;
    num_elem = FLA_Obj_max_dim( A );

    ldim_B_trans = ldim_B;
    inc_B_trans  = inc_B;
  }
  else
  {
    inc_A    = 1;
    inc_B    = 1;
    n_iter   = n_A;
    num_elem = m_A;

    if ( trans == FLA_NO_TRANSPOSE || trans == FLA_CONJ_NO_TRANSPOSE )
    {
      ldim_B_trans = ldim_B;
      inc_B_trans  = inc_B;
    }
    else if ( trans == FLA_TRANSPOSE || trans == FLA_CONJ_TRANSPOSE )
    {
      ldim_B_trans = inc_B;
      inc_B_trans  = ldim_B;
    }
  }


  switch ( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_B = ( float * ) FLA_FLOAT_PTR( B );

    for ( j = 0; j < n_iter; j++ )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_scopy( num_elem,
                   buff_A + j*ldim_A,       inc_A, 
                   buff_B + j*ldim_B_trans, inc_B_trans );
#else
      FLA_C2F( scopy )( &num_elem,
                        buff_A + j*ldim_A,       &inc_A, 
                        buff_B + j*ldim_B_trans, &inc_B_trans );
#endif
    }

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_B = ( double * ) FLA_DOUBLE_PTR( B );

    for ( j = 0; j < n_iter; j++ )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_dcopy( num_elem,
                   buff_A + j*ldim_A,       inc_A, 
                   buff_B + j*ldim_B_trans, inc_B_trans );
#else
      FLA_C2F( dcopy )( &num_elem,
                        buff_A + j*ldim_A,       &inc_A, 
                        buff_B + j*ldim_B_trans, &inc_B_trans );
#endif
    }

    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_B = ( scomplex * ) FLA_COMPLEX_PTR( B );

    for ( j = 0; j < n_iter; j++ )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_ccopy( num_elem,
                   buff_A + j*ldim_A,       inc_A, 
                   buff_B + j*ldim_B_trans, inc_B_trans );
#else
      FLA_C2F( ccopy )( &num_elem,
                        buff_A + j*ldim_A,       &inc_A, 
                        buff_B + j*ldim_B_trans, &inc_B_trans );
#endif
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_B = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( B );

    for ( j = 0; j < n_iter; j++ )
    {
#ifdef FLA_ENABLE_CBLAS_INTERFACE
      cblas_zcopy( num_elem,
                   buff_A + j*ldim_A,       inc_A, 
                   buff_B + j*ldim_B_trans, inc_B_trans );
#else
      FLA_C2F( zcopy )( &num_elem,
                        buff_A + j*ldim_A,       &inc_A, 
                        buff_B + j*ldim_B_trans, &inc_B_trans );
#endif
    }

    break;
  }

  }
  
  if ( trans == FLA_CONJ_NO_TRANSPOSE || trans == FLA_CONJ_TRANSPOSE )
    FLA_Conjugate( B );
  
  return FLA_SUCCESS;
}

void FLA_F2C( fla_copy_t_f ) 
     ( F_INT *trans, F_INT *A, F_INT *B, F_INT *IERROR )
{
  *IERROR = FLA_Copy_t( *( ( FLA_Trans * ) trans ),
                        *( ( FLA_Obj   * ) A     ),
                        *( ( FLA_Obj   * ) B     ) );
}

