/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Scal_r( FLA_Uplo uplo, FLA_Obj alpha, FLA_Obj A )
{
  FLA_Datatype datatype, dt_alpha;
  int          j, i_one = 1;
  int          num_elem;
  int          n_A, ldim_A;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Scal_r_check( uplo, alpha, A );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  dt_alpha = FLA_Obj_datatype( alpha );
  datatype = FLA_Obj_datatype( A );

  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  switch ( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_alpha = ( float * ) FLA_FLOAT_PTR( alpha );

    if ( uplo == FLA_LOWER_TRIANGULAR )
    {
      for ( j = 0; j < n_A; ++j )
      {
        num_elem = n_A - j;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
        cblas_sscal( num_elem, 
                     *buff_alpha,
                     buff_A + j*ldim_A + j, i_one );
#else
        FLA_C2F( sscal )( &num_elem,
                          buff_alpha,
                          buff_A + j*ldim_A + j, &i_one );
#endif
      }
    }
    else
    {
      for ( j = 0; j < n_A; ++j )
      {
        num_elem = j + 1;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
        cblas_sscal( num_elem, 
                     *buff_alpha,
                     buff_A + j*ldim_A, i_one );
#else
        FLA_C2F( sscal )( &num_elem,
                          buff_alpha,
                          buff_A + j*ldim_A, &i_one );
#endif
      }
    }

    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_alpha = ( double * ) FLA_DOUBLE_PTR( alpha );

    if ( uplo == FLA_LOWER_TRIANGULAR )
    {
      for ( j = 0; j < n_A; ++j )
      {
        num_elem = n_A - j;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
        cblas_dscal( num_elem,
                     *buff_alpha,
                     buff_A + j*ldim_A + j, i_one );
#else
        FLA_C2F( dscal )( &num_elem,
                          buff_alpha,
                          buff_A + j*ldim_A + j, &i_one );
#endif
      }
    }
    else
    {
      for ( j = 0; j < n_A; ++j )
      {
        num_elem = j + 1;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
        cblas_dscal( num_elem,
                     *buff_alpha,
                     buff_A + j*ldim_A, i_one );
#else
        FLA_C2F( dscal )( &num_elem,
                          buff_alpha,
                          buff_A + j*ldim_A, &i_one );
#endif
      }
    }

    break;
  }

  case FLA_COMPLEX:
  {
    if ( dt_alpha == FLA_COMPLEX )
    {
      scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
      scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
  
      if ( uplo == FLA_LOWER_TRIANGULAR )
      {
        for ( j = 0; j < n_A; ++j )
        {
          num_elem = n_A - j;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
          cblas_cscal( num_elem,
                       buff_alpha,
                       buff_A + j*ldim_A + j, i_one );
#else
          FLA_C2F( cscal )( &num_elem,
                            buff_alpha,
                            buff_A + j*ldim_A + j, &i_one );
#endif
        }
      }
      else
      {
        for ( j = 0; j < n_A; ++j )
        {
          num_elem = j + 1;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
          cblas_cscal( num_elem,
                       buff_alpha,
                       buff_A + j*ldim_A, i_one );
#else
          FLA_C2F( cscal )( &num_elem,
                            buff_alpha,
                            buff_A + j*ldim_A, &i_one );
#endif
        }
      }
    }
    else if ( dt_alpha == FLA_FLOAT )
    {
      scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
      float    *buff_alpha = ( float    * ) FLA_FLOAT_PTR( alpha );

      if ( uplo == FLA_LOWER_TRIANGULAR )
      {
        for ( j = 0; j < n_A; ++j )
        {
          num_elem = n_A - j;
#ifdef FLA_ENABLE_CBLAS_INTERFACE 
          cblas_csscal( num_elem,
                        *buff_alpha,
                        buff_A + j*ldim_A + j, i_one );
#else
          FLA_C2F( csscal )( &num_elem,
                             buff_alpha,
                             buff_A + j*ldim_A + j, &i_one );
#endif
        }
      }
      else
      {
        for ( j = 0; j < n_A; ++j )
        {
          num_elem = j + 1;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
          cblas_csscal( num_elem,
                        *buff_alpha,
                        buff_A + j*ldim_A, i_one );
#else
          FLA_C2F( csscal )( &num_elem,
                             buff_alpha,
                             buff_A + j*ldim_A, &i_one );
#endif
        }
      }
    }

    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    if ( dt_alpha == FLA_DOUBLE_COMPLEX )
    {
      dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
      dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
  
      if ( uplo == FLA_LOWER_TRIANGULAR )
      {
        for ( j = 0; j < n_A; ++j )
        {
          num_elem = n_A - j;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
          cblas_zscal( num_elem,
                       buff_alpha,
                       buff_A + j*ldim_A + j, i_one );
#else
          FLA_C2F( zscal )( &num_elem,
                            buff_alpha,
                            buff_A + j*ldim_A + j, &i_one );
#endif
        }
      }
      else
      {
        for ( j = 0; j < n_A; ++j )
        {
          num_elem = j + 1;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
          cblas_zscal( num_elem,
                       buff_alpha,
                       buff_A + j*ldim_A, i_one );
#else
          FLA_C2F( zscal )( &num_elem,
                            buff_alpha,
                            buff_A + j*ldim_A, &i_one );
#endif
        }
      }
    }
    else if ( dt_alpha == FLA_DOUBLE )
    {
      dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
      double   *buff_alpha = ( double   * ) FLA_DOUBLE_PTR( alpha );
  
      if ( uplo == FLA_LOWER_TRIANGULAR )
      {
        for ( j = 0; j < n_A; ++j )
        {
          num_elem = n_A - j;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
          cblas_zdscal( num_elem,
                        *buff_alpha,
                        buff_A + j*ldim_A + j, i_one );
#else
          FLA_C2F( zdscal )( &num_elem,
                             buff_alpha,
                             buff_A + j*ldim_A + j, &i_one );
#endif
        }
      }
      else
      {
        for ( j = 0; j < n_A; ++j )
        {
          num_elem = j + 1;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
          cblas_zdscal( num_elem,
                        *buff_alpha,
                        buff_A + j*ldim_A, i_one );
#else
          FLA_C2F( zdscal )( &num_elem,
                             buff_alpha,
                             buff_A + j*ldim_A, &i_one );
#endif
        }
      }
    }

    break;
  }

  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_scal_r_f ) 
     ( F_INT *uplo, F_INT *alpha, F_INT *A, F_INT *IERROR )
{
  *IERROR = FLA_Scal_r( *( ( FLA_Uplo * ) uplo  ),
                        *( ( FLA_Obj  * ) alpha ),
                        *( ( FLA_Obj  * ) A     ) );
}

