/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Gemv( FLA_Trans transa, FLA_Obj alpha, FLA_Obj A, FLA_Obj x, FLA_Obj beta, FLA_Obj y )
{
  FLA_Datatype datatype;
  int          m_A, n_A, ldim_A;
  int          m_x, ldim_x;
  int          m_y, ldim_y;
  int          inc_x, inc_y;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER     cblas_order = CblasColMajor;
  CBLAS_TRANSPOSE blas_transa;
#else
  char         blas_transa;
#endif
  FLA_Obj      A_copy;
  FLA_Trans    transa_copy;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Gemv_check( transa, alpha, A, x, beta, y );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  m_x      = FLA_Obj_length( x );
  ldim_x   = FLA_Obj_ldim( x );

  m_y      = FLA_Obj_length( y );
  ldim_y   = FLA_Obj_ldim( y );

  inc_x    = ( m_x == 1 ? ldim_x : 1 );
  inc_y    = ( m_y == 1 ? ldim_y : 1 );

  if ( transa == FLA_CONJ_NO_TRANSPOSE )
  {
    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, A, &A_copy );
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, A, A_copy );
    transa_copy = FLA_NO_TRANSPOSE;

    ldim_A      = FLA_Obj_ldim( A_copy );
  }
  else
  {
    A_copy      = A;
    transa_copy = transa;
  }

  FLA_Param_map_to_blas_trans( transa_copy, &blas_transa );


  switch( datatype ){
  
  case FLA_FLOAT:
  {
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_x     = ( float * ) FLA_FLOAT_PTR( x );
    float *buff_y     = ( float * ) FLA_FLOAT_PTR( y );
    float *buff_alpha = ( float * ) FLA_FLOAT_PTR( alpha );
    float *buff_beta  = ( float * ) FLA_FLOAT_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_sgemv( cblas_order,
                 blas_transa,
                 m_A, n_A, 
                 *buff_alpha,  
                 buff_A, ldim_A, 
                 buff_x, inc_x,
                 *buff_beta,  
                 buff_y, inc_y );
#else
    FLA_C2F( sgemv )( &blas_transa,
                      &m_A, &n_A, 
                      buff_alpha,  
                      buff_A, &ldim_A, 
                      buff_x, &inc_x,
                      buff_beta,  
                      buff_y, &inc_y );
#endif
    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_x     = ( double * ) FLA_DOUBLE_PTR( x );
    double *buff_y     = ( double * ) FLA_DOUBLE_PTR( y );
    double *buff_alpha = ( double * ) FLA_DOUBLE_PTR( alpha );
    double *buff_beta  = ( double * ) FLA_DOUBLE_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_dgemv( cblas_order,
                 blas_transa,
                 m_A, n_A, 
                 *buff_alpha,  
                 buff_A, ldim_A, 
                 buff_x, inc_x,
                 *buff_beta,  
                 buff_y, inc_y );
#else
    FLA_C2F( dgemv )( &blas_transa,
                      &m_A, &n_A, 
                      buff_alpha,  
                      buff_A, &ldim_A, 
                      buff_x, &inc_x,
                      buff_beta,  
                      buff_y, &inc_y );
#endif
    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A_copy );
    scomplex *buff_x     = ( scomplex * ) FLA_COMPLEX_PTR( x );
    scomplex *buff_y     = ( scomplex * ) FLA_COMPLEX_PTR( y );
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
    scomplex *buff_beta  = ( scomplex * ) FLA_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_cgemv( cblas_order,
                 blas_transa,
                 m_A, n_A, 
                 buff_alpha,  
                 buff_A, ldim_A, 
                 buff_x, inc_x,
                 buff_beta,  
                 buff_y, inc_y );
#else
    FLA_C2F( cgemv )( &blas_transa,
                      &m_A, &n_A, 
                      buff_alpha,  
                      buff_A, &ldim_A, 
                      buff_x, &inc_x,
                      buff_beta,  
                      buff_y, &inc_y );
#endif
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A_copy );
    dcomplex *buff_x     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( x );
    dcomplex *buff_y     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( y );
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
    dcomplex *buff_beta  = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_zgemv( cblas_order,
                 blas_transa,
                 m_A, n_A, 
                 buff_alpha,  
                 buff_A, ldim_A, 
                 buff_x, inc_x,
                 buff_beta,  
                 buff_y, inc_y );
#else
    FLA_C2F( zgemv )( &blas_transa,
                      &m_A, &n_A, 
                      buff_alpha,  
                      buff_A, &ldim_A, 
                      buff_x, &inc_x,
                      buff_beta,  
                      buff_y, &inc_y );
#endif
    break;
  }

  }

  if ( transa == FLA_CONJ_NO_TRANSPOSE )
    FLA_Obj_free( &A_copy );

  return FLA_SUCCESS;
}


void FLA_F2C( fla_gemv_f ) 
     ( F_INT *trans, F_INT *alpha, F_INT *A, F_INT *x, F_INT *beta, F_INT *y, F_INT *IERROR )
{
  *IERROR = FLA_Gemv( *( ( FLA_Trans * ) trans ), 
                      *( ( FLA_Obj   * ) alpha ),
                      *( ( FLA_Obj   * ) A     ),
                      *( ( FLA_Obj   * ) x     ),
                      *( ( FLA_Obj   * ) beta  ),
                      *( ( FLA_Obj   * ) y     ) );
}

