/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Hemv( FLA_Uplo uplo, FLA_Obj alpha, FLA_Obj A, FLA_Obj x, FLA_Obj beta, FLA_Obj y )
{
  FLA_Datatype datatype;
  int          m_A, ldim_A;
  int          m_x, inc_x, ldim_x;
  int          m_y, inc_y, ldim_y;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER  cblas_order = CblasColMajor;
  CBLAS_UPLO   blas_uplo;   
#else
  char         blas_uplo;
#endif

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Hemv_check( uplo, alpha, A, x, beta, y );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  ldim_A   = FLA_Obj_ldim( A );

  m_x      = FLA_Obj_length( x );
  ldim_x   = FLA_Obj_ldim( x );

  m_y      = FLA_Obj_length( y );
  ldim_y   = FLA_Obj_ldim( y );

  inc_x    = ( m_x == 1 ? ldim_x : 1 );
  inc_y    = ( m_y == 1 ? ldim_y : 1 );

  FLA_Param_map_to_blas_uplo( uplo, &blas_uplo );


  switch( datatype ){
  
  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_x     = ( scomplex * ) FLA_COMPLEX_PTR( x );
    scomplex *buff_y     = ( scomplex * ) FLA_COMPLEX_PTR( y );
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
    scomplex *buff_beta  = ( scomplex * ) FLA_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_chemv( cblas_order,
                 blas_uplo,
                 m_A,
                 buff_alpha,
                 buff_A, ldim_A, 
                 buff_x, inc_x,
                 buff_beta,
                 buff_y, inc_y );
#else
    FLA_C2F( chemv )( &blas_uplo,
                      &m_A,
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_x, &inc_x,
                      buff_beta,
                      buff_y, &inc_y );
#endif
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_x     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( x );
    dcomplex *buff_y     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( y );
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
    dcomplex *buff_beta  = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_zhemv( cblas_order,
                 blas_uplo,
                 m_A,
                 buff_alpha,
                 buff_A, ldim_A,
                 buff_x, inc_x,
                 buff_beta,
                 buff_y, inc_y );
#else
    FLA_C2F( zhemv )( &blas_uplo,
                      &m_A,
                      buff_alpha,
                      buff_A, &ldim_A,
                      buff_x, &inc_x,
                      buff_beta,
                      buff_y, &inc_y );
#endif
    break;
  }

  }
  
  return FLA_SUCCESS;
}

void FLA_F2C( fla_hemv_f ) 
     ( F_INT *uplo, F_INT *alpha, F_INT *A, F_INT *x, F_INT *beta, F_INT *y, F_INT *IERROR )
{
  *IERROR = FLA_Hemv( *( ( FLA_Uplo * ) uplo  ), 
                      *( ( FLA_Obj  * ) alpha ),
                      *( ( FLA_Obj  * ) A     ),
                      *( ( FLA_Obj  * ) x     ),
                      *( ( FLA_Obj  * ) beta  ),
                      *( ( FLA_Obj  * ) y     ) );
}

