/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Trmv( FLA_Uplo uplo, FLA_Trans transa, FLA_Diag diag, FLA_Obj A, FLA_Obj x ) 
{
  FLA_Datatype datatype;
  int          m_A, ldim_A;
  int          m_x, ldim_x, inc_x;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER     cblas_order = CblasColMajor;
  CBLAS_UPLO      blas_uplo;   
  CBLAS_TRANSPOSE blas_transa;
  CBLAS_DIAG      blas_diag;
#else
  char         blas_uplo;
  char         blas_transa;
  char         blas_diag; 
#endif
  FLA_Obj      x_copy;
  FLA_Trans    transa_copy;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Trmv_check( uplo, transa, diag, A, x );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  ldim_A   = FLA_Obj_ldim( A );

  m_x      = FLA_Obj_length( x );
  ldim_x   = FLA_Obj_ldim( x );

  inc_x    = ( m_x == 1 ? ldim_x : 1 );

  if ( transa == FLA_CONJ_NO_TRANSPOSE )
  {
    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, x, &x_copy );
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, x, x_copy );
    transa_copy = FLA_NO_TRANSPOSE;

    m_x      = FLA_Obj_length( x_copy );
    ldim_x   = FLA_Obj_ldim( x_copy );
    inc_x    = ( m_x == 1 ? ldim_x : 1 );
  }
  else
  {
    x_copy      = x;
    transa_copy = transa;
  }

  FLA_Param_map_to_blas_uplo( uplo, &blas_uplo );
  FLA_Param_map_to_blas_trans( transa_copy, &blas_transa );
  FLA_Param_map_to_blas_diag( diag, &blas_diag );


  switch( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_x = ( float * ) FLA_FLOAT_PTR( x );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_strmv( cblas_order,
                 blas_uplo,
                 blas_transa,
                 blas_diag,
                 m_A,
                 buff_A, ldim_A,
                 buff_x, inc_x );
#else
    FLA_C2F( strmv )( &blas_uplo,
                      &blas_transa,
                      &blas_diag,
                      &m_A,
                      buff_A, &ldim_A,
                      buff_x, &inc_x );
#endif
    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_x = ( double * ) FLA_DOUBLE_PTR( x );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_dtrmv( cblas_order,
                 blas_uplo,
                 blas_transa,
                 blas_diag,
                 m_A,
                 buff_A, ldim_A,
                 buff_x, inc_x );
#else
    FLA_C2F( dtrmv )( &blas_uplo,
                      &blas_transa,
                      &blas_diag,
                      &m_A,
                      buff_A, &ldim_A,
                      buff_x, &inc_x );
#endif
    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_x = ( scomplex * ) FLA_COMPLEX_PTR( x_copy );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_ctrmv( cblas_order,
                 blas_uplo,
                 blas_transa,
                 blas_diag,
                 m_A,
                 buff_A, ldim_A,
                 buff_x, inc_x );
#else
    FLA_C2F( ctrmv )( &blas_uplo,
                      &blas_transa,
                      &blas_diag,
                      &m_A,
                      buff_A, &ldim_A,
                      buff_x, &inc_x );
#endif
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_x = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( x_copy );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_ztrmv( cblas_order,
                 blas_uplo,
                 blas_transa,
                 blas_diag,
                 m_A,
                 buff_A, ldim_A,
                 buff_x, inc_x );
#else
    FLA_C2F( ztrmv )( &blas_uplo,
                      &blas_transa,
                      &blas_diag,
                      &m_A,
                      buff_A, &ldim_A,
                      buff_x, &inc_x );
#endif
    break;
  }

  }

  if ( transa == FLA_CONJ_NO_TRANSPOSE )
  {
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, x_copy, x );
    FLA_Obj_free( &x_copy );
  }

  return FLA_SUCCESS;
}

void FLA_F2C( fla_trmv_f ) 
     ( F_INT *uplo, F_INT *transa, F_INT *diag, F_INT *A, F_INT *x, F_INT *IERROR )
{
  *IERROR = FLA_Trmv( *( ( FLA_Uplo  * ) uplo   ), 
                      *( ( FLA_Trans * ) transa ),
                      *( ( FLA_Diag  * ) diag   ),
                      *( ( FLA_Obj   * ) A      ),
                      *( ( FLA_Obj   * ) x      ) );
}

