/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Gemm_external( FLA_Trans transa, FLA_Trans transb, FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  FLA_Datatype datatype;
  int          k_AB;
  int          m_A, n_A, ldim_A;
  int          ldim_B; 
  int          m_C, n_C, ldim_C;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER     cblas_order = CblasColMajor;
  CBLAS_TRANSPOSE blas_transa;
  CBLAS_TRANSPOSE blas_transb;
#else
  char         blas_transa; 
  char         blas_transb;
#endif
  FLA_Obj      A_copy;
  FLA_Obj      B_copy;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Gemm_check( transa, transb, alpha, A, B, beta, C );
#endif

  if ( FLA_Obj_has_zero_dim( C ) ) return FLA_SUCCESS;

  if ( FLA_Obj_has_zero_dim( A ) || FLA_Obj_has_zero_dim( B ) )
  {
    FLA_Scal( beta, C );
    return FLA_SUCCESS;
  }

  datatype = FLA_Obj_datatype( A );

  m_A      = FLA_Obj_length( A );
  n_A      = FLA_Obj_width( A );
  ldim_A   = FLA_Obj_ldim( A );

  ldim_B   = FLA_Obj_ldim( B );

  m_C      = FLA_Obj_length( C );
  n_C      = FLA_Obj_width( C );
  ldim_C   = FLA_Obj_ldim( C );

  if ( transa == FLA_NO_TRANSPOSE || transa == FLA_CONJ_NO_TRANSPOSE )
    k_AB = n_A;
  else
    k_AB = m_A;

  if ( transa == FLA_CONJ_NO_TRANSPOSE )
  {
    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, A, &A_copy );
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, A, A_copy );

    ldim_A = FLA_Obj_ldim( A_copy );
  }
  else
  {
    A_copy = A;
  }

  if ( transb == FLA_CONJ_NO_TRANSPOSE )
  {
    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, B, &B_copy );
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, B, B_copy );

    ldim_B = FLA_Obj_ldim( B_copy );
  }
  else
  {
    B_copy = B;
  }

  FLA_Param_map_to_blas_trans( transa, &blas_transa );
  FLA_Param_map_to_blas_trans( transb, &blas_transb );


  switch( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_B     = ( float * ) FLA_FLOAT_PTR( B );
    float *buff_C     = ( float * ) FLA_FLOAT_PTR( C );
    float *buff_alpha = ( float * ) FLA_FLOAT_PTR( alpha );
    float *buff_beta  = ( float * ) FLA_FLOAT_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_sgemm( cblas_order,
                 blas_transa, 
                 blas_transb,
                 m_C,
                 n_C, 
                 k_AB, 
                 *buff_alpha,
                 buff_A, ldim_A, 
                 buff_B, ldim_B,
                 *buff_beta,
                 buff_C, ldim_C );
#else
    FLA_C2F( sgemm )( &blas_transa, 
                      &blas_transb,
                      &m_C,
                      &n_C, 
                      &k_AB, 
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_B, &ldim_B,
                      buff_beta,
                      buff_C, &ldim_C );
#endif
    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_B     = ( double * ) FLA_DOUBLE_PTR( B );
    double *buff_C     = ( double * ) FLA_DOUBLE_PTR( C );
    double *buff_alpha = ( double * ) FLA_DOUBLE_PTR( alpha );
    double *buff_beta  = ( double * ) FLA_DOUBLE_PTR( beta );
                         
#ifdef FLA_ENABLE_CBLAS_INTERFACE           
    cblas_dgemm( cblas_order,
                 blas_transa, 
                 blas_transb,
                 m_C,
                 n_C, 
                 k_AB, 
                 *buff_alpha,
                 buff_A, ldim_A, 
                 buff_B, ldim_B,
                 *buff_beta,
                 buff_C, ldim_C );
#else
    FLA_C2F( dgemm )( &blas_transa, 
                      &blas_transb,
                      &m_C,
                      &n_C, 
                      &k_AB, 
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_B, &ldim_B,
                      buff_beta,
                      buff_C, &ldim_C );
#endif
    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A_copy );
    scomplex *buff_B     = ( scomplex * ) FLA_COMPLEX_PTR( B_copy );
    scomplex *buff_C     = ( scomplex * ) FLA_COMPLEX_PTR( C );
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );
    scomplex *buff_beta  = ( scomplex * ) FLA_COMPLEX_PTR( beta );
                         
#ifdef FLA_ENABLE_CBLAS_INTERFACE           
    cblas_cgemm( cblas_order,
                 blas_transa, 
                 blas_transb,
                 m_C,
                 n_C, 
                 k_AB, 
                 buff_alpha,
                 buff_A, ldim_A, 
                 buff_B, ldim_B,
                 buff_beta,
                 buff_C, ldim_C );
#else
    FLA_C2F( cgemm )( &blas_transa, 
                      &blas_transb,
                      &m_C,
                      &n_C, 
                      &k_AB, 
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_B, &ldim_B,
                      buff_beta,
                      buff_C, &ldim_C );
#endif
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A_copy );
    dcomplex *buff_B     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( B_copy );
    dcomplex *buff_C     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( C );
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );
    dcomplex *buff_beta  = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( beta );

#ifdef FLA_ENABLE_CBLAS_INTERFACE                                    
    cblas_zgemm( cblas_order,
                 blas_transa, 
                 blas_transb,
                 m_C,
                 n_C, 
                 k_AB, 
                 buff_alpha,
                 buff_A, ldim_A, 
                 buff_B, ldim_B,
                 buff_beta,
                 buff_C, ldim_C );
#else
    FLA_C2F( zgemm )( &blas_transa, 
                      &blas_transb,
                      &m_C,
                      &n_C, 
                      &k_AB, 
                      buff_alpha,
                      buff_A, &ldim_A, 
                      buff_B, &ldim_B,
                      buff_beta,
                      buff_C, &ldim_C );
#endif
    break;
  }

  }

  if ( transa == FLA_CONJ_NO_TRANSPOSE )
    FLA_Obj_free( &A_copy );

  if ( transb == FLA_CONJ_NO_TRANSPOSE )
    FLA_Obj_free( &B_copy );
 
  return FLA_SUCCESS;
}

void FLA_F2C( fla_gemm_external_f ) 
     ( F_INT *transa, F_INT *transb, F_INT *alpha, F_INT *A, F_INT *B, F_INT *beta, F_INT *C, F_INT *IERROR )
{
  *IERROR = FLA_Gemm_external( *( ( FLA_Trans * ) transa ), 
                               *( ( FLA_Trans * ) transb ), 
                               *( ( FLA_Obj   * ) alpha  ), 
                               *( ( FLA_Obj   * ) A      ), 
                               *( ( FLA_Obj   * ) B      ),
                               *( ( FLA_Obj   * ) beta   ), 
                               *( ( FLA_Obj   * ) C      ) );
}

FLA_Error FLA_Gemm_cc_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_CONJ_TRANSPOSE, FLA_CONJ_TRANSPOSE, alpha, A, B, beta, C );
}

FLA_Error FLA_Gemm_cn_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_CONJ_TRANSPOSE, FLA_NO_TRANSPOSE, alpha, A, B, beta, C );
}

FLA_Error FLA_Gemm_ct_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_CONJ_TRANSPOSE, FLA_TRANSPOSE, alpha, A, B, beta, C );
}

FLA_Error FLA_Gemm_nc_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_NO_TRANSPOSE, FLA_CONJ_TRANSPOSE, alpha, A, B, beta, C );
}

FLA_Error FLA_Gemm_nn_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_NO_TRANSPOSE, FLA_NO_TRANSPOSE, alpha, A, B, beta, C );
}

FLA_Error FLA_Gemm_nt_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_NO_TRANSPOSE, FLA_TRANSPOSE, alpha, A, B, beta, C );
}

FLA_Error FLA_Gemm_tc_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_TRANSPOSE, FLA_CONJ_TRANSPOSE, alpha, A, B, beta, C );
}

FLA_Error FLA_Gemm_tn_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_TRANSPOSE, FLA_NO_TRANSPOSE, alpha, A, B, beta, C );
}

FLA_Error FLA_Gemm_tt_external( FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  return FLA_Gemm_external( FLA_TRANSPOSE, FLA_TRANSPOSE, alpha, A, B, beta, C );
}

