/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Trsm_external( FLA_Side side, FLA_Uplo uplo, FLA_Trans trans, FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  FLA_Datatype datatype;
  int          ldim_A;
  int          m_B, n_B, ldim_B;
#ifdef FLA_ENABLE_CBLAS_INTERFACE
  CBLAS_ORDER     cblas_order = CblasColMajor;
  CBLAS_SIDE      blas_side;
  CBLAS_UPLO      blas_uplo;
  CBLAS_TRANSPOSE blas_trans;
  CBLAS_DIAG      blas_diag;
#else
  char         blas_side; 
  char         blas_uplo;
  char         blas_trans;
  char         blas_diag;
#endif
  FLA_Obj      A_copy;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() ) 
    FLA_Trsm_check( side, uplo, trans, diag, alpha, A, B );
#endif

  if ( FLA_Obj_has_zero_dim( B ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  ldim_A   = FLA_Obj_ldim( A );

  m_B      = FLA_Obj_length( B );
  n_B      = FLA_Obj_width( B );
  ldim_B   = FLA_Obj_ldim( B );

  if ( trans == FLA_CONJ_NO_TRANSPOSE )
  {
    FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, A, &A_copy );
    FLA_Copy_t( FLA_CONJ_NO_TRANSPOSE, A, A_copy );
    
    ldim_A = FLA_Obj_ldim( A_copy );
  }
  else 
  {
    A_copy = A;
  }

  FLA_Param_map_to_blas_side( side, &blas_side );
  FLA_Param_map_to_blas_uplo( uplo, &blas_uplo );
  FLA_Param_map_to_blas_trans( trans, &blas_trans );
  FLA_Param_map_to_blas_diag( diag, &blas_diag );


  switch( datatype ){

  case FLA_FLOAT:
  {
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_B     = ( float * ) FLA_FLOAT_PTR( B );
    float *buff_alpha = ( float * ) FLA_FLOAT_PTR( alpha );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_strsm( cblas_order,
                 blas_side,
                 blas_uplo,
                 blas_trans,
                 blas_diag,
                 m_B,
                 n_B,
                 *buff_alpha,
                 buff_A, ldim_A,
                 buff_B, ldim_B );
#else
    FLA_C2F( strsm )( &blas_side,
                      &blas_uplo,
                      &blas_trans,
                      &blas_diag,
                      &m_B,
                      &n_B,
                      buff_alpha,
                      buff_A, &ldim_A,
                      buff_B, &ldim_B );
#endif
    break;
  }

  case FLA_DOUBLE:
  {
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_B     = ( double * ) FLA_DOUBLE_PTR( B );
    double *buff_alpha = ( double * ) FLA_DOUBLE_PTR( alpha );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_dtrsm( cblas_order,
                 blas_side,
                 blas_uplo,
                 blas_trans,
                 blas_diag,
                 m_B,
                 n_B,
                 *buff_alpha,
                 buff_A, ldim_A,
                 buff_B, ldim_B );
#else
    FLA_C2F( dtrsm )( &blas_side,
                      &blas_uplo,
                      &blas_trans,
                      &blas_diag,
                      &m_B,
                      &n_B,
                      buff_alpha,
                      buff_A, &ldim_A,
                      buff_B, &ldim_B );
#endif
    break;
  }

  case FLA_COMPLEX:
  {
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A_copy );
    scomplex *buff_B     = ( scomplex * ) FLA_COMPLEX_PTR( B );
    scomplex *buff_alpha = ( scomplex * ) FLA_COMPLEX_PTR( alpha );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_ctrsm( cblas_order,
                 blas_side,
                 blas_uplo,
                 blas_trans,
                 blas_diag,
                 m_B,
                 n_B,
                 buff_alpha,
                 buff_A, ldim_A,
                 buff_B, ldim_B );
#else
    FLA_C2F( ctrsm )( &blas_side,
                      &blas_uplo,
                      &blas_trans,
                      &blas_diag,
                      &m_B,
                      &n_B,
                      buff_alpha,
                      buff_A, &ldim_A,
                      buff_B, &ldim_B );
#endif
    break;
  }

  case FLA_DOUBLE_COMPLEX:
  {
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A_copy );
    dcomplex *buff_B     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( B );
    dcomplex *buff_alpha = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( alpha );

#ifdef FLA_ENABLE_CBLAS_INTERFACE
    cblas_ztrsm( cblas_order,
                 blas_side,
                 blas_uplo,
                 blas_trans,
                 blas_diag,
                 m_B,
                 n_B,
                 buff_alpha,
                 buff_A, ldim_A,
                 buff_B, ldim_B );
#else
    FLA_C2F( ztrsm )( &blas_side,
                      &blas_uplo,
                      &blas_trans,
                      &blas_diag,
                      &m_B,
                      &n_B,
                      buff_alpha,
                      buff_A, &ldim_A,
                      buff_B, &ldim_B );
#endif
    break;
  }

  }
  
  if ( trans == FLA_CONJ_NO_TRANSPOSE )
    FLA_Obj_free( &A_copy );

  return FLA_SUCCESS;
}

void FLA_F2C( fla_trsm_external_f ) 
     ( F_INT *side, F_INT *uplo, F_INT *trans, F_INT *diag, F_INT *alpha, F_INT *A, F_INT *B, F_INT *IERROR )
{
  *IERROR = FLA_Trsm_external( *( ( FLA_Side  * ) side  ),
                               *( ( FLA_Uplo  * ) uplo  ),
                               *( ( FLA_Trans * ) trans ),
                               *( ( FLA_Diag  * ) diag  ),
                               *( ( FLA_Obj   * ) alpha ), 
                               *( ( FLA_Obj   * ) A     ),
                               *( ( FLA_Obj   * ) B     ) );
}

FLA_Error FLA_Trsm_llc_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_LEFT, FLA_LOWER_TRIANGULAR, FLA_CONJ_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_lln_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_LEFT, FLA_LOWER_TRIANGULAR, FLA_NO_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_llt_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_LEFT, FLA_LOWER_TRIANGULAR, FLA_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_luc_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_LEFT, FLA_UPPER_TRIANGULAR, FLA_CONJ_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_lun_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_LEFT, FLA_UPPER_TRIANGULAR, FLA_NO_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_lut_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_LEFT, FLA_UPPER_TRIANGULAR, FLA_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_rlc_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_RIGHT, FLA_LOWER_TRIANGULAR, FLA_CONJ_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_rln_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_RIGHT, FLA_LOWER_TRIANGULAR, FLA_NO_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_rlt_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_RIGHT, FLA_LOWER_TRIANGULAR, FLA_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_ruc_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_RIGHT, FLA_UPPER_TRIANGULAR, FLA_CONJ_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_run_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_RIGHT, FLA_UPPER_TRIANGULAR, FLA_NO_TRANSPOSE, diag, alpha, A, B );
}

FLA_Error FLA_Trsm_rut_external( FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  return FLA_Trsm_external( FLA_RIGHT, FLA_UPPER_TRIANGULAR, FLA_TRANSPOSE, diag, alpha, A, B );
}

