/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_Sylv_unb_external( FLA_Trans transa, FLA_Trans transb, FLA_Obj isgn, FLA_Obj A, FLA_Obj B, FLA_Obj C, FLA_Obj scale )
{
  FLA_Datatype datatype;
  int          ldim_A;
  int          ldim_B;
  int          m_C, n_C, ldim_C;
  int          info;
  char         blas_transa; 
  char         blas_transb;

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  if ( FLA_Check_parameters() )
    FLA_Sylv_check( transa, transb, isgn, A, B, C, scale );
#endif

  if ( FLA_Obj_has_zero_dim( A ) ) return FLA_SUCCESS;
  if ( FLA_Obj_has_zero_dim( B ) ) return FLA_SUCCESS;
  if ( FLA_Obj_has_zero_dim( C ) ) return FLA_SUCCESS;

  datatype = FLA_Obj_datatype( A );

  m_C      = FLA_Obj_length( C );
  n_C      = FLA_Obj_width( C );
  ldim_C   = FLA_Obj_ldim( C );

  ldim_A   = FLA_Obj_ldim( A );

  ldim_B   = FLA_Obj_ldim( B );

  FLA_Param_map_to_blas_trans( transa, &blas_transa );
  FLA_Param_map_to_blas_trans( transb, &blas_transb );


  switch( datatype ){

  case FLA_FLOAT:
  {
    int   *buff_isgn  = ( int   * ) FLA_INT_PTR( isgn );
    float *buff_A     = ( float * ) FLA_FLOAT_PTR( A );
    float *buff_B     = ( float * ) FLA_FLOAT_PTR( B );
    float *buff_C     = ( float * ) FLA_FLOAT_PTR( C );
    float *buff_scale = ( float * ) FLA_FLOAT_PTR( scale );

    FLA_C2F( strsyl )( &blas_transa,
                       &blas_transb,
                       buff_isgn,
                       &m_C,
                       &n_C,
                       buff_A, &ldim_A,
                       buff_B, &ldim_B,
                       buff_C, &ldim_C,
                       buff_scale,
                       &info );

    break;
  }

  case FLA_DOUBLE:
  {
    int    *buff_isgn  = ( int    * ) FLA_INT_PTR( isgn );
    double *buff_A     = ( double * ) FLA_DOUBLE_PTR( A );
    double *buff_B     = ( double * ) FLA_DOUBLE_PTR( B );
    double *buff_C     = ( double * ) FLA_DOUBLE_PTR( C );
    double *buff_scale = ( double * ) FLA_DOUBLE_PTR( scale );

    FLA_C2F( dtrsyl )( &blas_transa,
                       &blas_transb,
                       buff_isgn,
                       &m_C,
                       &n_C,
                       buff_A, &ldim_A,
                       buff_B, &ldim_B,
                       buff_C, &ldim_C,
                       buff_scale,
                       &info );

    break;
  } 

  case FLA_COMPLEX:
  {
    int      *buff_isgn  = ( int      * ) FLA_INT_PTR( isgn );
    scomplex *buff_A     = ( scomplex * ) FLA_COMPLEX_PTR( A );
    scomplex *buff_B     = ( scomplex * ) FLA_COMPLEX_PTR( B );
    scomplex *buff_C     = ( scomplex * ) FLA_COMPLEX_PTR( C );
    float    *buff_scale = ( float    * ) FLA_COMPLEX_PTR( scale );

    FLA_C2F( ctrsyl )( &blas_transa,
                       &blas_transb,
                       buff_isgn,
                       &m_C,
                       &n_C,
                       buff_A, &ldim_A,
                       buff_B, &ldim_B,
                       buff_C, &ldim_C,
                       buff_scale,
                       &info );

    break;
  } 

  case FLA_DOUBLE_COMPLEX:
  {
    int      *buff_isgn  = ( int      * ) FLA_INT_PTR( isgn );
    dcomplex *buff_A     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( A );
    dcomplex *buff_B     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( B );
    dcomplex *buff_C     = ( dcomplex * ) FLA_DOUBLE_COMPLEX_PTR( C );
    double   *buff_scale = ( double   * ) FLA_DOUBLE_COMPLEX_PTR( scale );

    FLA_C2F( ztrsyl )( &blas_transa,
                       &blas_transb,
                       buff_isgn,
                       &m_C,
                       &n_C,
                       buff_A, &ldim_A,
                       buff_B, &ldim_B,
                       buff_C, &ldim_C,
                       buff_scale,
                       &info );

    break;
  } 

  }

  // We don't provide a comprehensive strategy for handing scaling to avoid
  // overflow, so we just force the scale argument to 1.0.
  FLA_Obj_set_to_scalar( scale, FLA_ONE );

  return info;
}

void FLA_F2C( fla_sylv_unb_external_f )
     ( F_INT *transa, F_INT *transb, F_INT *isgn, F_INT *A, F_INT *B, F_INT *C, F_INT *scale, F_INT *IERROR )
{
  *IERROR = FLA_Sylv_unb_external( *( ( FLA_Trans * ) transa ), 
                                   *( ( FLA_Trans * ) transb ), 
                                   *( ( FLA_Obj   * ) isgn   ), 
                                   *( ( FLA_Obj   * ) A      ), 
                                   *( ( FLA_Obj   * ) B      ), 
                                   *( ( FLA_Obj   * ) C      ), 
                                   *( ( FLA_Obj   * ) scale  ) );
}

