/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Copy_x_check( PLA_Trans trans,
			    PLA_Obj   A,
			    PLA_Obj   B )
/*----------------------------------------------------------------------------

Purpose :  Check sqrt

IN      trans      option to transpose
IN      A          object to be copied
IN      B          object into which to copy

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      if ( PLA_Obj_objtype( A ) == PLA_CONSTANT )
	 PLA_Abort( "PLA_Copy: cannot copy from a constant", __LINE__, __FILE__ );

      if ( PLA_Obj_objtype( B ) == PLA_CONSTANT )
	 PLA_Abort( "PLA_Copy: cannot copy into a constant", __LINE__, __FILE__ );
      
      PLA_Check_valid_trans( trans, "PLA_Copy: invalid trans" );
      PLA_Check_consistent_datatype( PLA_Obj_datatype( A ), B, "PLA_Copy: inconsistent datatypes" );
      PLA_Check_consistent_dims( trans, A, B, "PLA_Copy: object sizes do not conform" );
      
      PLA_Check_consistent_templates( A, B, "PLA_Copy: different templates not yet implemented" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Reduce_x_check( PLA_Op    op,
			      PLA_Trans trans,
			      PLA_Obj   alpha,
			      PLA_Obj   A,
			      PLA_Obj   B )
/*----------------------------------------------------------------------------

Purpose :  Reduce contents using summation in the duplicated object
           given by A and overwrite B with the result

IN     op       reduction operation
IN     trans    option for transpose
IN     alpha    scaling factor for target object
IN     A        object to be reduced
IN     B        object into which to reduce

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      if ( PLA_Obj_objtype( A ) != PLA_MSCALAR     &&
	   PLA_Obj_objtype( A ) != PLA_PMROWVECTOR && 
	   PLA_Obj_objtype( A ) != PLA_PMCOLVECTOR )
	 PLA_Abort( "PLA_Reduce: cannot reduce an unduplicated object", __LINE__, __FILE__ );

      if ( PLA_Obj_owner_row( A ) != PLA_ALL_ROWS  &&
	   PLA_Obj_owner_col( A ) != PLA_ALL_COLS )
	 PLA_Abort( "PLA_Reduce: cannot reduce an unduplicated object", __LINE__, __FILE__ );
      
      if ( PLA_Obj_objtype( A ) == PLA_CONSTANT )
	 PLA_Abort( "PLA_Reduce: cannot reduce from a constant", __LINE__, __FILE__ );
      
      if ( PLA_Obj_objtype( B ) == PLA_CONSTANT )
	 PLA_Abort( "PLA_Reduce: cannot reduce into a constant", __LINE__, __FILE__ );
      
      PLA_Check_valid_op( op, "PLA_Reduce: invalid op" );
      PLA_Check_if_scalar( alpha, "PLA_Reduce: alpha must be 1x1" );
      
      PLA_Check_valid_trans( trans, "PLA_Reduce: invalid trans" );
      PLA_Check_consistent_datatype( PLA_Obj_datatype( A ), B, "PLA_Reduce: inconsistent datatypes" );
      PLA_Check_consistent_dims( trans, A, B, "PLA_Reduce: object sizes do not conform" );
      
      PLA_Check_consistent_templates( A, B, "PLA_Reduce: different templates not yet implemented" );     
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Sqrt_check( PLA_Obj alpha )
/*----------------------------------------------------------------------------

Purpose :  Check sqrt

IN      alpha      object must be a scalar

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_if_scalar( alpha, "PLA_Sqrt: alpha must be 1x1" );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Invert_check( PLA_Obj alpha )
/*----------------------------------------------------------------------------

Purpose :  Check invert

IN      alpha      object must be a scalar

----------------------------------------------------------------------------*/
{
   if ( PLA_Check_parameters( ) )
   {
      PLA_Check_if_scalar( alpha, "PLA_Invert: alpha must be 1x1" );
   }
   
   return PLA_SUCCESS;
}
