/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
static PLA_Bool PLA_check_parameters = TRUE;
static PLA_Bool PLA_error_check      = TRUE;
#else
static PLA_Bool PLA_check_parameters = FALSE;
static PLA_Bool PLA_error_check      = FALSE;
#endif


PLA_Bool PLA_Error_check( )
/*----------------------------------------------------------------------------

Purpose :   Extract error checking

Return :    if check errors

----------------------------------------------------------------------------*/
{
   return PLA_error_check;
}


void PLA_Error_check_set( PLA_Bool val )
/*----------------------------------------------------------------------------

Purpose :   Set error checking

IN      val        new value

----------------------------------------------------------------------------*/
{
   PLA_error_check = val;
   
   return;
}


PLA_Bool PLA_Check_parameters( )
/*----------------------------------------------------------------------------

Purpose :   Extract parameter checking

Return :    if check parameters

----------------------------------------------------------------------------*/
{
   return PLA_check_parameters;
}


void PLA_Check_parameters_set( PLA_Bool val )
/*----------------------------------------------------------------------------

Purpose :   Set parameter checking

IN      val        new value

----------------------------------------------------------------------------*/
{
   PLA_check_parameters = val;
   
   return;
}


void PLA_Check_valid_side( PLA_Side side, 
			   char     *message )
/*----------------------------------------------------------------------------

Purpose :   Check side option

IN      side       side option
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( side != PLA_TOP    && 
	side != PLA_BOTTOM &&
	side != PLA_LEFT   && 
	side != PLA_RIGHT )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_valid_quadrant( PLA_Quadrant quadrant, 
			       char         *message )
/*----------------------------------------------------------------------------

Purpose :   Check quadrant option

IN      quadrant   quadrant option
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( quadrant != PLA_TL &&
	quadrant != PLA_TR &&
	quadrant != PLA_BL &&
	quadrant != PLA_BR )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_valid_uplo( PLA_Uplo uplo, 
			   char     *message )
/*----------------------------------------------------------------------------

Purpose :   Check uplo option

IN      uplo       uplo option
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( uplo != PLA_LOWER_TRIANGULAR && 
	uplo != PLA_UPPER_TRIANGULAR )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_valid_trans( PLA_Trans trans, 
			    char      *message )
/*----------------------------------------------------------------------------

Purpose :   Check transpose option

IN      trans      transpose option
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( trans != PLA_NO_TRANSPOSE && 
	trans != PLA_TRANSPOSE    &&
	trans != PLA_CONJ_TRANSPOSE )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_valid_diag( PLA_Diag diag, 
			   char     *message )
/*----------------------------------------------------------------------------

Purpose :   Check diagonal option

IN      diag       diagonal option
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( diag != PLA_UNIT_DIAG    &&
	diag != PLA_NONUNIT_DIAG && 
	diag != PLA_ZERO_DIAG )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_valid_conj( PLA_Conj conj,
			   char     *message )
/*----------------------------------------------------------------------------

Purpose :   Check conjugate option

IN      conj       conjugate option
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( conj != PLA_NO_CONJUGATE &&
	conj != PLA_CONJUGATE )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_valid_datatype( PLA_Datatype datatype, 
			       char         *message )
/*----------------------------------------------------------------------------

Purpose :   Check datatype

IN      datatype   data type
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( datatype != PLA_FLOAT          &&
	datatype != PLA_DOUBLE         &&
	datatype != PLA_COMPLEX        &&
	datatype != PLA_DOUBLE_COMPLEX &&
	datatype != PLA_INT            &&
	datatype != PLA_CONSTANT )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_real_object( PLA_Obj A,
			    char    *message )
/*----------------------------------------------------------------------------

Purpose :   Check if real object

IN      A          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   PLA_Datatype datatype = PLA_Obj_datatype( A );
   
   if ( datatype != PLA_CONSTANT &&
        datatype != PLA_FLOAT    &&
	datatype != PLA_DOUBLE )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_complex_object( PLA_Obj A,
			       char    *message )
/*----------------------------------------------------------------------------

Purpose :   Check if complex object

IN      A          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   PLA_Datatype datatype = PLA_Obj_datatype( A );
   
   if ( datatype != PLA_CONSTANT &&
	datatype != PLA_COMPLEX  &&
	datatype != PLA_DOUBLE_COMPLEX )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_if_square( PLA_Obj A,
			  char    *message )
/*----------------------------------------------------------------------------

Purpose :   Check to see if object is square

IN      A          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_length( A ) != PLA_Obj_width( A ) )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_if_scalar( PLA_Obj alpha,
			  char    *message )
/*----------------------------------------------------------------------------

Purpose :   Check to see if object is a scalar

IN      alpha      object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_length( alpha ) != 1 ||
	PLA_Obj_width ( alpha ) != 1 )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_if_vector( PLA_Obj A,
			  char    *message )
/*----------------------------------------------------------------------------

Purpose :   Check to see if object is a vector

IN      A          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_length( A ) != 1 &&
	PLA_Obj_width ( A ) != 1 )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_consistent_datatype( PLA_Datatype datatype, 
				    PLA_Obj      A, 
				    char         *message )
/*----------------------------------------------------------------------------

Purpose :   Check to see if datatypes are consistent

IN      datatype   data type to check
IN      A          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_datatype( A ) != PLA_CONSTANT && 
	datatype              != PLA_CONSTANT )
   {
      if ( PLA_Obj_datatype( A ) != datatype )
	 PLA_Abort( message, __LINE__, __FILE__ );
   }
   
   return;
}


void PLA_Check_consistent_dims( PLA_Trans trans, 
				PLA_Obj   A, 
				PLA_Obj   B, 
				char      *message )
/*----------------------------------------------------------------------------

Purpose :   Check to see if dimensions are consistent

IN      trans      option for transpose
IN      A          object
IN      B          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( trans == PLA_NO_TRANSPOSE )
   {
      if ( PLA_Obj_length( A ) != PLA_Obj_length( B ) )
	 PLA_Abort( message, __LINE__, __FILE__ );
      
      if ( PLA_Obj_width ( A ) != PLA_Obj_width ( B ) )
	 PLA_Abort( message, __LINE__, __FILE__ );
   }
   else
   {
      if ( PLA_Obj_width ( A ) != PLA_Obj_length( B ) )
	 PLA_Abort( message, __LINE__, __FILE__ );
      
      if ( PLA_Obj_length( A ) != PLA_Obj_width ( B ) )
	 PLA_Abort( message, __LINE__, __FILE__ );
   }
}


void PLA_Check_valid_objtype( PLA_Objtype objtype,
			      char        *message )
/*----------------------------------------------------------------------------

Purpose :   Check object type

IN      objtype    object type
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( objtype != PLA_MATRIX      &&
	objtype != PLA_MROWVECTOR  &&
	objtype != PLA_MCOLVECTOR  &&
	objtype != PLA_MSCALAR     &&
	objtype != PLA_PMROWVECTOR &&
	objtype != PLA_PMCOLVECTOR &&
	objtype != PLA_CONSTANT )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_valid_op( PLA_Op op,
			 char   *message )
/*----------------------------------------------------------------------------

Purpose :   Check operation

IN      op         operation
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( op != PLA_SUM  &&
	op != PLA_PROD &&
	op != PLA_MAX  &&
	op != PLA_MIN )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}


void PLA_Check_if_local_object( PLA_Obj A,
				char    *message )
/*----------------------------------------------------------------------------

Purpose :   Check to see if object is local

IN      A          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_is_owner( A ) )
   {
      if ( PLA_Obj_length( A ) != PLA_Obj_local_length( A ) ||
	   PLA_Obj_width ( A ) != PLA_Obj_local_width ( A ) )
	 PLA_Abort( message, __LINE__, __FILE__ );
   }
   
   return;
}


void PLA_Check_consistent_owners( PLA_Obj A,
				  PLA_Obj B,
				  char    *message )
/*----------------------------------------------------------------------------

Purpose :   Check to see if objects owned by same nodes

IN      A          object
IN      B          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_is_owner( B ) )
   {
      if ( PLA_Obj_owner_all( A ) != PLA_Obj_owner_all( B ) )
	 PLA_Abort( message, __LINE__, __FILE__ );
   }
   
   return;
}


void PLA_Check_consistent_templates( PLA_Obj A,
				     PLA_Obj B,
				     char    *message )
/*----------------------------------------------------------------------------

Purpose :   Check to see if templates are consistent

IN      A          object
IN      B          object
IN      message    string for error message

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_template( A ) != PLA_Obj_template( B ) )
      PLA_Abort( message, __LINE__, __FILE__ );
   
   return;
}
