/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


int PLA_Obj_compute_local_ldim( PLA_Obj );


PLA_Error PLA_Obj_create( PLA_Objtype  objtype,
			  PLA_Datatype datatype,
			  int          m,
			  int          n,
			  int          alignM,
			  int          alignN,
			  int          ownerRow,
			  int          ownerCol,
			  PLA_Temp     templ,
			  PLA_Obj      *obj )
/*----------------------------------------------------------------------------

Purpose :   Helper function to create an object

IN      objtype      type of object
IN      datatype     datatype
IN      m            global length
IN      n            global width
IN      alignM       alignment to template
IN      alignN       alignment to template
IN      ownerRow     owner of the row
IN      ownerCol     owner of the column
IN      templ        template
OUT     obj          object created

----------------------------------------------------------------------------*/
{
   int
      localm, localn,
      typesize;
   
   if ( PLA_Error_check( ) )
      PLA_Obj_create_check( objtype, 
			    datatype, 
			    m, 
			    n, 
			    alignM, 
			    alignN, 
			    ownerRow, 
			    ownerCol, 
			    templ );
   
   obj->offm = 0;
   obj->offn = 0;
   obj->m    = m;
   obj->n    = n;
   obj->base = (PLA_Base_obj *) PLA_malloc( sizeof( PLA_Base_obj ) );
   
   obj->base->datatype = datatype;
   obj->base->m = m;
   obj->base->n = n;
   
   obj->base->objtype  = objtype;
   obj->base->alignM   = alignM;
   obj->base->alignN   = alignN;
   obj->base->ownerRow = ownerRow;
   obj->base->ownerCol = ownerCol;
   
   obj->base->templ = templ;
   
   obj->base->offset = -1; // undefined 
   obj->base->fp     = -1; // undefined 
   
   obj->base->ldim = PLA_Obj_compute_local_ldim( *obj );
   
   localm = PLA_Obj_local_length( *obj );
   localn = PLA_Obj_local_width( *obj );
   typesize = PLA_Type_size( datatype );
   
   if ( localm * localn != 0 ) 
   {
      obj->base->buffer = (void *) PLA_malloc( typesize * localm * localn );
      
      if ( objtype != PLA_CONSTANT )
	 PLA_Obj_set_to_zero( *obj );
   }
   else
      obj->base->buffer = NULL;
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Matrix_create( PLA_Datatype datatype,
			     int          m,
			     int          n,		
			     int          alignM,     
			     int          alignN,
			     PLA_Temp     templ,
			     PLA_Obj      *obj )
/*----------------------------------------------------------------------------

Purpose :   Create a distributed matrix

IN      datatype     datatype
IN      m            global length
IN      n            global width
IN      alignM       alignment to template
IN      alignN       alignment to template
IN      templ        template
OUT     obj          object describing created matrix

----------------------------------------------------------------------------*/
{
   PLA_Obj_create( PLA_MATRIX,
		   datatype,
		   m,
		   n,
		   alignM,
		   alignN,
		   PLA_ALL_ROWS, // undefined
		   PLA_ALL_COLS, // undefined
		   templ,
		   obj );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Mrowvector_create( PLA_Datatype datatype,
				 int          m, 
				 int          n,
				 int          alignN,
				 PLA_Temp     templ,
				 PLA_Obj      *obj )
/*----------------------------------------------------------------------------

Purpose :   Create a distributed multi-row-vector

IN      datatype     datatype
IN      m            global length
IN      n            global width
IN      alignN       alignment to template
IN      templ        template
OUT     obj          object describing created multi-row-vector

----------------------------------------------------------------------------*/
{
   PLA_Obj_create( PLA_MROWVECTOR,
		   datatype,
		   m,
		   n,
		   0,            // undefined
		   alignN,
		   PLA_ALL_ROWS, // undefined
		   PLA_ALL_COLS, // undefined
		   templ,
		   obj );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Mcolvector_create( PLA_Datatype datatype,
				 int          m,
				 int          n,
				 int          alignM,
				 PLA_Temp     templ,
				 PLA_Obj      *obj )
/*----------------------------------------------------------------------------

Purpose :   Create a distributed multi-column-vector

IN      datatype     datatype
IN      m            global length
IN      n            global width
IN      alignM       alignment to template
IN      templ        template
OUT     obj          object describing created multi-column-vector

----------------------------------------------------------------------------*/
{
   PLA_Obj_create( PLA_MCOLVECTOR,
		   datatype,
		   m,
		   n,
		   alignM,
		   0,            // undefined
		   PLA_ALL_ROWS, // undefined
		   PLA_ALL_COLS, // undefined
		   templ,
		   obj );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Mscalar_create( PLA_Datatype datatype,
			      int          m,
			      int          n,
			      int          ownerRow,
			      int          ownerCol,
			      PLA_Temp     templ,
			      PLA_Obj      *obj )
/*----------------------------------------------------------------------------

Purpose :   Create a distributed multi-scalar

IN      datatype     datatype
IN      m            global length
IN      n            global width
IN      ownerRow     owner of the row
IN      ownerCol     owner of the column
IN      templ        template
OUT     obj          object describing created multi-scalar

----------------------------------------------------------------------------*/
{
   PLA_Obj_create( PLA_MSCALAR,
		   datatype,
		   m,
		   n,
		   0, // undefined
		   0, // undefined
		   ownerRow,
		   ownerCol,
		   templ,
		   obj );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Pmrowvector_create( PLA_Datatype datatype,
				  int          m,
				  int          n,
				  int          alignN,
				  int          ownerRow,
				  PLA_Temp     templ,
				  PLA_Obj      *obj )
/*----------------------------------------------------------------------------

Purpose :   Create a distributed projected multi-row-vector

IN      datatype     datatype
IN      m            global length
IN      n            global width
IN      alignN       alignment to template
IN      ownerRow     owner of the row
IN      templ        template
OUT     obj          object describing created projected multi-row-vector

----------------------------------------------------------------------------*/
{
   PLA_Obj_create( PLA_PMROWVECTOR,
		   datatype,
		   m,
		   n,
		   0,            // undefined
		   alignN,
		   ownerRow,
		   PLA_ALL_COLS, // undefined
		   templ,
		   obj );  
   
   return PLA_SUCCESS;
}

  
PLA_Error PLA_Pmcolvector_create( PLA_Datatype datatype,
				  int          m,
				  int          n,
				  int          alignM,
				  int          ownerCol,
				  PLA_Temp     templ,
				  PLA_Obj      *obj )
/*----------------------------------------------------------------------------

Purpose :   Create a distributed projected multi-column-vector

IN      datatype     datatype
IN      m            global length
IN      n            global width
IN      alignM       alignment to template
IN      ownerCol     owner of the column
IN      templ        template
OUT     obj          object describing created projected multi-column-vector

----------------------------------------------------------------------------*/
{
   PLA_Obj_create( PLA_PMCOLVECTOR,
		   datatype,
		   m,
		   n,
		   alignM,
		   0,            // undefined
		   PLA_ALL_ROWS, // undefined
		   ownerCol,
		   templ,
		   obj );  
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Constant_complex_create( double  real,
				       double  imag,
				       PLA_Obj *obj )
/*----------------------------------------------------------------------------

Purpose :   Create a complex constant

IN      real         real value of constant
IN      imag         imaginary value of constant
OUT     obj          object describing created matrix

----------------------------------------------------------------------------*/
{
   float    *temp_s;
   double   *temp_d;
   scomplex *temp_c;
   dcomplex *temp_z;
   int      *temp_i;
   
   PLA_Obj_create( PLA_CONSTANT,
		   PLA_CONSTANT,
		   1,
		   1,
		   0, // undefined
		   0, // undefined
		   PLA_ALL_ROWS,
		   PLA_ALL_COLS,
		   PLA_TEMP_WORLD,
		   obj );
  
   temp_s = PLA_FLOAT_PTR( *obj );
   *temp_s = (float) real;
   
   temp_d = PLA_DOUBLE_PTR( *obj );
   *temp_d = real;
   
   temp_c = PLA_COMPLEX_PTR( *obj );
   temp_c->real = (float) real;
   temp_c->imag = (float) imag;
   
   temp_z = PLA_DOUBLE_COMPLEX_PTR( *obj );
   temp_z->real = real;
   temp_z->imag = imag;
   
   temp_i = PLA_INT_PTR( *obj );
   *temp_i = (int) real;
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Constant_create( double  real,
			       PLA_Obj *obj )
/*----------------------------------------------------------------------------

Purpose :   Create a constant

IN      real         value of constant
OUT     obj          object describing created matrix

----------------------------------------------------------------------------*/
{
   double imag = 0.0;
   
   PLA_Constant_complex_create( real, imag, obj );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Matrix_create_conf_to( PLA_Trans trans,
				     PLA_Obj   obj,
				     PLA_Obj   *newObj )
/*----------------------------------------------------------------------------

Purpose :   Create matrix conformal to given object (must be matrix)

IN      trans        option to transpose object
IN      obj          original object
OUT     newObj      created object

----------------------------------------------------------------------------*/
{
   if ( trans == PLA_NO_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Matrix_create( obj.base->datatype,
			       obj.m,
			       obj.n,		
			       obj.base->alignM + obj.offm,
			       obj.base->alignN + obj.offn,
			       obj.base->templ,
			       newObj );      
	    break;
	 default:
	    PLA_Abort( "Cannot create matrix conformal from anything but matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else if ( trans == PLA_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Matrix_create( obj.base->datatype,
			       obj.n,
			       obj.m,		
			       obj.base->alignN + obj.offn,
			       obj.base->alignM + obj.offm,
			       obj.base->templ,
			       newObj );      
	    break;
	 default:
	    PLA_Abort( "Cannot create matrix conformal from anything but matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else
      PLA_Abort( "Unknown value for transposition", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Mrowvector_create_conf_to( PLA_Trans trans,
					 PLA_Obj   obj,
					 int       numVects,
					 PLA_Obj   *newObj )
/*----------------------------------------------------------------------------

Purpose :   Create multi-row-vector conformal to given object

IN      trans        option to transpose object
IN      obj          original object
IN      numVects     number of vectors in multi-row-vector
OUT     newObj      created object

----------------------------------------------------------------------------*/
{
   if ( trans == PLA_NO_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Mrowvector_create( obj.base->datatype,
				   ( numVects == PLA_INHERIT ? 
				     obj.m : numVects ),
				   obj.n,
				   obj.base->alignN + obj.offn,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_MROWVECTOR:
	    PLA_Mrowvector_create( obj.base->datatype,
				   ( numVects == PLA_INHERIT ? 
				     obj.m : numVects ),
				   obj.n,
				   obj.base->alignN + obj.offn,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Abort( "Must transpose multi-column-vectors to make multi-row-vector conformal", __LINE__, __FILE__ );
	    break;
	 case PLA_PMROWVECTOR:
	    PLA_Mrowvector_create( obj.base->datatype,
				   ( numVects == PLA_INHERIT ? 
				     obj.m : numVects ),
				   obj.n,
				   obj.base->alignN + obj.offn,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Abort( "Must transpose projected multi-column-vectors to make multi-row-vector conformal", __LINE__, __FILE__ );
	    break;
	 default:
	    PLA_Abort( "Cannot create multi-row-vectors conformal to anything but multivectors or matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else if ( trans == PLA_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Mrowvector_create( obj.base->datatype,
				   ( numVects == PLA_INHERIT ? 
				     obj.n : numVects ),
				   obj.m,
				   obj.base->alignM + obj.offm,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_MROWVECTOR:
	    PLA_Abort( "Cannot transpose multi-row-vectors to another multi-row-vector", __LINE__, __FILE__ );
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Mrowvector_create( obj.base->datatype,
				   ( numVects == PLA_INHERIT ? 
				     obj.n : numVects ),
				   obj.m,  // switch order of m and n
				   obj.base->alignM + obj.offm,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_PMROWVECTOR:
	    PLA_Abort( "Cannot transpose projected multi-row-vectors to another multi-row-vector", __LINE__, __FILE__ );
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Mrowvector_create( obj.base->datatype,
				   ( numVects == PLA_INHERIT ? 
				     obj.n : numVects ),
				   obj.m,  // switch order of m and n
				   obj.base->alignM + obj.offm,
				   obj.base->templ,
				   newObj );
	    break;
	 default:
	    PLA_Abort( "Cannot create multi-row-vectors conformal to anything but multivectors and matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else
      PLA_Abort( "Unknown value for transposition", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Mcolvector_create_conf_to( PLA_Trans trans,
					 PLA_Obj   obj,
					 int       numVects,
					 PLA_Obj   *newObj )
/*----------------------------------------------------------------------------

Purpose :   Create multi-column-vector conformal to given object

IN      trans        option to transpose object
IN      obj          original object
IN      numVects     number of vectors in multi-column-vector
OUT     newObj      created object

----------------------------------------------------------------------------*/
{
   if ( trans == PLA_NO_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Mcolvector_create( obj.base->datatype,
				   obj.m,  
				   ( numVects == PLA_INHERIT ? 
				     obj.n : numVects ),
				   obj.base->alignM + obj.offm,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_MROWVECTOR:
	    PLA_Abort( "Must transpose multi-row-vectors to make multi-column-vector conformal", __LINE__, __FILE__ );	  
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Mcolvector_create( obj.base->datatype,
				   obj.m,  
				   ( numVects == PLA_INHERIT ? 
				     obj.n : numVects ),
				   obj.base->alignM + obj.offm,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_PMROWVECTOR:
	    PLA_Abort( "Must transpose projected multi-column-vectors to make multi-row-vector conformal", __LINE__, __FILE__ );
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Mcolvector_create( obj.base->datatype,
				   obj.m,  
				   ( numVects == PLA_INHERIT ? 
				     obj.n : numVects ),
				   obj.base->alignM + obj.offm,
				   obj.base->templ,
				   newObj );
	    break;
	 default:
	    PLA_Abort( "Cannot create multi-column-vectors conformal to anything but multivectors and matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else if ( trans == PLA_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Mcolvector_create( obj.base->datatype,
				   obj.n,  
				   ( numVects == PLA_INHERIT ? 
				     obj.m : numVects ),
				   obj.base->alignN + obj.offn,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_MROWVECTOR:
	    PLA_Mcolvector_create( obj.base->datatype,
				   obj.n,  // switch order of m and n
				   ( numVects == PLA_INHERIT ? 
				     obj.m : numVects ),
				   obj.base->alignN + obj.offn,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Abort( "Cannot transpose multi-column-vectors to another multi-column-vector", __LINE__, __FILE__ );
	    break;
	 case PLA_PMROWVECTOR:
	    PLA_Mcolvector_create( obj.base->datatype,
				   obj.n,  // switch order of m and n
				   ( numVects == PLA_INHERIT ? 
				     obj.m : numVects ),
				   obj.base->alignN + obj.offn,
				   obj.base->templ,
				   newObj );
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Abort( "Cannot transpose projected multi-column-vectors to another multi-column-vector", __LINE__, __FILE__ );
	    break;
	 default:
	    PLA_Abort( "Cannot create multi-column-vectors conformal to anything but multivectors and matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else
      PLA_Abort( "Unknown value for transposition", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Mscalar_create_conf_to( PLA_Trans trans,
				      PLA_Obj   obj,
				      int       ownerRow,
				      int       ownerCol,
				      PLA_Obj   *newObj )
/*----------------------------------------------------------------------------

Purpose :   Create multi-column-vector conformal to given object

IN      trans        option to transpose object
IN      obj          original object
IN      ownerRow     index of row of nodes which contains owning node(s)
IN      ownerCol     index of column of nodes which contains owning node(s)
OUT     newObj      created object

----------------------------------------------------------------------------*/
{
   if ( trans == PLA_NO_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.m,
				obj.n,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );            
	    break;
	 case PLA_MROWVECTOR:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.m,
				obj.n,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );            
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.m,
				obj.n,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ?
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );      
	    break;
	 case PLA_MSCALAR:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.m,
				obj.n,
				( ownerRow == PLA_INHERIT ? 
				  obj.base->ownerRow : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  obj.base->ownerCol : ownerCol ),
				obj.base->templ,
				newObj );      
	    break;
	 case PLA_PMROWVECTOR:     
	    PLA_Mscalar_create( obj.base->datatype,
				obj.m,
				obj.n,
				( ownerRow == PLA_INHERIT ? 
				  obj.base->ownerRow : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );      
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.m,
				obj.n,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  obj.base->ownerCol : ownerCol ),
				obj.base->templ,
				newObj );      
	    break;     
	 case PLA_CONSTANT:
	    PLA_Abort( "Cannot create multiscalar conformal to a constant", __LINE__, __FILE__ );
	    break;
	 default:
	    PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
	    break;
      }
   }
   else if ( trans == PLA_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.n,
				obj.m,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );            
	    break;
	 case PLA_MROWVECTOR:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.n,
				obj.m,
				( ownerRow == PLA_INHERIT ?
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );            
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.n,
				obj.m,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );      
	    break;
	 case PLA_MSCALAR:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.n,
				obj.m,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );      
	    break;
	 case PLA_PMROWVECTOR:     
	    PLA_Mscalar_create( obj.base->datatype,
				obj.n,
				obj.m,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );      
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Mscalar_create( obj.base->datatype,
				obj.n,
				obj.m,
				( ownerRow == PLA_INHERIT ? 
				  PLA_ALL_ROWS : ownerRow ),
				( ownerCol == PLA_INHERIT ? 
				  PLA_ALL_COLS : ownerCol ),
				obj.base->templ,
				newObj );      
	    break;     
	 case PLA_CONSTANT:
	    PLA_Abort( "Cannot create multiscalar conformal to a constant", __LINE__, __FILE__ );
	    break;
	 default:
	    PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
	    break;
      }
   }
   else
      PLA_Abort( "Unknown value for transposition", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Pmrowvector_create_conf_to( PLA_Trans trans,
					  PLA_Obj   obj,
					  int       numVects,
					  int       ownerRow,
					  PLA_Obj   *newObj )
/*----------------------------------------------------------------------------

Purpose :   Create projected multi-row-vector conformal to given object

IN      trans        option to transpose object
IN      obj          original object
IN      numVects     number of vectors in projected multi-row-vector
IN      ownerRow     index of row of nodes which contains owning node(s)
OUT     newObj      created object

----------------------------------------------------------------------------*/
{
   if ( trans == PLA_NO_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Pmrowvector_create( obj.base->datatype,
				    ( numVects == PLA_INHERIT ? 
				      obj.m : numVects ),
				    obj.n,
				    obj.base->alignN + obj.offn,
				    ( ownerRow == PLA_INHERIT ? 
				      PLA_ALL_ROWS : ownerRow ),  
				    obj.base->templ,
				    newObj );
	    break;
	 case PLA_MROWVECTOR:
	    PLA_Pmrowvector_create( obj.base->datatype,
				    ( numVects == PLA_INHERIT ? 
				      obj.m : numVects ),
				    obj.n,
				    obj.base->alignN + obj.offn,
				    ( ownerRow == PLA_INHERIT ? 
				      PLA_ALL_ROWS : ownerRow ),  
				    obj.base->templ,
				    newObj );
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Abort( "Must transpose multi-column-vectors to make multi-row-vector conformal", __LINE__, __FILE__ );	  
	    break;
	 case PLA_PMROWVECTOR:
	    PLA_Pmrowvector_create( obj.base->datatype,
				    ( numVects == PLA_INHERIT ? 
				      obj.m : numVects ),
				    obj.n,
				    obj.base->alignN + obj.offn,
				    ( ownerRow == PLA_INHERIT ? 
				      obj.base->ownerRow : ownerRow ),  
				    obj.base->templ,
				    newObj );
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Abort( "Must transpose projected multi-column-vectors to make multi-row-vector conformal", __LINE__, __FILE__ );	  
	    break;
	 default:
	    PLA_Abort( "Cannot create projected multi-row-vectors conformal to anything but multivectors and matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else if ( trans == PLA_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Pmrowvector_create( obj.base->datatype,
				    ( numVects == PLA_INHERIT ? 
				      obj.n : numVects ),
				    obj.m,
				    obj.base->alignM + obj.offm,
				    ( ownerRow == PLA_INHERIT ? 
				      PLA_ALL_ROWS : ownerRow ),  
				    obj.base->templ,
				    newObj );
	    break;
	 case PLA_MROWVECTOR:
	    PLA_Abort( "Cannot transpose projected multi-row-vectors to another multi-row-vector", __LINE__, __FILE__ );
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Pmrowvector_create( obj.base->datatype,
				    ( numVects == PLA_INHERIT ? 
				      obj.n : numVects ),
				    obj.m, // switch order of m and n
				    obj.base->alignM + obj.offm,
				    ( ownerRow == PLA_INHERIT ? 
				      PLA_ALL_ROWS : ownerRow ),  
				    obj.base->templ,
				    newObj );
	    break;
	 case PLA_PMROWVECTOR:
	    PLA_Abort( "Cannot transpose projected multi-row-vectors to another multi-row-vector", __LINE__, __FILE__ );
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Pmrowvector_create( obj.base->datatype,
				    ( numVects == PLA_INHERIT ? 
				      obj.n : numVects ),
				    obj.m, // switch order of m and n
				    obj.base->alignM + obj.offm,
				    ( ownerRow == PLA_INHERIT ? 
				      PLA_ALL_ROWS : ownerRow ),  
				    obj.base->templ,
				    newObj );
	    break;
	 default:
	    PLA_Abort( "Cannot create projected multi-row-vectors conformal to anything but multivectors and matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else
      PLA_Abort( "Unknown value for transposition", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Pmcolvector_create_conf_to( PLA_Trans trans,
					  PLA_Obj   obj,
					  int       numVects,
					  int       ownerCol,
					  PLA_Obj   *newObj )
/*----------------------------------------------------------------------------

Purpose :   Create projected multi-column-vector conformal to given object

IN      trans        option to transpose object
IN      obj          original object
IN      numVects     number of vectors in projected multi-column-vector
IN      ownerCol     index of column of nodes which contains owning node(s)
OUT     newObj      created object

----------------------------------------------------------------------------*/
{
   if ( trans == PLA_NO_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Pmcolvector_create( obj.base->datatype,
				    obj.m,
				    ( numVects == PLA_INHERIT ? 
				      obj.n : numVects ),
				    obj.base->alignM + obj.offm,
				    ( ownerCol == PLA_INHERIT ? 
				      PLA_ALL_COLS : ownerCol ),  
				    obj.base->templ,
				    newObj );
	    break;     
	 case PLA_MROWVECTOR:
	    PLA_Abort( "Must transpose multi-row-vectors to make multi-column-vector conformal", __LINE__, __FILE__ );	  
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Pmcolvector_create( obj.base->datatype,
				    obj.m,
				    ( numVects == PLA_INHERIT ? 
				      obj.n : numVects ),
				    obj.base->alignM + obj.offm,
				    ( ownerCol == PLA_INHERIT ? 
				      PLA_ALL_COLS : ownerCol ),  
				    obj.base->templ,
				    newObj );
	    break;
	 case PLA_PMROWVECTOR:
	    PLA_Abort( "Must transpose projected multi-row-vectors to make multi-column-vector conformal", __LINE__, __FILE__ );	  
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Pmcolvector_create( obj.base->datatype,
				    obj.m,
				    ( numVects == PLA_INHERIT ? 
				      obj.n : numVects ),
				    obj.base->alignM + obj.offm,
				    ( ownerCol == PLA_INHERIT ? 
				      obj.base->ownerCol : ownerCol ),  
				    obj.base->templ,
				    newObj );
	    break;
	 default:
	    PLA_Abort( "Cannot create projected multi-column-vectors conformal to anything but multivectors and matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else if ( trans == PLA_TRANSPOSE )
   {
      switch ( obj.base->objtype )
      {
	 case PLA_MATRIX:
	    PLA_Pmcolvector_create( obj.base->datatype,
				    obj.n,
				    ( numVects == PLA_INHERIT ? 
				      obj.m : numVects ),
				    obj.base->alignN + obj.offn,
				    ( ownerCol == PLA_INHERIT ? 
				      PLA_ALL_COLS : ownerCol ),  
				    obj.base->templ,
				    newObj );
	    break;     
	 case PLA_MROWVECTOR:
	    PLA_Pmcolvector_create( obj.base->datatype,
				    obj.n, // switch order of m and n
				    ( numVects == PLA_INHERIT ?
				      obj.m : numVects ),
				    obj.base->alignN + obj.offn,
				    ( ownerCol == PLA_INHERIT ? 
				      PLA_ALL_COLS : ownerCol ),  
				    obj.base->templ,
				    newObj );
	    break;
	 case PLA_MCOLVECTOR:
	    PLA_Abort( "Cannot transpose multi-column-vectors to another multi-column-vector", __LINE__, __FILE__ );
	    break;
	 case PLA_PMROWVECTOR:
	    PLA_Pmcolvector_create( obj.base->datatype,
				    obj.n, // switch order of m and n
				    ( numVects == PLA_INHERIT ? 
				      obj.m : numVects ),
				    obj.base->alignN + obj.offn,
				    ( ownerCol == PLA_INHERIT ? 
				      PLA_ALL_COLS : ownerCol ),  
				    obj.base->templ,
				    newObj );
	    break;
	 case PLA_PMCOLVECTOR:
	    PLA_Abort( "Cannot transpose projected multi-column-vectors to another multi-column-vector", __LINE__, __FILE__ );
	    break;
	 default:
	    PLA_Abort( "Cannot create projected multi-column-vectors conformal to anything but multivectors and matrices", __LINE__, __FILE__ );
	    break;
      }
   }
   else
      PLA_Abort( "Unknown value for transposition", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Obj_free( PLA_Obj *obj )
/*----------------------------------------------------------------------------

Purpose :   Free the object

IN      obj        object

----------------------------------------------------------------------------*/
{
   if ( obj == NULL ) 
      PLA_Abort( "Free called on NULL pointer", __LINE__, __FILE__ );
   else 
   {
      if ( obj->base == NULL ) 
	 PLA_Abort( "PLA_Obj base pointer NULL", __LINE__, __FILE__ );
      else 
      {
	 PLA_free( obj->base->buffer );
	 PLA_free( (void *) obj->base );
	 obj->offm = 0;
	 obj->offn = 0;
	 obj->m    = 0;
	 obj->n    = 0;
      }
   }
   
   return PLA_SUCCESS;
}


PLA_Objtype PLA_Obj_objtype( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract the object type

IN      obj       object

Return :    object type

----------------------------------------------------------------------------*/
{
   return obj.base->objtype;
}


PLA_Datatype PLA_Obj_datatype( PLA_Obj obj ) 
/*----------------------------------------------------------------------------

Purpose :   Extract the datatype

IN      obj       object

Return :    datatype

----------------------------------------------------------------------------*/
{
   return obj.base->datatype;
}


PLA_Temp PLA_Obj_template( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract the template

IN      obj     object

Return :    template

----------------------------------------------------------------------------*/
{
   return obj.base->templ;
}


PLA_Error PLA_Obj_global_info( PLA_Obj obj,
			       int     *m,
			       int     *n,
			       int     *alignM,
			       int     *alignN,
			       int     *ownerRow,
			       int     *ownerCol )
/*----------------------------------------------------------------------------

Purpose :   Extract global information

IN      obj            object
OUT     m              global length
OUT     n              global width
OUT     alignM         alignment to template
OUT     alignN         alignment to template
OUT     ownerRow       owner of the row
OUT     ownerCol       owner of the column

----------------------------------------------------------------------------*/
{
   *m        = obj.m;
   *n        = obj.n; 
   *alignM   = obj.base->alignM;
   *alignN   = obj.base->alignN;
   *ownerRow = obj.base->ownerRow;
   *ownerCol = obj.base->ownerCol;
   
   return PLA_SUCCESS;
}


int PLA_Obj_length( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract the global length

IN      obj       object

Return :    global length

----------------------------------------------------------------------------*/
{
   return obj.m;
}


int PLA_Obj_width( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract the global width

IN      obj       object

Return :    global width

----------------------------------------------------------------------------*/
{
   return obj.n;
}


int PLA_Obj_align_m( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract alignment to template

IN      obj            object

Return :    alignment to template

----------------------------------------------------------------------------*/
{
   return obj.base->alignM;
}


int PLA_Obj_align_n( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract alignment to template

IN      obj            object

Return :    alignment to template

----------------------------------------------------------------------------*/
{
   return obj.base->alignN;
}


int PLA_Obj_owner_row( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract owner of the row

IN      obj            object

Return :    owner of the row

----------------------------------------------------------------------------*/
{
   return obj.base->ownerRow;
}


int PLA_Obj_owner_col( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract owner of the column

IN      obj            object

Return :     owner of the column

----------------------------------------------------------------------------*/
{
   return obj.base->ownerCol;
}


/*----------------------------------------------------------------------------

   Let's say we have a 3 x 2 mesh of nodes, and 
   we do a wrapping of both inducing vectors:  
           ---------------------------------------------------
   P_{0,0} | A_{0,0} A_{0,2} A{0,4} | A_{0,1} A_{0,3} A{0,5} | P_{0,1}
           | A_{3,0} A_{3,2} A{3,4} | A_{3,1} A_{3,3} A{3,5} |
           ---------------------------------------------------
   P_{1,0} | A_{1,0} A_{1,2} A{1,4} | A_{1,1} A_{1,3} A{1,5} | P_{1,1}
           | A_{4,0} A_{4,2} A{4,4} | A_{4,1} A_{4,3} A{4,5} |
           ---------------------------------------------------
   P_{2,0} | A_{2,0} A_{2,2} A{2,4} | A_{2,1} A_{2,3} A{2,5} | P_{2,1}
           | A_{5,0} A_{5,2} A{5,4} | A_{5,1} A_{5,3} A{5,5} |
           ---------------------------------------------------
	   
   If the row dimension of the matrix is slightly large, m = 8 * mb, 
   then the matrix would be distributed this way:
           ---------------------------------------------------
   P_{0,0} | A_{0,0} A_{0,2} A{0,4} | A_{0,1} A_{0,3} A{0,5} | P_{0,1}
           | A_{3,0} A_{3,2} A{3,4} | A_{3,1} A_{3,3} A{3,5} |
           | A_{6,0} A_{6,2} A{6,4} | A_{6,1} A_{6,3} A{6,5} |
           ---------------------------------------------------
   P_{1,0} | A_{1,0} A_{1,2} A{1,4} | A_{1,1} A_{1,3} A{1,5} | P_{1,1}
           | A_{4,0} A_{4,2} A{4,4} | A_{4,1} A_{4,3} A{4,5} |
           | A_{7,0} A_{7,2} A{7,4} | A_{3,1} A_{7,3} A{7,5} |
           ---------------------------------------------------
   P_{2,0} | A_{2,0} A_{2,2} A{2,4} | A_{2,1} A_{2,3} A{2,5} | P_{2,1}
           | A_{5,0} A_{5,2} A{5,4} | A_{5,1} A_{5,3} A{5,5} |
           ---------------------------------------------------
   
	            A_{i,j} -> P_{i%r,j%c}


   Here is the matrix distributed if no wrapping is done:
           ---------------------------------------------------
   P_{0,0} | A_{0,0} A_{0,1} A{0,2} | A_{0,3} A_{0,4} A{0,5} | P_{0,1}
           | A_{1,0} A_{1,1} A{1,2} | A_{1,3} A_{1,4} A{1,5} |
           ---------------------------------------------------
   P_{1,0} | A_{2,0} A_{2,1} A{2,2} | A_{2,3} A_{2,4} A{2,5} | P_{1,1}
           | A_{3,0} A_{3,1} A{3,2} | A_{3,3} A_{3,4} A{3,5} |
           ---------------------------------------------------
   P_{2,0} | A_{4,0} A_{4,1} A{4,2} | A_{4,3} A_{4,4} A{4,5} | P_{2,1}
           | A_{5,0} A_{5,1} A{5,2} | A_{5,3} A_{5,4} A{5,5} |
           ---------------------------------------------------  

   The matrix distributed with no wrapping, and 
   the row dimension is slightly large: 
           ---------------------------------------------------
   P_{0,0} | A_{0,0} A_{0,1} A{0,2} | A_{0,3} A_{0,4} A{0,5} | P_{0,1}
           | A_{1,0} A_{1,1} A{1,2} | A_{1,3} A_{1,4} A{1,5} |
           | A_{6,0} A_{6,1} A{6,2} | A_{6,3} A_{6,4} A{6,5} |
           | A_{7,0} A_{7,1} A{7,2} | A_{7,3} A_{7,4} A{7,5} |
           ---------------------------------------------------
   P_{1,0} | A_{2,0} A_{2,1} A{2,2} | A_{2,3} A_{2,4} A{2,5} | P_{1,1}
           | A_{3,0} A_{3,1} A{3,2} | A_{3,3} A_{3,4} A{3,5} |
           ---------------------------------------------------
   P_{2,0} | A_{4,0} A_{4,1} A{4,2} | A_{4,3} A_{4,4} A{4,5} | P_{2,1}
           | A_{5,0} A_{5,1} A{5,2} | A_{5,3} A_{5,4} A{5,5} |
           ---------------------------------------------------
 	  
	            A_{i,j} -> P_{i/c%r,j/r%c}
  
----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
  
   MPI uses row major ordering

----------------------------------------------------------------------------*/
#define rowORcol( distr, stride ) ( (distr) == PLA_COL_MAJOR ? (stride) : 1 )


PLA_Bool PLA_Obj_is_valid_state( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Check to see if local processor is part of object's template

IN      obj         object

Return :    if local processor is part of object's template

----------------------------------------------------------------------------*/
{
   if ( obj.base->templ->me == PLA_UNDEFINED )
      return FALSE;
   
   return TRUE;
}


int PLA_Type_size( PLA_Datatype datatype )
/*----------------------------------------------------------------------------

Purpose :   Extract the size of the datatype

IN      datatype        datatype

Return :    size of the datatype

----------------------------------------------------------------------------*/
{
   int typesize = 0;
   
   switch ( datatype )
   {
      case PLA_FLOAT:
	 MPI_Type_size( MPI_FLOAT, &typesize );
	 break;
      case PLA_DOUBLE:
	 MPI_Type_size( MPI_DOUBLE, &typesize );
	 break;
      case PLA_COMPLEX:
	 MPI_Type_size( MPI_COMPLEX, &typesize );
	 break;
      case PLA_DOUBLE_COMPLEX:
	 MPI_Type_size( MPI_DOUBLE_COMPLEX, &typesize );
	 break;
      case PLA_INT:
	 MPI_Type_size( MPI_INT, &typesize );
	 break;
      case PLA_CONSTANT:
      {
	 int tempsize;
	 
	 MPI_Type_size( MPI_DOUBLE, &tempsize );
	 typesize += tempsize * 2; // float + double
	 
	 MPI_Type_size( MPI_COMPLEX, &tempsize );
	 typesize += tempsize;
	 
	 MPI_Type_size( MPI_DOUBLE_COMPLEX, &tempsize );
	 typesize += tempsize;
	 
	 MPI_Type_size( MPI_INT, &tempsize );
	 typesize += tempsize * 2; // alignment
	 
	 break;
      }
      default:
	 PLA_Abort( "Datatype not yet implemented", __LINE__, __FILE__ );
	 break;
   }
   
   return typesize;
}


int PLA_compute_local_size( int m_global, 
			    int strideCol,
			    int strideRow,
			    int nb, 
			    int nprocs, 
			    int me )
/*----------------------------------------------------------------------------

Purpose :   Commute the local size of object for one dimension

IN     m_global      global length of vector
IN     strideCol     stride for indexing processors within columns
IN     strideRow     stride for indexing processors within rows
IN     nb            block size with which vector is cyclicly wrapped
IN     nprocs        number of processors onto which the vector is wrapped
IN     me            rank among the processors

Return  :   Local size of vector
  
Assumption :   Vector is aligned at the first processor in the simplest way

----------------------------------------------------------------------------*/
{
   int 
      result, nblocks, ncycles;
   
   // A block is a subvector of length nb  
   // A cycle is a subvector of length nb * nprocs
   
   // Compute the number of complete blocks in vector of length m_global
   nblocks = m_global / nb;
   
   // Compute the number of complete cycles in vector of length m_global
   ncycles = nblocks / nprocs;
   
   // Compute contribution to result from ncycles (one block per cycle)
   result  = ncycles * nb;
   
   // Update the number of complete blocks to how many are left after ncycles
   nblocks = nblocks - ncycles * nprocs;
   
   // Only if me < nblocks is there a left-over block on this processor
   if ( ( me / strideCol + me % strideCol * strideRow ) < nblocks ) 
      result += nb;
   
   // If me == nblocks, the final partial block (if any) is on this processor
   if ( ( me / strideCol + me % strideCol * strideRow ) == nblocks ) 
      result += m_global % nb;
   
   return result;
}


int PLA_compute_mypart( int m_global, 
			int align, 
			int strideCol,
			int strideRow,
			int nb, 
			int nprocs, 
			int me )
/*----------------------------------------------------------------------------

Purpose :   Commute the local size of object for one dimension 
            without the alignment

IN     m_global      global length of vector
IN     align         alignment of first element of vector to template
IN     strideCol     stride for indexing processors within columns
IN     strideRow     stride for indexing processors within rows
IN     nb            block size with which vector is cyclicly wrapped
IN     nprocs        number of processors onto which the vector is wrapped
IN     me            rank among the processors

Return :    Local length of vector

----------------------------------------------------------------------------*/
{
   return
      PLA_compute_local_size( m_global + align, strideCol, strideRow, 
			      nb, nprocs, me ) 
      -
      PLA_compute_local_size( align, strideCol, strideRow, nb, nprocs, me );
}


int PLA_compute_local_size_matrix( int m_global, 
				   int xdim,
				   int nb, 
				   int nprocs, 
				   int me )
/*----------------------------------------------------------------------------

Purpose :   Commute the local size of object for one dimension

IN     m_global      global length of vector
IN     xdim          extra dimension
IN     nb            block size with which vector is cyclicly wrapped
IN     nprocs        number of processors onto which the vector is wrapped
IN     me            rank among the processors

Return  :   Local size of vector
  
Assumption :   Vector is aligned at the first processor in the simplest way, 
               and matrix is stored not wrapped in this dimension
----------------------------------------------------------------------------*/
{
   int 
      result, nblocks, ncycles;
   
   // A block is a subvector of length nb
   // A cycle is a subvector of length nb * nprocs
   
   // Compute the number of complete blocks in vector of length m_global
   nblocks = m_global / nb;
   
   // Compute the number of complete cycles in vector of length m_global
   ncycles = nblocks / nprocs;
   
   // Compute contribution to result from ncycles (one block per cycle)
   result  = ncycles * nb * xdim;
   
   // Update the number of complete blocks to how many are left after ncycles
   nblocks = nblocks - ncycles * nprocs;
   
   // Compute the number of rows of processors that have full cycle left
   ncycles = nblocks / xdim;
   
   // The extra blocks not in a row of processors
   nblocks = nblocks - ncycles * xdim;
   
   // Only if me < ncycles is there a left-over block on this processor
   if ( me < ncycles ) 
      result += nb * xdim;
   
   // If me == ncycles, the final partial block (if any) is on this processor
   if ( me == ncycles ) 
      result += nblocks * nb + m_global % nb;
   
   return result;
}


int PLA_compute_mypart_matrix( int m_global, 
			       int align, 
			       int xdim,
			       int nb, 
			       int nprocs, 
			       int me )
/*----------------------------------------------------------------------------

Purpose :   Commute the local size of object for one dimension 
            without the alignment

IN     m_global      global length of vector
IN     align         alignment of first element of vector to template
IN     xdim          extra dimension
IN     nb            block size with which vector is cyclicly wrapped
IN     nprocs        number of processors onto which the vector is wrapped
IN     me            rank among the processors

Return :    Local length of vector

----------------------------------------------------------------------------*/
{
   return
      PLA_compute_local_size_matrix( m_global + align, xdim, nb, nprocs, me ) 
      -
      PLA_compute_local_size_matrix( align, xdim, nb, nprocs, me );
}


int PLA_Obj_compute_local_ldim( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Compute the local leading dimension

IN      obj         object

Return :    local leading dimension of array holding data

----------------------------------------------------------------------------*/
{
   int ldim = 0;

   if ( !PLA_Obj_is_valid_state( obj ) )
      return 0;
   
   switch ( obj.base->objtype ) 
   {
      case PLA_MATRIX:
	 if ( obj.base->templ->colDistr == PLA_ROW_MAJOR )
	    ldim = PLA_compute_mypart( obj.base->m,
				       obj.base->alignM,
				       1,
				       1,
				       obj.base->templ->mb,
				       obj.base->templ->nprows,
				       obj.base->templ->myrow );
	 else
	    ldim = PLA_compute_mypart_matrix( obj.base->m,
					      obj.base->alignM,
					      obj.base->templ->npcols,
					      obj.base->templ->mb,
					      obj.base->templ->nprocs,
					      obj.base->templ->myrow );
	 break;
	 
      case PLA_MROWVECTOR: 
	 ldim = obj.base->m;
	 break;
	 
      case PLA_MCOLVECTOR:
	 ldim = PLA_compute_mypart( obj.base->m,
				    obj.base->alignM,
				    rowORcol( obj.base->templ->colDistr, 
					      obj.base->templ->npcols ),
				    rowORcol( obj.base->templ->colDistr, 
					      obj.base->templ->nprows ),
				    obj.base->templ->mb,
				    obj.base->templ->nprocs,
				    obj.base->templ->me );
	 break;
	 
      case PLA_MSCALAR:
	 if ( ( obj.base->ownerRow == PLA_ALL_ROWS || 
		obj.base->ownerRow == obj.base->templ->myrow ) &&
	      ( obj.base->ownerCol == PLA_ALL_COLS ||
		obj.base->ownerCol == obj.base->templ->mycol ) )
	    ldim = obj.base->m;
	 else 
	    ldim = 0;
	 break;
	 
      case PLA_PMROWVECTOR:
	 if ( obj.base->ownerRow == PLA_ALL_ROWS ||
	      obj.base->ownerRow == obj.base->templ->myrow )
	    ldim = obj.base->m;
	 else
	    ldim = 0;
	 break;
	 
      case PLA_PMCOLVECTOR:
	 if ( obj.base->ownerCol == PLA_ALL_COLS ||
	      obj.base->ownerCol == obj.base->templ->mycol )
	    ldim = PLA_compute_mypart( obj.base->m,
				       obj.base->alignM,
				       1,
				       1,
				       obj.base->templ->mb,
				       obj.base->templ->nprows, 
				       obj.base->templ->myrow );
	 else
	    ldim = 0;
	 break;
	 
      case PLA_CONSTANT:
	 ldim = 1;
	 break;
	 
      default:
	 PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
	 break;
   }  
   
   return ldim;
}


int PLA_Obj_local_length( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract the local length

IN      obj         object

Return :    local length

----------------------------------------------------------------------------*/
{ 
   int localm = 0;

   if ( !PLA_Obj_is_valid_state( obj ) )
      return 0;
   
   switch ( obj.base->objtype ) 
   {
      case PLA_MATRIX:
	 if ( obj.base->templ->colDistr == PLA_ROW_MAJOR )
	    localm = PLA_compute_mypart( obj.m,
					 obj.base->alignM + obj.offm,
					 1,
					 1,
					 obj.base->templ->mb,
					 obj.base->templ->nprows,
					 obj.base->templ->myrow );
	 else
	    localm = PLA_compute_mypart_matrix( obj.m,
						obj.base->alignM + obj.offm,
						obj.base->templ->npcols,
						obj.base->templ->mb,
						obj.base->templ->nprocs,
						obj.base->templ->myrow );
	 break;

      case PLA_MROWVECTOR: 
	 localm = obj.m;
	 break;
	 
      case PLA_MCOLVECTOR:
	 localm = PLA_compute_mypart( obj.m,
				      obj.base->alignM + obj.offm,
				      rowORcol( obj.base->templ->colDistr,
						obj.base->templ->npcols ),
				      rowORcol( obj.base->templ->colDistr, 
						obj.base->templ->nprows ),
				      obj.base->templ->mb,
				      obj.base->templ->nprocs,
				      obj.base->templ->me );
	 break;
	 	 
      case PLA_MSCALAR:
	 if ( ( obj.base->ownerRow == PLA_ALL_ROWS || 
		obj.base->ownerRow == obj.base->templ->myrow ) &&
	      ( obj.base->ownerCol == PLA_ALL_COLS ||
		obj.base->ownerCol == obj.base->templ->mycol ) )
	    localm = obj.m;
	 else 
	    localm = 0;
	 break;
	 
      case PLA_PMROWVECTOR:
	 if ( obj.base->ownerRow == PLA_ALL_ROWS ||
	      obj.base->ownerRow == obj.base->templ->myrow )
	    localm = obj.m;
	 else
	    localm = 0;
	 break;
	 
      case PLA_PMCOLVECTOR:
	 if ( obj.base->ownerCol == PLA_ALL_COLS ||
	      obj.base->ownerCol == obj.base->templ->mycol )
	    localm = PLA_compute_mypart( obj.m,
					 obj.base->alignM + obj.offm,
					 1,
					 1,
					 obj.base->templ->mb,
					 obj.base->templ->nprows, 
					 obj.base->templ->myrow );
	 else
	    localm = 0;
	 break;
	 
      case PLA_CONSTANT:
	 localm = 1;
	 break;
	 
      default:
	 PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
	 break;
   }
   
   return localm;
}


int PLA_Obj_local_width( PLA_Obj obj )			 
/*----------------------------------------------------------------------------

Purpose :   Extract the local width

IN      obj          object

Return :    local width

----------------------------------------------------------------------------*/
{
   int localn = 0;

   if ( !PLA_Obj_is_valid_state( obj ) )
      return 0;
   
   switch ( obj.base->objtype ) 
   {
      case PLA_MATRIX: 
	 if ( obj.base->templ->rowDistr == PLA_COL_MAJOR )
	    localn = PLA_compute_mypart( obj.n,
					 obj.base->alignN + obj.offn,
					 1,
					 1,
					 obj.base->templ->nb, 
					 obj.base->templ->npcols, 
					 obj.base->templ->mycol );
	 else
	    localn = PLA_compute_mypart_matrix( obj.n,
						obj.base->alignN + obj.offn,
						obj.base->templ->nprows,
						obj.base->templ->nb,
						obj.base->templ->nprocs,
						obj.base->templ->mycol ); 
	 break;
	 
      case PLA_MROWVECTOR: 
	 localn = PLA_compute_mypart( obj.n,
				      obj.base->alignN + obj.offn,
				      rowORcol( obj.base->templ->rowDistr, 
						obj.base->templ->npcols ),
				      rowORcol( obj.base->templ->rowDistr, 
						obj.base->templ->nprows ),
				      obj.base->templ->nb, 
				      obj.base->templ->nprocs, 
				      obj.base->templ->me );
	 break;
	 
      case PLA_MCOLVECTOR:
	 localn = obj.n;
	 break;
	 
      case PLA_MSCALAR:
	 if ( ( obj.base->ownerRow == PLA_ALL_ROWS || 
		obj.base->ownerRow == obj.base->templ->myrow ) &&
	      ( obj.base->ownerCol == PLA_ALL_COLS || 
		obj.base->ownerCol == obj.base->templ->mycol ) )
	    localn = obj.n;
	 else
	    localn = 0;
	 break;
	 
      case PLA_PMROWVECTOR:
	 if ( obj.base->ownerRow == PLA_ALL_ROWS ||
	      obj.base->ownerRow == obj.base->templ->myrow )
	    localn = PLA_compute_mypart( obj.n,
					 obj.base->alignN + obj.offn,
					 1,
					 1,
					 obj.base->templ->nb,
					 obj.base->templ->npcols,
					 obj.base->templ->mycol );
	 else
	    localn = 0;
	 break;
	 
      case PLA_PMCOLVECTOR:   
	 if ( obj.base->ownerCol == PLA_ALL_COLS ||
	      obj.base->ownerCol == obj.base->templ->mycol )
	    localn = obj.n;
	 else
	    localn = 0;
	 break;
	 
      case PLA_CONSTANT:
	 localn = 1;
	 break;
	 
      default:
	 PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
	 break;
   }
   
   return localn;
}


int PLA_Obj_local_offm( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract the local offm

IN      obj         object

Return :    local offm

----------------------------------------------------------------------------*/
{ 
   int localoffm = 0;

   if ( !PLA_Obj_is_valid_state( obj ) )
      return 0;
   
   switch ( obj.base->objtype ) 
   {
      case PLA_MATRIX:
	 if ( obj.base->templ->colDistr == PLA_ROW_MAJOR )
	    localoffm = PLA_compute_mypart( obj.offm,
					    obj.base->alignM,
					    1,
					    1,
					    obj.base->templ->mb,
					    obj.base->templ->nprows,
					    obj.base->templ->myrow );
	 else
	    localoffm = PLA_compute_mypart_matrix( obj.offm,
						   obj.base->alignM,
						   obj.base->templ->npcols,
						   obj.base->templ->mb,
						   obj.base->templ->nprocs,
						   obj.base->templ->myrow );
	 break;
	 
      case PLA_MROWVECTOR: 
	 localoffm = obj.offm;
	 break;
	 
      case PLA_MCOLVECTOR:
	 localoffm = PLA_compute_mypart( obj.offm,
					 obj.base->alignM,
					 rowORcol( obj.base->templ->colDistr, 
						   obj.base->templ->npcols ),
					 rowORcol( obj.base->templ->colDistr, 
						   obj.base->templ->nprows ),
					 obj.base->templ->mb,
					 obj.base->templ->nprocs,
					 obj.base->templ->me );
	 break;
	 
      case PLA_MSCALAR:
	 if ( ( obj.base->ownerRow == PLA_ALL_ROWS || 
		obj.base->ownerRow == obj.base->templ->myrow ) &&
	      ( obj.base->ownerCol == PLA_ALL_COLS ||
		obj.base->ownerCol == obj.base->templ->mycol ) )
	    localoffm = obj.offm;
	 else 
	    localoffm = 0;
	 break;
	 
      case PLA_PMROWVECTOR:
	 if ( obj.base->ownerRow == PLA_ALL_ROWS ||
	      obj.base->ownerRow == obj.base->templ->myrow )
	    localoffm = obj.offm;
	 else
	    localoffm = 0;
	 break;
	 
      case PLA_PMCOLVECTOR:
	 if ( obj.base->ownerCol == PLA_ALL_COLS ||
	      obj.base->ownerCol == obj.base->templ->mycol )
	    localoffm = PLA_compute_mypart( obj.offm,
					    obj.base->alignM,
					    1,
					    1,
					    obj.base->templ->mb,
					    obj.base->templ->nprows, 
					    obj.base->templ->myrow );
	 else
	    localoffm = 0;
	 break;
	 
      case PLA_CONSTANT:
	 localoffm = 0;
	 break;
	 
      default:
	 PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
	 break;
   }
   
   return localoffm;
}


int PLA_Obj_local_offn( PLA_Obj obj )			 
/*----------------------------------------------------------------------------

Purpose :   Extract the local offn

IN      obj          object

Return :    local offn

----------------------------------------------------------------------------*/
{
   int localoffn = 0;

   if ( !PLA_Obj_is_valid_state( obj ) )
      return 0;
   
   switch ( obj.base->objtype ) 
   {
      case PLA_MATRIX: 
	 if ( obj.base->templ->rowDistr == PLA_COL_MAJOR )
	    localoffn = PLA_compute_mypart( obj.offn,
					    obj.base->alignN,
					    1,
					    1,
					    obj.base->templ->nb, 
					    obj.base->templ->npcols, 
					    obj.base->templ->mycol );
	 else
	    localoffn = PLA_compute_mypart_matrix( obj.offn,
						   obj.base->alignN,
						   obj.base->templ->nprows,
						   obj.base->templ->nb,
						   obj.base->templ->nprocs,
						   obj.base->templ->mycol ); 
	 break;
	 
      case PLA_MROWVECTOR: 
	 localoffn = PLA_compute_mypart( obj.offn,
					 obj.base->alignN,
					 rowORcol( obj.base->templ->rowDistr, 
						   obj.base->templ->npcols ),
					 rowORcol( obj.base->templ->rowDistr, 
						   obj.base->templ->nprows ),
					 obj.base->templ->nb, 
					 obj.base->templ->nprocs, 
					 obj.base->templ->me );
	 break;
	 
      case PLA_MCOLVECTOR:
	 localoffn = obj.offn;
	 break;
	 
      case PLA_MSCALAR:
	 if ( ( obj.base->ownerRow == PLA_ALL_ROWS || 
		obj.base->ownerRow == obj.base->templ->myrow ) &&
	      ( obj.base->ownerCol == PLA_ALL_COLS || 
		obj.base->ownerCol == obj.base->templ->mycol ) )
	    localoffn = obj.offn;
	 else
	    localoffn = 0;
	 break;
	 
      case PLA_PMROWVECTOR:
	 if ( obj.base->ownerRow == PLA_ALL_ROWS ||
	      obj.base->ownerRow == obj.base->templ->myrow )
	    localoffn = PLA_compute_mypart( obj.offn,
					    obj.base->alignN,
					    1,
					    1,
					    obj.base->templ->nb,
					    obj.base->templ->npcols,
					    obj.base->templ->mycol );
	 else
	    localoffn = 0;
	 break;
	 
      case PLA_PMCOLVECTOR:   
	 if ( obj.base->ownerCol == PLA_ALL_COLS ||
	      obj.base->ownerCol == obj.base->templ->mycol )
	    localoffn = obj.offn;
	 else
	    localoffn = 0;
	 break;
	 
      case PLA_CONSTANT:
	 localoffn = 0;
	 break;
	 
      default:
	 PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
	 break;
   }
   
   return localoffn;
}


int PLA_Obj_local_ldim( PLA_Obj obj )
/*----------------------------------------------------------------------------

Purpose :   Extract the local leading dimension

IN      obj         object

Return :    local leading dimension of array holding data

----------------------------------------------------------------------------*/
{
   return obj.base->ldim;
}


PLA_Error PLA_Obj_local_info( PLA_Obj obj,
			      int     *localm,
			      int     *localn,
			      int     *ldim,
			      void    **buffer )
/*----------------------------------------------------------------------------

Purpose :   Extract local information

IN      obj         object
OUT     localm      local row dimension
OUT     localn      local column dimension
OUT     ldim        local leading dimension of array holding data
OUT     buffer      address of local data

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_objtype( obj ) == PLA_CONSTANT )
      PLA_Abort( "Cannot extract from constant", __LINE__, __FILE__ );
   else
   {
      *localm = PLA_Obj_local_length( obj );
      *localn = PLA_Obj_local_width( obj );
      *ldim   = PLA_Obj_local_ldim( obj );
      *buffer = obj.base->buffer + 
	 ( PLA_Obj_local_offn( obj ) *
	   PLA_Obj_local_ldim( obj ) +
	   PLA_Obj_local_offm( obj ) ) *
	 PLA_Type_size( PLA_Obj_datatype( obj ) );
   }
   
   return PLA_SUCCESS;
}
