/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


int PLA_sqrt_floor( int nprocs ) 
/*----------------------------------------------------------------------------

Purpose : Finds the floor( sqrt( nprocs ) )

IN      nprocs     number of processors

Return  : floor( sqrt( nprocs ) )

----------------------------------------------------------------------------*/
{
   int i = 0;
   while ( i * i <= nprocs ) 
      i++;
   return i - 1;
}


void PLA_factorization( int nprocs,
			int nprows, 
			int npcols, 
			int *dims )
/*----------------------------------------------------------------------------

Purpose :   Create a 2D integer factorization for a number of processors
            and a suggested ratio

IN      nprocs     number of processors
IN      nprows     number of rows
IN      npcols     number of columns
    Suggested ratio - nprows x npcols
OUT     dims       two dimensional decomposition of processors

----------------------------------------------------------------------------*/
{
   /*
      Set values of two dimensions
      Unless specified, number of rows is always larger of the two
   */
   
   // Set the dimensions if rows and columns match number of processors
   if ( nprocs == nprows * npcols ) 
   {
      dims[0] = nprows;
      dims[1] = npcols;
   }
   // Try to preserve the ratio as closely as possible
   else if ( nprocs % ( nprows * npcols ) == 0 )
   {
      int num = nprocs / ( nprows * npcols );
      int sqrt_num = PLA_sqrt_floor( num );
      while ( num % sqrt_num != 0 ) 
	 sqrt_num = sqrt_num - 1;
      
      if ( nprows > npcols )
      {
	 npcols *= ( num / sqrt_num );
	 nprows *= sqrt_num;
	 
	 if ( nprows > npcols ) 
	 {
	    dims[0] = nprows;
	    dims[1] = npcols;
	 }
	 else 
	 {
	    dims[0] = npcols;
	    dims[1] = nprows;
	 }
      }
      else
      {
	 nprows *= ( num / sqrt_num );
	 npcols *= sqrt_num;
	 
	 if ( nprows > npcols ) 
	 {
	    dims[0] = nprows;
	    dims[1] = npcols;
	 }
	 else
	 {
	    dims[0] = npcols;
	    dims[1] = nprows; 
	 }
      }      
   }
   // Ratio does not match the total number of processors
   else
   {
      int num = PLA_sqrt_floor( nprocs );
      while ( nprocs % num != 0 ) 
	 num = num - 1;
      
      if ( nprows > npcols )
      {
	 dims[0] = nprocs / num; 
	 dims[1] = num;
      }
      else
      { 
	 dims[0] = num; 
	 dims[1] = nprocs / num;
      }
   }
   
   return;
}


PLA_Error PLA_Temp_create_without_buffer( MPI_Comm  comm,
					  int       nprows, 
					  int       npcols,  
					  int       mb, 
					  PLA_Distr colDistr, 
					  int       nb, 
					  PLA_Distr rowDistr,
					  PLA_Temp  templ )
/*----------------------------------------------------------------------------

Purpose :   Create a template without creating the buffer

IN      comm       MPI communicator describing all processors
IN      nprows     number of rows
IN      npcols     number of columns
    Suggested ratio - nprows x npcols
IN      mb         block size in rows
IN      colDistr   inducing vector distribution in columns
IN      nb         block size in columns
IN      rowDistr   inducing vector distribution in rows
OUT     templ      new template

----------------------------------------------------------------------------*/
{
   int
      dims[2],  
      periods[2],
      remain_dims[2],   
      error;
   
   MPI_Comm 
      comm2d;
   
   /*
      Error checking
   */
   if ( nprows < 1 || npcols < 1 )
      PLA_Abort( "Malformed ratio of processors", __LINE__, __FILE__ );
   
   if ( mb < 1 || nb < 1 )
      PLA_Abort( "Malformed blocking size", __LINE__, __FILE__ );
   
   if ( colDistr != PLA_ROW_MAJOR && colDistr != PLA_COL_MAJOR )
      PLA_Abort( "Undefined column vector distribution", __LINE__, __FILE__ );
   
   if ( rowDistr != PLA_ROW_MAJOR && rowDistr != PLA_COL_MAJOR )
      PLA_Abort( "Undefined row vector distribution", __LINE__, __FILE__ );

   /*
      Store the parameters passed
   */
   templ->mb       = mb;
   templ->colDistr = colDistr;
   templ->nb       = nb;
   templ->rowDistr = rowDistr;

   /*
      Cannot extract info from empty communicator
   */
   if ( comm == MPI_COMM_NULL )
   {
      templ->commAll = MPI_COMM_NULL;
      templ->commRow = MPI_COMM_NULL;
      templ->commCol = MPI_COMM_NULL;
      
      templ->me    = PLA_UNDEFINED;
      templ->myrow = PLA_UNDEFINED;
      templ->mycol = PLA_UNDEFINED;

      templ->nprocs = 0;
      templ->nprows = 0;
      templ->npcols = 0;

      return PLA_FAILURE;
   }
  
   error = MPI_Comm_dup( comm, &(templ->commAll) );
   if ( error )
      PLA_Abort( "Error duplicating communicator", __LINE__, __FILE__ );
   
   error = MPI_Comm_size( templ->commAll, &(templ->nprocs) );
   if ( error )
      PLA_Abort( "Error computing nprocs", __LINE__, __FILE__ );
   
   error = MPI_Comm_rank( templ->commAll, &(templ->me) );
   if ( error )
      PLA_Abort( "Error computing me", __LINE__, __FILE__ );  
   
   /*
      Create a communicator with a 2D topology
   */
   PLA_factorization( templ->nprocs, nprows, npcols, dims );
   
   
   periods[0] = TRUE; periods[1] = TRUE;
   
   error = MPI_Cart_create( templ->commAll, 2, dims, periods, FALSE, &comm2d );
   if ( error )
      PLA_Abort( "Error creating 2D communicator", __LINE__, __FILE__ );
   
   /* 
      Set the other 2 communicators, row, and column
      remain_dims tells which dimensions to keep and which to throw away
      For instance remain_dims[0] = 1, remain_dims[1] = 0 
      keeps the rows and throws away the columns creating a col communicator
   */
   remain_dims[0] = 1;
   remain_dims[1] = 0;
   templ->commCol = MPI_COMM_NULL;
   error = MPI_Cart_sub( comm2d, remain_dims, &(templ->commCol) ); 
   if ( error )
      PLA_Abort( "Error creating column communicator", __LINE__, __FILE__ );
   
   remain_dims[0] = 0;
   remain_dims[1] = 1;
   templ->commRow = MPI_COMM_NULL;
   error = MPI_Cart_sub( comm2d, remain_dims, &(templ->commRow) );
   if ( error )
      PLA_Abort( "Error creating row communicator", __LINE__, __FILE__ );
   
   /* 
      To compute the row dimension, look at size of the column communicator
      To compute the size of the column dimension 
      look at the size of the row communicator
   */
   error = MPI_Comm_size( templ->commCol, &(templ->nprows) );
   if ( error )
      PLA_Abort( "Error computing nprows", __LINE__, __FILE__ );
   
   error = MPI_Comm_size( templ->commRow, &(templ->npcols) );
   if ( error )
      PLA_Abort( "Error computing npcols", __LINE__, __FILE__ );
   
   /* 
      Store the position of this node within the processing grid
      The postion in the row communicator is the column index and vice versa
   */
   error = MPI_Comm_rank( templ->commCol, &(templ->myrow) );
   if ( error )
      PLA_Abort( "Error computing myrow", __LINE__, __FILE__ );
   
   error = MPI_Comm_rank( templ->commRow, &(templ->mycol) );
   if ( error )
      PLA_Abort( "Error computing mycol", __LINE__, __FILE__ );
   
   
   MPI_Comm_free( &comm2d );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Temp_create( MPI_Comm  comm,
			   int       nprows, 
			   int       npcols,  
			   int       mb, 
			   PLA_Distr colDistr, 
			   int       nb, 
			   PLA_Distr rowDistr,
			   PLA_Temp  *templ_ret )
/*----------------------------------------------------------------------------

Purpose :   Create a template

IN      comm       MPI communicator describing all processors
IN      nprows     number of rows
IN      npcols     number of columns
    Suggested ratio - nprows x npcols
IN      mb         block size in rows
IN      colDistr   inducing vector distribution in columns
IN      nb         block size in columns
IN      rowDistr   inducing vector distribution in rows
OUT     templ_ret  new template

----------------------------------------------------------------------------*/
{
   PLA_Error
      info;

   PLA_Temp
      templ = (PLA_Temp) PLA_malloc( sizeof( PLA_Template ) );
   
   info = PLA_Temp_create_without_buffer( comm,
					  nprows, npcols,
					  mb,     colDistr,
					  nb,     rowDistr,
					  templ );  
   
   *templ_ret = templ;
   
   return info;
}


PLA_Error PLA_Temp_free_without_buffer( PLA_Temp *templ )
/*----------------------------------------------------------------------------

Purpose :   Free the template

IN      templ     template to destroy

----------------------------------------------------------------------------*/
{
   if ( templ != NULL ) 
   { 
      if ( (*templ)->commAll != MPI_COMM_NULL )
	 MPI_Comm_free( &((*templ)->commAll) );

      if ( (*templ)->commCol != MPI_COMM_NULL )
	 MPI_Comm_free( &((*templ)->commCol) );
      
      if ( (*templ)->commRow != MPI_COMM_NULL )
	 MPI_Comm_free( &((*templ)->commRow) );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Temp_free( PLA_Temp *templ )
/*----------------------------------------------------------------------------

Purpose :   Free the template

IN      templ     template to destroy

----------------------------------------------------------------------------*/
{
   if ( templ != NULL ) 
   { 
      PLA_Temp_free_without_buffer( templ );
      
      PLA_free( *templ );
      *templ = NULL;
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Temp_comm_all_info( PLA_Temp templ, 
				  MPI_Comm *comm,
				  int      *me, 
				  int      *nprocs )
/*----------------------------------------------------------------------------

Purpose :   Extract information for all processors

IN      templ     template
OUT     comm      MPI communication
OUT     me        rank of processor
OUT     nprocs    number of processors

----------------------------------------------------------------------------*/
{
   *comm   = templ->commAll;
   *me     = templ->me;
   *nprocs = templ->nprocs;
   
   return PLA_SUCCESS;
}


MPI_Comm PLA_Temp_comm_all( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract communicator for all processors

IN      templ     template

Return :    MPI communicator

----------------------------------------------------------------------------*/
{
   return templ->commAll;
}


int PLA_Temp_comm_all_rank( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract rank for all processors

IN      templ     template

REturn :    rank of processor

----------------------------------------------------------------------------*/
{
   return templ->me;
}


int PLA_Temp_comm_all_size( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract number of processors for all processors

IN      templ     template

Return :    number of processors

----------------------------------------------------------------------------*/
{
   return templ->nprocs;
}


PLA_Error PLA_Temp_comm_row_info( PLA_Temp templ,
				  MPI_Comm *commRow,
				  int      *mycol,
				  int      *npcols )
/*----------------------------------------------------------------------------

Purpose :   Extract information for row of processors

IN      templ     template
OUT     commRow   MPI communication
OUT     mycol     rank of processor
OUT     npcols    number of processors in rows

----------------------------------------------------------------------------*/
{
   *commRow = templ->commRow;
   *mycol   = templ->mycol;
   *npcols  = templ->npcols;
   
   return PLA_SUCCESS;
}


MPI_Comm PLA_Temp_comm_row( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract communicator for row of processors

IN      templ     template

Return :    MPI communicator

----------------------------------------------------------------------------*/
{
   return templ->commRow;
}


int PLA_Temp_comm_row_rank( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract rank for row of processors

IN      templ     template

Return :     rank of processor

----------------------------------------------------------------------------*/
{
   return templ->mycol;
}


int PLA_Temp_comm_row_size( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract number of processors for row of processors

IN      templ     template

Return :    number of processors in rows

----------------------------------------------------------------------------*/
{
   return templ->npcols;
}


PLA_Error PLA_Temp_comm_col_info( PLA_Temp templ,
				  MPI_Comm *commCol,
				  int      *myrow,
				  int      *nprows )
/*----------------------------------------------------------------------------

Purpose :   Extract information for column of processors

IN      templ     template
OUT     commCol   MPI communication
OUT     myrow     rank of processor
OUT     nprows    number of processors in columns

----------------------------------------------------------------------------*/
{
   *commCol = templ->commCol;
   *myrow   = templ->myrow;
   *nprows  = templ->nprows;
   
   return PLA_SUCCESS;
}


MPI_Comm PLA_Temp_comm_col( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract communicator for column of processors

IN      templ     template

Return :    MPI communicator

----------------------------------------------------------------------------*/
{
   return templ->commCol;
}


int PLA_Temp_comm_col_rank( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract rank for column of processors

IN      templ     template

Return :    rank of processor

----------------------------------------------------------------------------*/
{
   return templ->myrow;
}


int PLA_Temp_comm_col_size( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract number of processors for column of processors

IN      templ     template

Return :    number of processors in columns

----------------------------------------------------------------------------*/
{
   return templ->nprows;
}


int PLA_Temp_mb( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract block size for rows

IN      templ     template

Return :    block size for rows

----------------------------------------------------------------------------*/
{
   return templ->mb;
}


PLA_Distr PLA_Temp_col_distr( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract inducing vector distribution for columns vectors

IN      templ     template

Return :    inducing vector distribution for columns vectors

----------------------------------------------------------------------------*/
{
   return templ->colDistr;
}


int PLA_Temp_nb( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract block size for columns

IN      templ     template

Return :    block size for columns

----------------------------------------------------------------------------*/
{
   return templ->nb;
}


PLA_Distr PLA_Temp_row_distr( PLA_Temp templ )
/*----------------------------------------------------------------------------

Purpose :   Extract inducing vector distribution for rows vectors

IN      templ     template

Return :    inducing vector distribution for rows vectors

----------------------------------------------------------------------------*/
{
   return templ->rowDistr;
}
