/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"
#include "mpi.h"


PLA_Error create_problem( PLA_Obj, PLA_Obj, PLA_Obj );
PLA_Error PLA_Local_cholesky( PLA_Obj );


int main( int argc, char *argv[] )
{
   int
      size,
      ownerRow = PLA_ALL_ROWS,
      ownerCol = PLA_ALL_COLS,
      me, nprocs;
   
   MPI_Comm
      comm = MPI_COMM_WORLD;
   
   PLA_Temp
      templ = PLA_TEMP_WORLD;
   
   PLA_Error
      info;
   
   PLA_Datatype
      datatype = PLA_DOUBLE;
   
   PLA_Obj
      A_matrix,
      A, x, b,
      b_norm;
   
   double
      b_norm_value,
      time, flops;
   
   
   MPI_Init( &argc, &argv );
   PLA_Init( );
   

   MPI_Comm_rank( comm, &me );
   if ( me == 0 )
   {
      printf( "Enter matrix dimension ( n ): ");
      scanf ( "%d", &size );
   }
   MPI_Bcast( &size, 1, MPI_INT, 0, comm );
   
   nprocs = PLA_Temp_comm_all_size( templ );


   PLA_Matrix_create( datatype,
		      size, size,
		      0, 0,
		      templ,
		      &A_matrix );
   PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, A_matrix,
			       ownerRow, ownerCol,
			       &A );
   PLA_Mscalar_create( datatype,
		       size, 1,
		       ownerRow, ownerCol,
		       templ,
		       &x );
   PLA_Mscalar_create( datatype,
		       size, 1,
		       ownerRow, ownerCol,
		       templ,
		       &b );
   PLA_Mscalar_create( datatype,
		       1, 1,
		       ownerRow, ownerCol,
		       templ,
		       &b_norm );
   
   create_problem( A_matrix, x, b );
   PLA_Copy( A_matrix, A );

   MPI_Barrier( comm );
   time = MPI_Wtime( );
   
   info = PLA_Local_cholesky( A );  
   
   if ( info != PLA_SUCCESS )
      PLA_printf( "Nonpositive definite matrix detected\n" );
   else 
   { 
      PLA_Local_trsv( PLA_LOWER_TRIANGULAR, PLA_NO_TRANSPOSE, PLA_NONUNIT_DIAG,
		      A, b ); 
      PLA_Local_trsv( PLA_LOWER_TRIANGULAR, PLA_TRANSPOSE, PLA_NONUNIT_DIAG, 
		      A, b );
      
      time = MPI_Wtime( ) - time;
      flops = 1.0 / 3.0 * size * size * size;
      PLA_printf( "n = %d, time = %lf, FLOPS/node = %lf\n", 
		  size, time, flops / time / nprocs );
      
      PLA_Local_axpy( PLA_MINUS_ONE, x, b );
      PLA_Local_nrm2( b, b_norm );
      
      if ( me == 0 )
      {
	 PLA_Obj_get_local_contents( PLA_NO_TRANSPOSE, b_norm, 1, 1, 
				     &b_norm_value, 1 );
	 PLA_printf( "Norm of difference between original x and computed x: %le\n", b_norm_value );
      }
   }

   PLA_Reduce( A, A_matrix );
   
   PLA_Obj_free( &A_matrix );
   PLA_Obj_free( &A );
   PLA_Obj_free( &x );
   PLA_Obj_free( &b );
   PLA_Obj_free( &b_norm );
   
   PLA_Finalize( );
   MPI_Finalize( );
   return 0;
}
