/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Copy_from_to( PLA_Trans, PLA_Obj, PLA_Obj );
PLA_Error PLA_Copy_stride( PLA_Obj, PLA_Obj );
PLA_Error PLA_Copy_stride_trans( PLA_Obj, PLA_Obj );
PLA_Error PLA_Copy_aligned( PLA_Trans, PLA_Obj, PLA_Obj );
PLA_Error PLA_Copy_comm_bcast( PLA_Trans, PLA_Obj, PLA_Obj, int );
PLA_Error PLA_Copy_comm_send( PLA_Trans, PLA_Obj, PLA_Obj, int );


PLA_Error PLA_Copy( PLA_Obj A, 
		    PLA_Obj B )
/*----------------------------------------------------------------------------

Purpose :  Copy contents (not description) between linear algebra objects

IN     A        object to be copied
OUT    B        object into which to copy

----------------------------------------------------------------------------*/
{
   PLA_Copy_x( PLA_NO_TRANSPOSE, A, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Copy_x( PLA_Trans trans, 
		      PLA_Obj   A, 
		      PLA_Obj   B )
/*----------------------------------------------------------------------------

Purpose :  Copy contents (not description) between linear algebra objects

IN     trans    option for transpose of object
IN     A        object to be copied
OUT    B        object into which to copy

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Copy_x_check( trans, A, B );
   
   if ( PLA_Obj_is_valid_state( A ) &&
	PLA_Obj_is_valid_state( B ) )
      PLA_Copy_from_to( trans, A, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Copy_from_to( PLA_Trans trans, 
			    PLA_Obj   A, 
			    PLA_Obj   B )
/*----------------------------------------------------------------------------

Purpose :  Copy contents (not description) between linear algebra objects

IN     trans    option for transpose of object
IN     A        object to be copied
OUT    B        object into which to copy

----------------------------------------------------------------------------*/
{
   if ( trans == PLA_NO_TRANSPOSE )
      PLA_Copy_stride( A, B );
   else if ( trans == PLA_TRANSPOSE )
      PLA_Copy_stride_trans( A, B );
   else
      PLA_Abort( "Unknown value for transposition", __LINE__, __FILE__ );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Copy_stride( PLA_Obj A, 
			   PLA_Obj B )
/*----------------------------------------------------------------------------

Purpose :  Copy contents (not description) between linear algebra objects
           with shifting

IN     A        object to be copied
OUT    B        object into which to copy

----------------------------------------------------------------------------*/
{
   int 
      bm, bn;

   PLA_Obj
      AL, AR,    A0, A1, A2, 
      AT1,           A01,
      AB1,           A11,
                     A21,
      BL, BR,    B0, B1, B2,
      BT1,           B01,
      BB1,           B11,
                     B21;
   
   PLA_Part_1x2( A,
		 &AL,  &AR,
		 0, PLA_LEFT );
   PLA_Part_1x2( B,
		 &BL,  &BR,
		 0, PLA_LEFT );
   
   while ( PLA_Obj_width( AL ) < PLA_Obj_width( A ) )
   {
      bn = min( PLA_Obj_split_size( AR, PLA_LEFT ),
		PLA_Obj_split_size( BR, PLA_LEFT ) );
      
      PLA_Repart_1x2_to_1x3( AL, /**/ AR,   &A0, /**/ &A1, &A2,
                             bn, PLA_RIGHT );
      PLA_Repart_1x2_to_1x3( BL, /**/ BR,   &B0, /**/ &B1, &B2,
                             bn, PLA_RIGHT );

      /*------------------------------------------------------------*/
      
      PLA_Part_2x1( A1,   &AT1,
		          &AB1,
                    0, PLA_TOP );
      PLA_Part_2x1( B1,   &BT1,
		          &BB1,
                    0, PLA_TOP );

      while ( PLA_Obj_length( AT1 ) < PLA_Obj_length( A1 ) )
      {
	 bm = min( PLA_Obj_split_size( AB1, PLA_TOP ),
		   PLA_Obj_split_size( BB1, PLA_TOP ) );
	 
	 PLA_Repart_2x1_to_3x1(  AT1,        &A01,
			      /* *** */   /* **** */
				             &A11,
				 AB1,        &A21,
				 bm, PLA_BOTTOM );
	 PLA_Repart_2x1_to_3x1(  BT1,        &B01,
			      /* *** */   /* **** */
				             &B11,
				 BB1,        &B21,
				 bm, PLA_BOTTOM );
	 
	 /*------------------------------------------------------------*/
	 
	 PLA_Copy_aligned( PLA_NO_TRANSPOSE, A11, B11 );
	 
	 /*------------------------------------------------------------*/
	 
	 PLA_Cont_with_3x1_to_2x1( &AT1,        A01,
				                A11,
				/* **** */   /* *** */
				   &AB1,        A21,
				   PLA_TOP );
	 PLA_Cont_with_3x1_to_2x1( &BT1,        B01,
				                B11,
				/* **** */   /* *** */
				   &BB1,        B21,
				   PLA_TOP );
      }
      
      /*------------------------------------------------------------*/
      
      PLA_Cont_with_1x3_to_1x2( &AL, /**/ &AR,   A0, A1, /**/ A2,
                                PLA_LEFT );
      PLA_Cont_with_1x3_to_1x2( &BL, /**/ &BR,   B0, B1, /**/ B2,
                                PLA_LEFT );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Copy_stride_trans( PLA_Obj A, 
				 PLA_Obj B )
/*----------------------------------------------------------------------------

Purpose :  Copy contents (not description) between linear algebra objects
           with shifting and transposition

IN     A        object to be copied
OUT    B        object into which to copy

----------------------------------------------------------------------------*/
{
   int 
      bm, bn;
  
   PLA_Obj
      AL, AR,    A0, A1, A2,
      AT1,           A01,
      AB1,           A11,
                     A21,
      BT,               B0,
      BB,               B1,
                        B2,
      BL1, BR1,    B10, B11, B12; 
   
   PLA_Part_1x2( A,
		 &AL,  &AR,
		 0, PLA_LEFT );
   PLA_Part_2x1( B,   &BT,
		      &BB,
		 0, PLA_TOP );
   
   while ( PLA_Obj_width( AL ) < PLA_Obj_width( A ) )
   {
      bn = min( PLA_Obj_split_size( AR, PLA_LEFT ),
		PLA_Obj_split_size( BB, PLA_TOP ) );
      
      PLA_Repart_1x2_to_1x3( AL, /**/ AR,   &A0, /**/ &A1, &A2,
                             bn, PLA_RIGHT );
      PLA_Repart_2x1_to_3x1( BT,        &B0,
			  /* ** */   /* *** */
			                &B1,
			     BB,        &B2,
			     bn, PLA_BOTTOM );
      
      /*------------------------------------------------------------*/
      
      PLA_Part_2x1( A1,   &AT1,
		          &AB1,
                    0, PLA_TOP );
      PLA_Part_1x2( B1,
		    &BL1,  &BR1,
		    0, PLA_LEFT );

      while ( PLA_Obj_length( AT1 ) < PLA_Obj_length( A1 ) )
      {
	 bm = min( PLA_Obj_split_size( AB1, PLA_TOP ),
		   PLA_Obj_split_size( BR1, PLA_LEFT ) );
	 
	 PLA_Repart_2x1_to_3x1( AT1,        &A01,
			     /* *** */   /* **** */
				            &A11,
				AB1,        &A21,
				bm, PLA_BOTTOM );
	 PLA_Repart_1x2_to_1x3( BL1, /**/ BR1,   &B10, /**/ &B11, &B12,
				bm, PLA_RIGHT );
	 
	 /*------------------------------------------------------------*/
	 
	 PLA_Copy_aligned( PLA_TRANSPOSE, A11, B11 );	  
	 
	 /*------------------------------------------------------------*/
	 
	 PLA_Cont_with_3x1_to_2x1( &AT1,        A01,
				                A11,
				 /* **** */   /* *** */
				   &AB1,        A21,
				   PLA_TOP );
	 PLA_Cont_with_1x3_to_1x2( &BL1, /**/ &BR1,   B10, B11, /**/ B12,
				   PLA_LEFT );
      }
      
      /*------------------------------------------------------------*/
      
      PLA_Cont_with_1x3_to_1x2( &AL, /**/ &AR,   A0, A1, /**/ A2,
                                PLA_LEFT );
      PLA_Cont_with_3x1_to_2x1( &BT,        B0,
				            B1,
			     /* *** */   /* ** */
				&BB,        B2,
				PLA_TOP );
   }
   
   return PLA_SUCCESS;
}


#define PLA_all  20201
#define PLA_row  20202
#define PLA_col  20203


PLA_Error PLA_Copy_aligned( PLA_Trans trans,
			    PLA_Obj   A, 
			    PLA_Obj   B )
/*----------------------------------------------------------------------------

Purpose :  Copy contents (not description) between linear algebra objects
           when both objected are properly aligned

IN     trans    option to transpose
IN     A        object to be copied
OUT    B        object into which to copy

----------------------------------------------------------------------------*/
{
   int
      rowA, colA,
      rowB, colB;

   PLA_Obj_owner( A, &rowA, &colA );
   
   // A is duplicated on all processors
   if ( rowA == PLA_ALL_ROWS && colA == PLA_ALL_COLS )
   {
      if ( PLA_Obj_is_owner( A ) &&
	   PLA_Obj_is_owner( B ) )
	 PLA_Local_copy_x( trans, A, B );
   }
   else 
   {
      PLA_Obj_owner( B, &rowB, &colB );

      // A is duplicated in a column of processors
      if ( rowA == PLA_ALL_ROWS && colA != PLA_ALL_COLS )
      {
	 // B is duplicated on all processors
	 if ( rowB == PLA_ALL_ROWS && colB == PLA_ALL_COLS )
	 {
	    PLA_Copy_comm_bcast( trans, A, B, PLA_row ); 
	 }
	 // B is duplicated in a column of processors
	 else if ( rowB == PLA_ALL_ROWS && colB != PLA_ALL_COLS )
	 {
	    if ( colA == colB )
	    {
	       if ( PLA_Temp_comm_row_rank( PLA_Obj_template( B ) ) == colB )
		  PLA_Local_copy_x( trans, A, B );
	    }
	    else	    
	       PLA_Copy_comm_send( trans, A, B, PLA_row );
	 }
	 // B is duplicated in a row of processors
	 else if ( rowB != PLA_ALL_ROWS && colB == PLA_ALL_COLS )
	 {
	    if ( PLA_Temp_comm_col_rank( PLA_Obj_template( B ) ) == rowB )
	       PLA_Copy_comm_bcast( trans, A, B, PLA_row );
	 }
	 // B is unduplicated
	 else
	 {
	    if ( colA == colB )
	    {
	       if ( PLA_Temp_comm_row_rank( PLA_Obj_template( B ) ) == colB &&
		    PLA_Temp_comm_col_rank( PLA_Obj_template( B ) ) == rowB )
		  PLA_Local_copy_x( trans, A, B );
	    }
	    else
	    {
	       if ( PLA_Temp_comm_col_rank( PLA_Obj_template( B ) ) == rowB )
		  PLA_Copy_comm_send( trans, A, B, PLA_row );
	    }
	 }
      }
      // A is duplicated in a row of processors
      else if ( rowA != PLA_ALL_ROWS && colA == PLA_ALL_COLS )
      {
	 // B is duplicated on all processors
	 if ( rowB == PLA_ALL_ROWS && colB == PLA_ALL_COLS )
	 {
	    PLA_Copy_comm_bcast( trans, A, B, PLA_col );
	 }
	 // B is duplicated in a column of processors
	 else if ( rowB == PLA_ALL_ROWS && colB != PLA_ALL_COLS )
	 {
	    if ( PLA_Temp_comm_row_rank( PLA_Obj_template( B ) ) == colB )
	       PLA_Copy_comm_bcast( trans, A, B, PLA_col );	    
	 }
	 // B is duplicated in a row of processors
	 else if ( rowB != PLA_ALL_ROWS && colB == PLA_ALL_COLS )
	 {
	    if ( rowA == rowB )
	    {
	       if ( PLA_Temp_comm_col_rank( PLA_Obj_template( B ) ) == rowB )
		  PLA_Local_copy_x( trans, A, B );
	    }
	    else	    
	       PLA_Copy_comm_send( trans, A, B, PLA_col );
	 }
	 // B is unduplicated
	 else
	 {
	    if ( rowA == rowB )
	    {
	       if ( PLA_Temp_comm_row_rank( PLA_Obj_template( B ) ) == colB &&
		    PLA_Temp_comm_col_rank( PLA_Obj_template( B ) ) == rowB )
		  PLA_Local_copy_x( trans, A, B );
	    }
	    else
	    {
	       if ( PLA_Temp_comm_row_rank( PLA_Obj_template( B ) ) == colB )
		  PLA_Copy_comm_send( trans, A, B, PLA_col );
	    }	    
	 }
      }
      // A is unduplicated
      else
      {
	 // B is duplicated on all processors
	 if ( rowB == PLA_ALL_ROWS && colB == PLA_ALL_COLS )
	 {
	    PLA_Copy_comm_bcast( trans, A, B, PLA_all );
	 }
	 // B is duplicated in a column of processors
	 else if ( rowB == PLA_ALL_ROWS && colB != PLA_ALL_COLS )
	 {
	    if ( colA == colB )
	    {
	       if ( PLA_Temp_comm_row_rank( PLA_Obj_template( B ) ) == colB )
		  PLA_Copy_comm_bcast( trans, A, B, PLA_col );
	    }
	    else
	    {       
	       PLA_Obj
		  temp;
	       
	       PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, A,
					   rowA, colB,
					   &temp );
	       
	       if ( PLA_Temp_comm_col_rank( PLA_Obj_template( A ) ) == rowA )
		  PLA_Copy_comm_send( trans, A, temp, PLA_row );

	       if ( PLA_Temp_comm_row_rank( PLA_Obj_template( B ) ) == colB )
		  PLA_Copy_comm_bcast( trans, temp, B, PLA_col );
	       
	       PLA_Obj_free( &temp );
	    }
	 }
	 // B is duplicated in a row of processors
	 else if ( rowB != PLA_ALL_ROWS && colB == PLA_ALL_COLS )
	 {
	    if ( rowA == rowB )
	    {
	       if ( PLA_Temp_comm_col_rank( PLA_Obj_template( B ) ) == rowB )
		  PLA_Copy_comm_bcast( trans, A, B, PLA_row );
	    }
	    else
	    {
	       PLA_Obj
		  temp;
	       
	       PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, A,
					   rowB, colA,
					   &temp );
	       
	       if ( PLA_Temp_comm_row_rank( PLA_Obj_template( A ) ) == colA )
		  PLA_Copy_comm_send( trans, A, temp, PLA_col );

	       if ( PLA_Temp_comm_col_rank( PLA_Obj_template( B ) ) == rowB )
		  PLA_Copy_comm_bcast( trans, temp, B, PLA_row );
	       
	       PLA_Obj_free( &temp );
	    }
	 }
	 // B is unduplicated
	 else
	 {
	    if ( PLA_Obj_owner_all( A ) == PLA_Obj_owner_all( B ) )
	       PLA_Local_copy_x( trans, A, B );
	    else	    
	       PLA_Copy_comm_send( trans, A, B, PLA_all );
	 }
      }
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Copy_comm_bcast( PLA_Trans trans,
			       PLA_Obj   A, 
			       PLA_Obj   B,
			       int       opt )
/*----------------------------------------------------------------------------

Purpose :  Copy contents (not description) between linear algebra objects
           when both objected are properly aligned, 
           perform communication with broadcast

IN     trans    option to transpose
IN     A        object to be copied
OUT    B        object into which to copy
IN     opt      option for which communnicator:  all, row, col

----------------------------------------------------------------------------*/
{
   int     
      root,
      rowA, colA,
      rowB, colB;

   PLA_Obj
      packed;

   MPI_Comm
      comm;

   MPI_Datatype
      datatype;

   root = PLA_Obj_owner( A, &rowA, &colA );
   PLA_Obj_owner( B, &rowB, &colB );

   datatype = PLA_MPI_Datatype( PLA_Obj_datatype( B ) );
  
   if ( opt == PLA_all )
   {
      comm = PLA_Temp_comm_all( PLA_Obj_template( B ) );
   }
   else if ( opt == PLA_row )
   {
      comm = PLA_Temp_comm_row( PLA_Obj_template( B ) );
      root = colA;
   }
   else if ( opt == PLA_col )
   {
      comm = PLA_Temp_comm_col( PLA_Obj_template( B ) );
      root = rowA;
   }
   
   PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, B,
			       rowB, colB,
			       &packed );
   
   if ( PLA_Obj_is_owner( A ) )
      PLA_Local_copy_x( trans, A, packed );

   MPI_Bcast( (void *) packed.base->buffer,
              PLA_Obj_length( packed ) * PLA_Obj_width( packed ),
	      datatype, root, comm );
    
   if ( PLA_Obj_is_owner( B ) )
      PLA_Local_copy_x( PLA_NO_TRANSPOSE, packed, B );

   PLA_Obj_free( &packed );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Copy_comm_send( PLA_Trans trans,
			      PLA_Obj   A, 
			      PLA_Obj   B,
			      int       opt )
/*----------------------------------------------------------------------------

Purpose :  Copy contents (not description) between linear algebra objects
           when both objected are properly aligned, 
           perform communication with sending

IN     trans    option to transpose
IN     A        object to be copied
OUT    B        object into which to copy
IN     opt      option for which communnicator:  all, row, col

----------------------------------------------------------------------------*/
{
   int     
      srce, dest,
      rowA, colA,
      rowB, colB;

   PLA_Obj
      A_packed, B_packed;

   MPI_Comm
      comm;

   MPI_Datatype
      datatype;

   MPI_Status
      status;

   srce = PLA_Obj_owner( A, &rowA, &colA );
   dest = PLA_Obj_owner( B, &rowB, &colB );

   datatype = PLA_MPI_Datatype( PLA_Obj_datatype( B ) );

   if ( opt == PLA_all )
   {
      comm = PLA_Temp_comm_all( PLA_Obj_template( B ) );
   }
   else if ( opt == PLA_row )
   {
      comm = PLA_Temp_comm_row( PLA_Obj_template( B ) );
      srce = colA;
      dest = colB;
   }
   else if ( opt == PLA_col )
   {
      comm = PLA_Temp_comm_col( PLA_Obj_template( B ) );
      srce = rowA;
      dest = rowB;
   }
   
   PLA_Mscalar_create_conf_to( trans, A,
			       rowA, colA,
			       &A_packed );
   PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, B,
			       rowB, colB,
			       &B_packed );
   
   if ( PLA_Obj_is_owner( A ) )
   {
      PLA_Local_copy_x( trans, A, A_packed );  

      MPI_Send( (void *) A_packed.base->buffer, 
		PLA_Obj_length( A_packed ) * PLA_Obj_width( A_packed ),
		datatype, dest, 0, comm );
   }
   if ( PLA_Obj_is_owner( B ) )
   {
      MPI_Recv( (void *) B_packed.base->buffer,
		PLA_Obj_length( B_packed ) * PLA_Obj_width( B_packed ),
		datatype, srce, 0, comm, &status );

      PLA_Local_copy_x( PLA_NO_TRANSPOSE, B_packed, B );
   }

   PLA_Obj_free( &A_packed );
   PLA_Obj_free( &B_packed );

   return PLA_SUCCESS;
}
