/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


PLA_Error PLA_Reduce_from_to( PLA_Op, PLA_Trans, PLA_Obj, PLA_Obj, PLA_Obj );
PLA_Error PLA_Reduce_stride( PLA_Op, PLA_Obj, PLA_Obj );
PLA_Error PLA_Reduce_aligned( PLA_Op, PLA_Obj, PLA_Obj );


PLA_Error PLA_Reduce( PLA_Obj A, 
		      PLA_Obj B )
/*----------------------------------------------------------------------------

Purpose :  Reduce contents using summation in the duplicated object
           given by A and overwrite B with the result

IN     A        object to be reduced
OUT    B        object into which to reduce

----------------------------------------------------------------------------*/
{
   PLA_Reduce_x( PLA_SUM, PLA_NO_TRANSPOSE, PLA_ONE, A, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Reduce_x( PLA_Op    op,
			PLA_Trans trans, 
			PLA_Obj   alpha,
			PLA_Obj   A, 
			PLA_Obj   B )
/*----------------------------------------------------------------------------

Purpose :  Reduce contents using summation in the duplicated object
           given by A and overwrite B with the result

IN     op       reduction operation
IN     trans    option for transpose
IN     alpha    scaling factor for target object
IN     A        object to be reduced
OUT    B        object into which to reduce

----------------------------------------------------------------------------*/
{
   if ( PLA_Error_check( ) )
      PLA_Reduce_x_check( op, trans, alpha, A, B );

   if ( PLA_Obj_is_valid_state( A ) &&
        PLA_Obj_is_valid_state( B ) )
      PLA_Reduce_from_to( op, trans, alpha, A, B );
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Reduce_from_to( PLA_Op    op,
			      PLA_Trans trans,
			      PLA_Obj   alpha,
			      PLA_Obj   A,
			      PLA_Obj   B )
/*----------------------------------------------------------------------------

Purpose :  Reduce contents using summation in the duplicated object
           given by A and overwrite B with the result

IN     op       reduction operation
IN     trans    option for transpose
IN     alpha    scaling factor for target object
IN     A        object to be reduced
OUT    B        object into which to reduce

----------------------------------------------------------------------------*/
{
   PLA_Obj
      temp, alpha_cpy;

   switch ( PLA_Obj_objtype( A ) )
   {
      case PLA_MATRIX:
	 return PLA_FAILURE;
	 
      case PLA_MROWVECTOR:
	 return PLA_FAILURE;
	 
      case PLA_MCOLVECTOR:
	 return PLA_FAILURE;
	 
      case PLA_MSCALAR:
	 if ( PLA_Obj_owner_row( A ) != PLA_ALL_ROWS &&
	      PLA_Obj_owner_col( A ) != PLA_ALL_COLS )
	    return PLA_FAILURE;
	 
	 PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, A,
				     PLA_INHERIT, PLA_INHERIT,
				     &temp );
	 break;
      case PLA_PMROWVECTOR:
	 if ( PLA_Obj_owner_row( A ) != PLA_ALL_ROWS )
	    return PLA_FAILURE;

	 PLA_Pmrowvector_create_conf_to( PLA_NO_TRANSPOSE, A,
					 PLA_INHERIT, PLA_INHERIT,
					 &temp );
	 break;
      case PLA_PMCOLVECTOR:	
	 if ( PLA_Obj_owner_col( A ) != PLA_ALL_COLS )
	    return PLA_FAILURE;

	 PLA_Pmcolvector_create_conf_to( PLA_NO_TRANSPOSE, A,
					 PLA_INHERIT, PLA_INHERIT,
					 &temp );
	 break;
      case PLA_CONSTANT:
	 PLA_Abort( "PLA_Reduce: cannot reduce a constant", __LINE__, __FILE__ );
         break;
      default:
         PLA_Abort( "Undefined objtype", __LINE__, __FILE__ );
         break;
   }
   
   PLA_Copy( A, temp );
   PLA_Scalar_copy( alpha, &alpha_cpy );

   PLA_Reduce_stride( op, alpha_cpy, temp );

   PLA_Copy_x( trans, temp, B );

   PLA_Obj_free( &temp );
   PLA_Obj_free( &alpha_cpy );

   return PLA_SUCCESS;
}


PLA_Error PLA_Reduce_stride( PLA_Op    op,
			     PLA_Obj   alpha,
			     PLA_Obj   A )
/*----------------------------------------------------------------------------

Purpose :  Reduce contents using summation in the duplicated object
           given by A with shifting

IN     op       reduction operation
IN     alpha    scaling factor for target object
OUT    A        object to be reduced

----------------------------------------------------------------------------*/
{
   int 
      bm, bn;

   PLA_Obj
      AL, AR,    A0, A1, A2, 
      AT1,           A01,
      AB1,           A11,
                     A21;
   
   PLA_Part_1x2( A,
		 &AL,  &AR,
		 0, PLA_LEFT );
   
   while ( PLA_Obj_width( AL ) < PLA_Obj_width( A ) )
   {
      bn = PLA_Obj_split_size( AR, PLA_LEFT );
      
      PLA_Repart_1x2_to_1x3( AL, /**/ AR,   &A0, /**/ &A1, &A2,
                             bn, PLA_RIGHT );

      /*------------------------------------------------------------*/
      
      PLA_Part_2x1( A1,   &AT1,
		          &AB1,
                    0, PLA_TOP );

      while ( PLA_Obj_length( AT1 ) < PLA_Obj_length( A1 ) )
      {
	 bm = PLA_Obj_split_size( AB1, PLA_TOP );
	 
	 PLA_Repart_2x1_to_3x1(  AT1,        &A01,
			      /* *** */   /* **** */
				             &A11,
				 AB1,        &A21,
				 bm, PLA_BOTTOM );
	 
	 /*------------------------------------------------------------*/
	 
	 PLA_Reduce_aligned( op, alpha, A11 );
	 
	 /*------------------------------------------------------------*/
	 
	 PLA_Cont_with_3x1_to_2x1( &AT1,        A01,
				                A11,
				/* **** */   /* *** */
				   &AB1,        A21,
				   PLA_TOP );
      }
      
      /*------------------------------------------------------------*/
      
      PLA_Cont_with_1x3_to_1x2( &AL, /**/ &AR,   A0, A1, /**/ A2,
                                PLA_LEFT );
   }
   
   return PLA_SUCCESS;
}


PLA_Error PLA_Reduce_aligned( PLA_Op    op,
			      PLA_Obj   alpha,
			      PLA_Obj   A )
/*----------------------------------------------------------------------------

Purpose :  Reduce contents using summation in the duplicated object
           given by A with shifting and alignment

IN     op       reduction operation
IN     alpha    scaling factor for target object
OUT    A        object to be reduced

----------------------------------------------------------------------------*/
{
   int
      rowA, colA;
    
   PLA_Obj
      packed, packed_reduce;

   MPI_Comm
      comm;

   MPI_Datatype
      datatype;

   MPI_Op
      oper;

   PLA_Obj_owner( A, &rowA, &colA );

   datatype = PLA_MPI_Datatype( PLA_Obj_datatype( A ) );
   oper = PLA_MPI_Op( op );

   if ( rowA == PLA_ALL_ROWS && colA == PLA_ALL_COLS )
   {
      comm = PLA_Temp_comm_all( PLA_Obj_template( A ) );
   }
   else if ( rowA == PLA_ALL_ROWS && colA != PLA_ALL_COLS )
   {
      comm = PLA_Temp_comm_col( PLA_Obj_template( A ) );
   }
   else if ( rowA != PLA_ALL_ROWS && colA == PLA_ALL_COLS )
   {
      comm = PLA_Temp_comm_row( PLA_Obj_template( A ) );
   }
   else
   {
      comm = PLA_Temp_comm_all( PLA_Obj_template( A ) );
      PLA_Abort( "PLA_Reduce: cannot reduce unduplicated object", __LINE__, __FILE__ );
   }

   PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, A,
                               rowA, colA,
                               &packed );
   PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, A,
                               rowA, colA,
                               &packed_reduce );

   if ( PLA_Obj_is_owner( A ) )
   {
      PLA_Local_copy_x( PLA_NO_TRANSPOSE, A, packed );

      MPI_Allreduce( packed.base->buffer, packed_reduce.base->buffer,
		     PLA_Obj_length( A ) * PLA_Obj_width( A ),
		     datatype, oper, comm );
      
      PLA_Local_copy_x( PLA_NO_TRANSPOSE, packed_reduce, A );
   }

   PLA_Obj_free( &packed );
   PLA_Obj_free( &packed_reduce );
   
   return PLA_SUCCESS;
}
