/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "PLA.h"


MPI_Datatype PLA_MPI_Datatype( PLA_Datatype datatype )
/*----------------------------------------------------------------------------

Purpose :  Return the corresponding MPI datatype

IN     datatype  reduction operation

Return :   MPI datatype

----------------------------------------------------------------------------*/
{
   MPI_Datatype dtype;

   switch( datatype )
   {
      case PLA_FLOAT:
	 dtype = MPI_FLOAT;
	 break;
      case PLA_DOUBLE:
	 dtype = MPI_DOUBLE;
	 break;
      case PLA_COMPLEX:
	 dtype = MPI_COMPLEX;
	 break;
      case PLA_DOUBLE_COMPLEX:
	 dtype = MPI_DOUBLE_COMPLEX;
	 break;
      case PLA_INT:
	 dtype = MPI_INT;
	 break;
      default:
	 dtype = MPI_DOUBLE;
	 PLA_Abort( "Undefined datatype", __LINE__, __FILE__ );
	 break;
   }

   return dtype;
}


MPI_Op PLA_MPI_Op( PLA_Op op )
/*----------------------------------------------------------------------------

Purpose :  Return the corresponding MPI operation

IN     op        reduction operation

Return :   MPI operation

----------------------------------------------------------------------------*/
{
   MPI_Op oper;

   switch ( op )
   {
      case PLA_SUM:
	 oper = MPI_SUM;
	 break;
      case PLA_PROD:
	 oper = MPI_PROD;
	 break;
      case PLA_MAX:
	 oper = MPI_MAX;
	 break;
      case PLA_MIN:
	 oper = MPI_MIN;
	 break;
      default:
	 oper = MPI_SUM;
	 PLA_Abort( "Undefined reduction operation", __LINE__, __FILE__ );
	 break;
   }

   return oper;
}


PLA_Error PLA_Scalar_copy( PLA_Obj alpha, PLA_Obj *alpha_cpy )
/*----------------------------------------------------------------------------

Purpose :  Create a constant from the given scalar object

IN     alpha        scalar object
OUT    alpha_cpy    constant

----------------------------------------------------------------------------*/
{
   if ( PLA_Obj_objtype( alpha ) == PLA_CONSTANT )
   {
      dcomplex *temp_z = PLA_DOUBLE_COMPLEX_PTR( alpha );

      PLA_Constant_complex_create( temp_z->real, temp_z->imag,
				   alpha_cpy );
   }
   else
   {
      PLA_Mscalar_create_conf_to( PLA_NO_TRANSPOSE, alpha,
				  PLA_ALL_ROWS, PLA_ALL_COLS,
				  alpha_cpy );

      PLA_Copy( alpha, *alpha_cpy );
   }

   return PLA_SUCCESS;
}
