/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

extern fla_gemm_t*      fla_gemm_cntl_blas;

extern fla_gemm_t*      fla_gemm_cntl_pb_bb;
extern fla_gemm_t*      fla_gemm_cntl_bp_bb;
extern fla_gemm_t*      fla_gemm_cntl_ip_bb;

extern fla_gemm_t*      fla_gemm_cntl_mp_ip;
extern fla_gemm_t*      fla_gemm_cntl_mp_ip_bb;
extern fla_gemm_t*      fla_gemm_cntl_op_bp;
extern fla_gemm_t*      fla_gemm_cntl_op_bp_bb;
extern fla_gemm_t*      fla_gemm_cntl_pm_ip;
extern fla_gemm_t*      fla_gemm_cntl_pm_ip_bb;
extern fla_gemm_t*      fla_gemm_cntl_op_pb;
extern fla_gemm_t*      fla_gemm_cntl_op_pb_bb;
extern fla_gemm_t*      fla_gemm_cntl_mp_pb;
extern fla_gemm_t*      fla_gemm_cntl_mp_pb_bb;
extern fla_gemm_t*      fla_gemm_cntl_pm_bp;
extern fla_gemm_t*      fla_gemm_cntl_pm_bp_bb;

extern fla_gemm_t*      fla_gemm_cntl_mm_pm;
extern fla_gemm_t*      fla_gemm_cntl_mm_pm_ip;
extern fla_gemm_t*      fla_gemm_cntl_mm_pm_ip_bb;
extern fla_gemm_t*      fla_gemm_cntl_mm_mp;
extern fla_gemm_t*      fla_gemm_cntl_mm_mp_ip;
extern fla_gemm_t*      fla_gemm_cntl_mm_mp_ip_bb;
extern fla_gemm_t*      fla_gemm_cntl_mm_op;
extern fla_gemm_t*      fla_gemm_cntl_mm_op_bp;
extern fla_gemm_t*      fla_gemm_cntl_mm_op_bp_bb;

extern fla_blocksize_t* fla_gemm_var1_bsize;
extern fla_blocksize_t* fla_gemm_var3_bsize;
extern fla_blocksize_t* fla_gemm_var5_bsize;

FLA_Error FLA_Gemm( FLA_Trans transa, FLA_Trans transb, FLA_Obj alpha, FLA_Obj A, FLA_Obj B, FLA_Obj beta, FLA_Obj C )
{
  FLA_Error    r_val = FLA_SUCCESS;
#ifdef FLA_ENABLE_BLAS3_FRONT_END_CNTL_TREES
  FLA_Datatype datatype;
  int          m_C, n_C, k_AB;
  int          FLA_GEMM_VAR1_BLOCKSIZE;
  int          FLA_GEMM_VAR3_BLOCKSIZE;
  int          FLA_GEMM_VAR5_BLOCKSIZE;
#endif

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Bool     check_params;
  
  // Check parameters.
  if ( FLA_Check_parameters() )
    FLA_Gemm_check( transa, transb, alpha, A, B, beta, C );

  // Temporarily disable parameter checking.
  check_params = FLA_Check_parameters_set( FALSE );
#endif

#ifdef FLA_ENABLE_BLAS3_FRONT_END_CNTL_TREES
  // Determine the datatype of the operation.
  datatype = FLA_Obj_datatype( A );

  // Extract the appropriate blocksize for the given datatype.
  FLA_GEMM_VAR1_BLOCKSIZE = FLA_Blocksize_extract( datatype, fla_gemm_var1_bsize );
  FLA_GEMM_VAR3_BLOCKSIZE = FLA_Blocksize_extract( datatype, fla_gemm_var3_bsize );
  FLA_GEMM_VAR5_BLOCKSIZE = FLA_Blocksize_extract( datatype, fla_gemm_var5_bsize );

  // Determine the m and n dimensions from C.
  m_C = FLA_Obj_length( C );
  n_C = FLA_Obj_width( C );

  // Determine the k dimension of A and B.
  if ( transa == FLA_NO_TRANSPOSE )
    k_AB = FLA_Obj_width( A );
  else
    k_AB = FLA_Obj_length( A );

  // Invoke FLA_Gemm_internal() with the appropriate control tree.
  if      ( FLA_GEMM_VAR1_BLOCKSIZE < m_C && FLA_GEMM_VAR3_BLOCKSIZE < n_C && FLA_GEMM_VAR5_BLOCKSIZE < k_AB )
  {
    r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_mm_op );
    //r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_mm_mp );
    //r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_mm_pm );
  }
  else if ( m_C <= FLA_GEMM_VAR1_BLOCKSIZE && FLA_GEMM_VAR3_BLOCKSIZE < n_C && FLA_GEMM_VAR5_BLOCKSIZE < k_AB )
  {
    r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_pm_bp );
    //r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_pm_ip );
  }
  else if ( FLA_GEMM_VAR1_BLOCKSIZE < m_C && n_C <= FLA_GEMM_VAR3_BLOCKSIZE && FLA_GEMM_VAR5_BLOCKSIZE < k_AB )
  {
    r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_mp_pb );
    //r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_mp_ip );
  }
  else if ( FLA_GEMM_VAR1_BLOCKSIZE < m_C && FLA_GEMM_VAR3_BLOCKSIZE < n_C && k_AB <= FLA_GEMM_VAR5_BLOCKSIZE )
  {
    r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_op_bp );
    //r_val = FLA_Gemm_internal( transa, transb, alpha, A, B, beta, C, fla_gemm_cntl_op_pb );
  }
  else
  {
    r_val = FLA_Gemm_external( transa, transb, alpha, A, B, beta, C );
  }
#else
  r_val = FLA_Gemm_external( transa, transb, alpha, A, B, beta, C );
#endif

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  // Restore parameter checking to its previous status.
  FLA_Check_parameters_set( check_params );
#endif

  return r_val;
}

void FLA_F2C( fla_gemm_f ) 
     ( F_INT *transa, F_INT *transb, F_INT *alpha, F_INT *A, F_INT *B, F_INT *beta, F_INT *C, F_INT *IERROR )
{
  *IERROR = FLA_Gemm( *( ( FLA_Trans * ) transa ),
                      *( ( FLA_Trans * ) transb ),
                      *( ( FLA_Obj   * ) alpha  ),
                      *( ( FLA_Obj   * ) A      ),
                      *( ( FLA_Obj   * ) B      ),
                      *( ( FLA_Obj   * ) beta   ),
                      *( ( FLA_Obj   * ) C      ) );
}

