/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

extern fla_trmm_t*      fla_trmm_cntl_bp;
extern fla_trmm_t*      fla_trmm_cntl_mp;
extern fla_trmm_t*      fla_trmm_cntl_mm;
extern fla_blocksize_t* fla_trmm_var2_bsize;
extern fla_blocksize_t* fla_trmm_var3_bsize;

FLA_Error FLA_Trmm( FLA_Side side, FLA_Uplo uplo, FLA_Trans trans, FLA_Diag diag, FLA_Obj alpha, FLA_Obj A, FLA_Obj B )
{
  FLA_Error    r_val = FLA_SUCCESS;
#ifdef FLA_ENABLE_BLAS3_FRONT_END_CNTL_TREES
  FLA_Datatype datatype;
  int          m_A, nrhs_B;
  int          FLA_TRMM_VAR2_BLOCKSIZE;
  int          FLA_TRMM_VAR3_BLOCKSIZE;
#endif

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  FLA_Bool     check_params;

  // Check parameters.
  if ( FLA_Check_parameters() )
    FLA_Trmm_check( side, uplo, trans, diag, alpha, A, B );

  // Temporarily disable parameter checking.
  check_params = FLA_Check_parameters_set( FALSE );
#endif

#ifdef FLA_ENABLE_BLAS3_FRONT_END_CNTL_TREES
  // Determine the datatype of the operation.
  datatype = FLA_Obj_datatype( A );

  // Extract the appropriate blocksize for the given datatype.
  FLA_TRMM_VAR2_BLOCKSIZE = FLA_Blocksize_extract( datatype, fla_trmm_var2_bsize );
  FLA_TRMM_VAR3_BLOCKSIZE = FLA_Blocksize_extract( datatype, fla_trmm_var3_bsize );

  // Determine the dimensions of A.
  m_A = FLA_Obj_length( A );

  // Determine the number of right-hand sides of B.
  if ( side == FLA_LEFT )
    nrhs_B = FLA_Obj_width( B );
  else
    nrhs_B = FLA_Obj_length( B );

  // Invoke FLA_Trmm_internal() with the appropriate control tree.
  if      ( FLA_TRMM_VAR2_BLOCKSIZE < m_A && FLA_TRMM_VAR3_BLOCKSIZE < nrhs_B )
  {
    r_val = FLA_Trmm_internal( side, uplo, trans, diag, alpha, A, B, fla_trmm_cntl_mm );
  }
  else if ( FLA_TRMM_VAR2_BLOCKSIZE < m_A && nrhs_B <= FLA_TRMM_VAR3_BLOCKSIZE )
  {
    r_val = FLA_Trmm_internal( side, uplo, trans, diag, alpha, A, B, fla_trmm_cntl_mp );
  }
  else if ( m_A <= FLA_TRMM_VAR2_BLOCKSIZE && FLA_TRMM_VAR3_BLOCKSIZE < nrhs_B )
  {
    r_val = FLA_Trmm_internal( side, uplo, trans, diag, alpha, A, B, fla_trmm_cntl_bp );
  }
  else if ( m_A <= FLA_TRMM_VAR2_BLOCKSIZE && nrhs_B <= FLA_TRMM_VAR3_BLOCKSIZE )
  {
    r_val = FLA_Trmm_external( side, uplo, trans, diag, alpha, A, B );
  }
#else
  r_val = FLA_Trmm_external( side, uplo, trans, diag, alpha, A, B );
#endif 

#ifdef FLA_ENABLE_INTERNAL_ERROR_CHECKING
  // Restore parameter checking to its previous status.
  FLA_Check_parameters_set( check_params );
#endif
  
  return r_val;
}

void FLA_F2C( fla_trmm_f ) 
     ( F_INT *side, F_INT *uplo, F_INT *trans, F_INT *diag, F_INT *alpha, F_INT *A, F_INT *B, F_INT *IERROR )
{
  *IERROR = FLA_Trmm( *( ( FLA_Side  * ) side  ),
                      *( ( FLA_Uplo  * ) uplo  ),
                      *( ( FLA_Trans * ) trans ),
                      *( ( FLA_Diag  * ) diag  ),
                      *( ( FLA_Obj   * ) alpha ),
                      *( ( FLA_Obj   * ) A     ),
                      *( ( FLA_Obj   * ) B     ) );
}

