*
*  libflame
*  An object-based infrastructure for developing high-performance
*  dense linear algebra libraries.
*
*  Copyright (C) 2008, The University of Texas
*
*  libflame is free software; you can redistribute it and/or modify
*  it under the terms of the GNU Lesser General Public License as
*  published by the Free Software Foundation; either version 2.1 of
*  the License, or (at your option) any later version.
*
*  libflame is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
*  Lesser General Public License for more details.
*
*  You should have received a copy of the GNU Lesser General Public
*  License along with libflame; if you did not receive a copy, see
*  http://www.gnu.org/licenses/.
*
*  For more information, please contact us at flame@cs.utexas.edu or
*  send mail to:
*
*  Field G. Van Zee and/or
*  Robert A. van de Geijn
*  The University of Texas at Austin
*  Department of Computer Sciences
*  1 University Station C0500
*  Austin TX 78712
*
      SUBROUTINE DTRSM ( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA,
     $                   B, LDB )
*     .. Scalar Arguments ..
      CHARACTER*1        SIDE, UPLO, TRANSA, DIAG
      INTEGER            M, N, LDA, LDB
      DOUBLE PRECISION   ALPHA
*     .. Array Arguments ..
      DOUBLE PRECISION   A( LDA, * ), B( LDB, * )
*     ..
*
*  Purpose
*  =======
*
*  DTRSM  solves one of the matrix equations
*
*     op( A )*X = alpha*B,   or   X*op( A ) = alpha*B,
*
*  where alpha is a scalar, X and B are m by n matrices, A is a unit, or
*  non-unit,  upper or lower triangular matrix  and  op( A )  is one  of
*
*     op( A ) = A   or   op( A ) = A'.
*
*  The matrix X is overwritten on B.
*
*  Parameters
*  ==========
*
*  SIDE   - CHARACTER*1.
*           On entry, SIDE specifies whether op( A ) appears on the left
*           or right of X as follows:
*
*              SIDE = 'L' or 'l'   op( A )*X = alpha*B.
*
*              SIDE = 'R' or 'r'   X*op( A ) = alpha*B.
*
*           Unchanged on exit.
*
*  UPLO   - CHARACTER*1.
*           On entry, UPLO specifies whether the matrix A is an upper or
*           lower triangular matrix as follows:
*
*              UPLO = 'U' or 'u'   A is an upper triangular matrix.
*
*              UPLO = 'L' or 'l'   A is a lower triangular matrix.
*
*           Unchanged on exit.
*
*  TRANSA - CHARACTER*1.
*           On entry, TRANSA specifies the form of op( A ) to be used in
*           the matrix multiplication as follows:
*
*              TRANSA = 'N' or 'n'   op( A ) = A.
*
*              TRANSA = 'T' or 't'   op( A ) = A'.
*
*              TRANSA = 'C' or 'c'   op( A ) = A'.
*
*           Unchanged on exit.
*
*  DIAG   - CHARACTER*1.
*           On entry, DIAG specifies whether or not A is unit triangular
*           as follows:
*
*              DIAG = 'U' or 'u'   A is assumed to be unit triangular.
*
*              DIAG = 'N' or 'n'   A is not assumed to be unit
*                                  triangular.
*
*           Unchanged on exit.
*
*  M      - INTEGER.
*           On entry, M specifies the number of rows of B. M must be at
*           least zero.
*           Unchanged on exit.
*
*  N      - INTEGER.
*           On entry, N specifies the number of columns of B.  N must be
*           at least zero.
*           Unchanged on exit.
*
*  ALPHA  - DOUBLE PRECISION.
*           On entry,  ALPHA specifies the scalar  alpha. When  alpha is
*           zero then  A is not referenced and  B need not be set before
*           entry.
*           Unchanged on exit.
*
*  A      - DOUBLE PRECISION array of DIMENSION ( LDA, k ), where k is m
*           when  SIDE = 'L' or 'l'  and is  n  when  SIDE = 'R' or 'r'.
*           Before entry  with  UPLO = 'U' or 'u',  the  leading  k by k
*           upper triangular part of the array  A must contain the upper
*           triangular matrix  and the strictly lower triangular part of
*           A is not referenced.
*           Before entry  with  UPLO = 'L' or 'l',  the  leading  k by k
*           lower triangular part of the array  A must contain the lower
*           triangular matrix  and the strictly upper triangular part of
*           A is not referenced.
*           Note that when  DIAG = 'U' or 'u',  the diagonal elements of
*           A  are not referenced either,  but are assumed to be  unity.
*           Unchanged on exit.
*
*  LDA    - INTEGER.
*           On entry, LDA specifies the first dimension of A as declared
*           in the calling (sub) program.  When  SIDE = 'L' or 'l'  then
*           LDA  must be at least  max( 1, m ),  when  SIDE = 'R' or 'r'
*           then LDA must be at least max( 1, n ).
*           Unchanged on exit.
*
*  B      - DOUBLE PRECISION array of DIMENSION ( LDB, n ).
*           Before entry,  the leading  m by n part of the array  B must
*           contain  the  right-hand  side  matrix  B,  and  on exit  is
*           overwritten by the solution matrix  X.
*
*  LDB    - INTEGER.
*           On entry, LDB specifies the first dimension of B as declared
*           in  the  calling  (sub)  program.   LDB  must  be  at  least
*           max( 1, m ).
*           Unchanged on exit.
*
*
*  Level 3 Blas routine.
*
*
*  -- Written on 8-February-1989.
*     Jack Dongarra, Argonne National Laboratory.
*     Iain Duff, AERE Harwell.
*     Jeremy Du Croz, Numerical Algorithms Group Ltd.
*     Sven Hammarling, Numerical Algorithms Group Ltd.
*
*
*     .. External Functions ..
      LOGICAL            LSAME
      EXTERNAL           LSAME
*     .. External Subroutines ..
      EXTERNAL           XERBLA
*     .. Intrinsic Functions ..
      INTRINSIC          MAX
*     .. Local Scalars ..
      LOGICAL            LSIDE, NOUNIT, UPPER
      INTEGER            I, INFO, J, NROWA, NCOLA
*     .. Parameters ..
      DOUBLE PRECISION   ONE         , ZERO
      PARAMETER        ( ONE = 1.0D+0, ZERO = 0.0D+0 )
*     ..        
*     .. Declare FLAME related stuff
      include "FLAMEf.h"
*     
      INTEGER            A_obj( FLA_OBJ_SIZE ),
     $                   B_obj( FLA_OBJ_SIZE ),
     $                   alpha_obj( FLA_OBJ_SIZE ),
     $                   iside,
     $                   iuplo,
     $                   itransa,
     $                   idiag,
     $                   datatype
      PARAMETER        ( datatype = FLA_DOUBLE )
*     ..
*     .. Executable Statements ..
*    
*     Initialize FLAME.
      CALL FLA_INIT_F( INFO )
*
*     Test the input parameters.
*
      LSIDE  = LSAME( SIDE  , 'L' )
      IF( LSIDE )THEN
         NROWA = M
         NCOLA = N
      ELSE
         NROWA = N
         NCOLA = M
      END IF
      NOUNIT = LSAME( DIAG  , 'N' )
      UPPER  = LSAME( UPLO  , 'U' )
*
      INFO   = 0
      IF(      ( .NOT.LSIDE                ).AND.
     $         ( .NOT.LSAME( SIDE  , 'R' ) )      )THEN
         INFO = 1
      ELSE IF( ( .NOT.UPPER                ).AND.
     $         ( .NOT.LSAME( UPLO  , 'L' ) )      )THEN
         INFO = 2
      ELSE IF( ( .NOT.LSAME( TRANSA, 'N' ) ).AND.
     $         ( .NOT.LSAME( TRANSA, 'T' ) ).AND.
     $         ( .NOT.LSAME( TRANSA, 'C' ) )      )THEN
         INFO = 3
      ELSE IF( ( .NOT.LSAME( DIAG  , 'U' ) ).AND.
     $         ( .NOT.LSAME( DIAG  , 'N' ) )      )THEN
         INFO = 4
      ELSE IF( M  .LT.0               )THEN
         INFO = 5
      ELSE IF( N  .LT.0               )THEN
         INFO = 6
      ELSE IF( LDA.LT.MAX( 1, NROWA ) )THEN
         INFO = 9
      ELSE IF( LDB.LT.MAX( 1, M     ) )THEN
         INFO = 11
      END IF
      IF( INFO.NE.0 )THEN
         CALL XERBLA( 'DTRSM ', INFO )
         RETURN
      END IF
*
*     Quick return if possible.
*
      IF( N.EQ.0 )
     $   RETURN
*
*     And when  alpha.eq.zero.
*
      IF( ALPHA.EQ.ZERO )THEN
         DO 20, J = 1, N
            DO 10, I = 1, M
               B( I, J ) = ZERO
   10       CONTINUE
   20    CONTINUE
         RETURN
      END IF
*     
*     Set FLAME "side" argument for A.
*
      IF     ( LSAME( SIDE, 'L') )THEN
         iside = FLA_LEFT
      ELSE IF( LSAME( SIDE, 'R') )THEN
         iside = FLA_RIGHT
      END IF
*     
*     Set FLAME "uplo" argument for A.
*    
      IF     ( LSAME( UPLO, 'L') )THEN
         iuplo = FLA_LOWER_TRIANGULAR
      ELSE IF( LSAME( UPLO, 'U') )THEN
         iuplo = FLA_UPPER_TRIANGULAR
      END IF
*
*     Set FLAME transposition argument for A.
*
      IF     ( LSAME( TRANSA, 'N' ) )THEN
         itransa = FLA_NO_TRANSPOSE
      ELSE IF( LSAME( TRANSA, 'T' ) )THEN
         itransa = FLA_TRANSPOSE
      ELSE IF( LSAME( TRANSA, 'C' ) )THEN
         itransa = FLA_CONJ_TRANSPOSE
      END IF
*     
*     Set FLAME unit/non-unit diagonal argument for A.
*    
      IF     ( LSAME( DIAG, 'N') )THEN
         idiag = FLA_NONUNIT_DIAG
      ELSE IF( LSAME( DIAG, 'U') )THEN
         idiag = FLA_UNIT_DIAG
      END IF
*
*     Create base object for A_obj and attach matrix buffer.
*
      CALL FLA_OBJ_CREATE_W_OUT_BUFF_F( datatype, NROWA, NCOLA, A_obj,
     $                                  INFO )
      CALL FLA_OBJ_ATTACH_BUFF_F( A, LDA, A_obj, INFO )
*
*     Create base object for B_obj and attach matrix buffer.
*
      CALL FLA_OBJ_CREATE_W_OUT_BUFF_F( datatype, M, N, B_obj,
     $                                  INFO )
      CALL FLA_OBJ_ATTACH_BUFF_F( B, LDB, B_obj, INFO )
*
*     Create base object for alpha_obj and attach scalar buffer.
*
      CALL FLA_OBJ_CREATE_W_OUT_BUFF_F( datatype, 1, 1, alpha_obj,
     $                                  INFO )
      CALL FLA_OBJ_ATTACH_BUFF_F( ALPHA, 1, alpha_obj, INFO )
*
*     Call FLAME trsm code.
*
      CALL FLA_TRSM_F( iside, iuplo, itransa, idiag, alpha, A, B, INFO )
*
*     Free A_obj base object (but leave A buffer alone).
*
      CALL FLA_OBJ_FREE_W_OUT_BUFF_F( A_obj, INFO )
*
*     Free B_obj base object (but leave B buffer alone).
*
      CALL FLA_OBJ_FREE_W_OUT_BUFF_F( B_obj, INFO )
*
*     Free alpha_obj base object (but leave alpha buffer alone).
*
      CALL FLA_OBJ_FREE_W_OUT_BUFF_F( alpha_obj, INFO )
*
      RETURN
*
*     End of DTRSM .
*
      END
