/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_QR_blk_var1( FLA_Obj A, FLA_Obj t, FLA_Obj S, fla_qr_t* cntl )
{
  FLA_Obj ATL,   ATR,      A00, A01, A02, 
          ABL,   ABR,      A10, A11, A12,
                           A20, A21, A22;

  FLA_Obj tT,              t0,
          tB,              t1,
                           t2;

  FLA_Obj SL,    SR,       S0,  S1,  W12;

  FLA_Obj W12T, W12B;

  FLA_Obj S1T,   S2B,      AB1;

  int b;

  FLA_Part_2x2( A,    &ATL, &ATR,
                      &ABL, &ABR,     0, 0, FLA_TL );

  FLA_Part_2x1( t,    &tT, 
                      &tB,            0, FLA_TOP );

  FLA_Part_1x2( S,    &SL,  &SR,      0, FLA_LEFT );

  while ( FLA_Obj_width ( ATL ) != FLA_Obj_width ( A ) &&
          FLA_Obj_length( ATL ) != FLA_Obj_length( A ) ){

    b = FLA_Determine_blocksize( ABR, FLA_BR, FLA_Cntl_blocksize( cntl ) );

    FLA_Repart_2x2_to_3x3( ATL, /**/ ATR,       &A00, /**/ &A01, &A02,
                        /* ************* */   /* ******************** */
                                                &A10, /**/ &A11, &A12,
                           ABL, /**/ ABR,       &A20, /**/ &A21, &A22,
                           b, b, FLA_BR );

    FLA_Repart_2x1_to_3x1( tT,                &t0, 
                        /* ** */            /* ****** */
                                              &t1, 
                           tB,                &t2,        b, FLA_BOTTOM );

    FLA_Repart_1x2_to_1x3( SL,  /**/ SR,        &S0, /**/ &S1, &W12,
                           b, FLA_RIGHT );

    /*------------------------------------------------------------*/

    /*
    S1T = FLA_Top_part( S1, b );
    */

    FLA_Part_2x1( S1,   &S1T, 
                        &S2B, b, FLA_TOP );

    /*
    [ U1, S1T ] = FLA_QR_internal( [ A11
                                     A21 ], t1, S1T );
    */

    FLA_Merge_2x1( A11,
                   A21,   &AB1 );

    FLA_QR_internal( AB1, t1, S1T, 
                     FLA_Cntl_sub_qr( cntl ) );


    if ( FLA_Obj_width( A12 ) > 0 ){

      /*
      W12T = FLA_Top_part( W12, b );
      */

      FLA_Part_2x1( W12,  &W12T, 
                          &W12B, b, FLA_TOP );

      /*
      U11 = trilu( A11 );
      U21 = A21;
      
      W12T = triu( S1T )' * ( U11' * A12 + U21' * A22 );
      */

      FLA_Copy( A12, W12T );
      FLA_Trmm_external( FLA_LEFT, FLA_LOWER_TRIANGULAR, 
                         FLA_TRANSPOSE, FLA_UNIT_DIAG,
                         FLA_ONE, A11, W12T );
      FLA_Gemm_external( FLA_TRANSPOSE, FLA_NO_TRANSPOSE,
                         FLA_ONE, A21, A22, FLA_ONE, W12T );
      FLA_Trmm_external( FLA_LEFT, FLA_UPPER_TRIANGULAR, 
                         FLA_TRANSPOSE, FLA_NONUNIT_DIAG, 
                         FLA_ONE, S1T, W12T );

      /*
        A22 = A22 - U21 * W12T;
        A12 = A12 - U11 * W12T;
      */

      FLA_Gemm_external( FLA_NO_TRANSPOSE, FLA_NO_TRANSPOSE, 
                         FLA_MINUS_ONE, A21, W12T, FLA_ONE, A22 );
      FLA_Trmm_external( FLA_LEFT, FLA_LOWER_TRIANGULAR, 
                         FLA_NO_TRANSPOSE, FLA_UNIT_DIAG,
                         FLA_MINUS_ONE, A11, W12T );
      FLA_Axpy( FLA_ONE, W12T, A12 );
    }

    /*------------------------------------------------------------*/

    FLA_Cont_with_3x3_to_2x2( &ATL, /**/ &ATR,       A00, A01, /**/ A02,
                                                     A10, A11, /**/ A12,
                            /* ************** */  /* ****************** */
                              &ABL, /**/ &ABR,       A20, A21, /**/ A22,
                              FLA_TL );

    FLA_Cont_with_3x1_to_2x1( &tT,                t0, 
                                                  t1, 
                            /* ** */           /* ****** */
                              &tB,                t2,     FLA_TOP );

    FLA_Cont_with_1x3_to_1x2( &SL,  /**/ &SR,        S0, S1, /**/ W12,
                              FLA_LEFT );

  }

  return FLA_SUCCESS;
}

