/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/

#include "FLAME.h"

FLA_Error FLA_QR_UT_Accum_T_unb_var2( FLA_Obj A, FLA_Obj t, FLA_Obj T )
{
  FLA_Obj ATL,   ATR,      A00,  a01,     A02, 
          ABL,   ABR,      a10t, alpha11, a12t,
                           A20,  a21,     A22;

  FLA_Obj tT,              t0,
          tB,              tau1,
                           t2;

  FLA_Part_2x1( t,    &tT, 
                      &tB,            0, FLA_TOP );

  FLA_Obj TTL,   TTR,      T00,  t01,     T02, 
          TBL,   TBR,      t10t, tau11,   t12t,
                           T20,  t21,     T22;

  FLA_Part_2x2( A,    &ATL, &ATR,
                      &ABL, &ABR,     0, 0, FLA_TL );

  FLA_Part_2x2( T,    &TTL, &TTR,
                      &TBL, &TBR,     0, 0, FLA_TL );

  while ( FLA_Obj_width ( ATL ) != FLA_Obj_width ( A )   &&
	  FLA_Obj_length( ATL ) != FLA_Obj_length( A )){

    FLA_Repart_2x2_to_3x3( ATL, /**/ ATR,       &A00,  /**/ &a01,     &A02,
                        /* ************* */   /* ************************** */
                                                &a10t, /**/ &alpha11, &a12t,
                           ABL, /**/ ABR,       &A20,  /**/ &a21,     &A22,
                           1, 1, FLA_BR );

    FLA_Repart_2x1_to_3x1( tT,                &t0, 
                        /* ** */            /* ****** */
                                              &tau1, 
                           tB,                &t2,        1, FLA_BOTTOM );

    FLA_Repart_2x2_to_3x3( TTL, /**/ TTR,       &T00,  /**/ &t01,     &T02,
                        /* ************* */   /* ************************** */
                                                &t10t, /**/ &tau11,   &t12t,
                           TBL, /**/ TBR,       &T20,  /**/ &t21,     &T22,
                           1, 1, FLA_BR );

    /*------------------------------------------------------------*/

    /* Update the current column by applying ( I - U inv(T') U' ) to it
    where U and T are the parts of these matrices that have been
    accumulated so far.
    
    w01 = - triu( inv(T00) )' * ( trilu( A00 )' * a01 + alpha11 * a10t' + A20' * a21 );  
    
    a01     = a01     + trilu( A00 ) * w01;
    alpha11 = alpha11 + a10t         * w01;
    a21     = a21     + A20          * w01;

    w01 is temporarily stored in t01
    */

    FLA_Copy( a01, t01 );
    FLA_Trmv( FLA_LOWER_TRIANGULAR, FLA_TRANSPOSE, FLA_UNIT_DIAG, 
	      A00, t01 );
    FLA_Axpy_t( FLA_TRANSPOSE, alpha11, a10t, t01 );
    FLA_Gemv( FLA_TRANSPOSE, FLA_ONE, A20, a21, FLA_ONE, t01 );
    FLA_Trsv( FLA_UPPER_TRIANGULAR, FLA_TRANSPOSE, FLA_NONUNIT_DIAG,
	      T00, t01 );
    FLA_Scal( FLA_MINUS_ONE, t01 );
    FLA_Dot_s( FLA_ONE, a10t, t01, FLA_ONE, alpha11 );
    FLA_Gemv( FLA_NO_TRANSPOSE, FLA_ONE, A20, t01, FLA_ONE, a21 );
    FLA_Trmv( FLA_LOWER_TRIANGULAR, FLA_NO_TRANSPOSE, FLA_UNIT_DIAG,
	      A00, t01 );
    FLA_Axpy( FLA_ONE, t01, a01 );

    /* Compute the current Householder vector (UT transform style)

    [ alpha11, ...
      a21, tau11 ] = Househ2_UT( alpha11, ...
                                 a21 );
    */
    FLA_Househ2_UT( alpha11,
                    a21, tau11 );

    /* tau1 = tau11 */
    FLA_Copy( tau11, tau1 );

    /* Update T
    t01  = a10t' + A20' * a21
    */

    FLA_Copy_t( FLA_TRANSPOSE, a10t, t01 );
    FLA_Gemv( FLA_TRANSPOSE, FLA_ONE, A20, a21, FLA_ONE, t01 );

    /*------------------------------------------------------------*/

    FLA_Cont_with_3x3_to_2x2( &ATL, /**/ &ATR,       A00,  a01,     /**/ A02,
                                                     a10t, alpha11, /**/ a12t,
                            /* ************** */  /* ************************ */
                              &ABL, /**/ &ABR,       A20,  a21,     /**/ A22,
                              FLA_TL );

    FLA_Cont_with_3x1_to_2x1( &tT,                t0, 
                                                  tau1, 
                            /* ** */           /* ****** */
                              &tB,                t2,     FLA_TOP );

    FLA_Cont_with_3x3_to_2x2( &TTL, /**/ &TTR,       T00,  t01,     /**/ T02,
                                                     t10t, tau11, /**/ t12t,
                            /* ************** */  /* ************************ */
                              &TBL, /**/ &TBR,       T20,  t21,     /**/ T22,
                              FLA_TL );

  }

  return FLA_SUCCESS;
}



