/*
   libflame
   An object-based infrastructure for developing high-performance
   dense linear algebra libraries.

   Copyright (C) 2008, The University of Texas

   libflame is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   libflame is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with libflame; if you did not receive a copy, see
   http://www.gnu.org/licenses/.

   For more information, please contact us at flame@cs.utexas.edu or
   send mail to:

   Field G. Van Zee and/or
   Robert A. van de Geijn
   The University of Texas at Austin
   Department of Computer Sciences
   1 University Station C0500
   Austin TX 78712
*/
#include "FLAME.h"

#define FLA_ALG_REFERENCE 0
#define FLA_ALG_BLOCKED   1
#define FLA_ALG_UNBLOCKED 2

#define MATRIX_SIZE 1000


void time_LU_piv(
                  int variant, int type, int nrepeats, int percent_B, 
                  int nb_alg,
                  FLA_Obj A, FLA_Obj A_ref, FLA_Obj p,
                  double *dtime, double *diff, double *gflops );


void time_LU_piv(
                  int variant, int type, int nrepeats, int percent_B, int nb_alg,
                  FLA_Obj A, FLA_Obj A_ref, FLA_Obj p,
                  double *dtime, double *diff, double *gflops )
{
  int
    irep, m_A, m_B, m_E;

  double
    dtime_old = 1.0e9;

  FLA_Obj
    A_old, B, C, D, E, p_B, p_D, p_E, Ls, y, y_T, y_B, x;

  fla_blocksize_t*
    bp;
  fla_lu_t*
    cntl_lu_var;
  fla_trsm_t*
    cntl_trsm_blas;
  fla_gemm_t*
    cntl_gemm_blas;

  bp             = FLA_Blocksize_create( nb_alg, nb_alg, nb_alg, nb_alg );
  cntl_trsm_blas = FLA_Cntl_trsm_obj_create( FLA_SUBPROBLEM, NULL, NULL, NULL );
  cntl_gemm_blas = FLA_Cntl_gemm_obj_create( FLA_SUBPROBLEM, NULL, NULL );
  cntl_lu_var    = FLA_Cntl_lu_obj_create( variant, bp, cntl_gemm_blas, cntl_gemm_blas, cntl_gemm_blas, cntl_trsm_blas, cntl_trsm_blas );


  FLA_Obj_create_conf_to( FLA_NO_TRANSPOSE, A, &A_old );

  FLA_Copy( A, A_old );

  m_A = FLA_Obj_width( A );
  m_B = m_A * percent_B / 100;
  m_E = m_A - m_B;

  FLA_Part_2x2( A, &B, &C, 
  		   &D, &E, m_B, m_B, FLA_TL );

  FLA_Part_2x1( p, &p_B,
   		   &p_E, m_B, FLA_TOP );

  FLA_Obj_create( FLA_DOUBLE, m_B, nb_alg, &Ls );

  FLA_Obj_create( FLA_INT, m_B, 1, &p_D );
  
  FLA_Obj_create( FLA_DOUBLE, m_A, 1, &x );

  FLA_Obj_create( FLA_DOUBLE, m_A, 1, &y );

  FLA_Random_matrix( x );

  FLA_Gemv( FLA_NO_TRANSPOSE, FLA_ONE, A, x, FLA_ZERO, y );

  FLA_Part_2x1( y, &y_T,
   		   &y_B, m_B, FLA_TOP );

  for ( irep = 0 ; irep < nrepeats; irep++ ){
    FLA_Copy( A_old, A );

    FLA_LU_piv( B, p_B );


    *dtime = FLA_Clock();

    switch( variant ){
    case 0:
      /* Time reference implementation */
      REF_LU_piv( A, p );

      break;

    case 3:{

      // Time variant 3 
      switch( type ){
      case FLA_ALG_UNBLOCKED:
	//	FLA_General_solve_2x2( FALSE, B, C,
	//			              D, E, p_B, p_D, p_E, Ls, 
	//			              y_T, y_B, nb_alg );

	//	FLA_LU_2x2( TRUE , B, C,
	FLA_LU_2x2( FALSE, B, C,
	 	           D, E, p_B, p_D, p_E, Ls, 
			         nb_alg );

        break;
      case FLA_ALG_BLOCKED:
	FLA_LU_piv( A, p );
        break;
      default:
        printf("trouble\n");
      }

      break;
    }

    case 4:{

      // Time variant 4 
      switch( type ){
      case FLA_ALG_UNBLOCKED:
        FLA_LU_piv_unb_var4( A, p );
        break;
      case FLA_ALG_BLOCKED:
        FLA_LU_piv_blk_var4( A, p, cntl_lu_var );
        break;
      default:
        printf("trouble\n");
      }

      break;
    }

    case 5:{

      // Time variant 5 
      switch( type ){
      case FLA_ALG_UNBLOCKED:
        FLA_LU_piv_unb_var5( A, p );
        break;
      case FLA_ALG_BLOCKED:
        FLA_LU_piv_blk_var5( A, p, cntl_lu_var );
        break;
      default:
        printf("trouble\n");
      }

      break;
    }
    }

    *dtime = FLA_Clock() - *dtime;
    dtime_old = min( *dtime, dtime_old );
  }

  FLA_Cntl_obj_free( cntl_lu_var );
  FLA_Cntl_obj_free( cntl_trsm_blas );
  FLA_Cntl_obj_free( cntl_gemm_blas );
  FLA_Blocksize_free( bp );


  if ( variant == 0 ){
    FLA_Copy( A, A_ref );
    *diff = 0.0;
  }
  else{
    if ( type != FLA_ALG_BLOCKED ) {
      *diff = FLA_Max_elemwise_diff( x, y );
    }
    else{
      *diff = FLA_Max_elemwise_diff( A, A_ref );
    }
  }

  *gflops = 2.0/3.0  * 
            FLA_Obj_length( A ) * 
            FLA_Obj_length( A ) * 
            FLA_Obj_length( A ) / 
            dtime_old / 1e9;

  *dtime = dtime_old;

  FLA_Copy( A_old, A );

  FLA_Obj_free( &A_old );
  FLA_Obj_free( &Ls );
  FLA_Obj_free( &p_D );
  FLA_Obj_free( &x );
  FLA_Obj_free( &y );
}

